/* $Id: Read_Input.c 1108 2011-01-23 18:10:46Z martin $ */

/* ===================================================================== */
/*     Read_Input.c                                                      */
/*                                                                       */
/* Routine to read crystallographic information and Fourier Map files    */
/* ===================================================================== */
/* Routines included in this file:
 *
 *   check_atom_name - compare two atom name-type variables
 *   convert_pos - convert string of form n/m to float
 *   end_flp - convert endedness of an integer
 *   end_flp_real - convert endedness of a float
 *   Error_Box - print error message to stderr
 *   get_next_token - extract the next CIF token from the file
 *   vec_dif - checks if two vectors are equal
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_ext.h"
#include <ctype.h>
int Block_CIF = 0;

/* ************************************************************** */
/* ************************************************************** */

int check_atom_name(char * name1, char * name2)
{
/* determine if two atom names are the same. Return 1 (TRUE) if they are */

    if (name1[0] == name2[0] && name1[1] == name2[1] &&
        name1[2] == name2[2] && name1[3] == name2[3])
        return 1;
    else
        return 0;
}

/* ************************************************************** */
/* ************************************************************** */

float convert_pos(char * string)
{
/* routine to convert an atom string to a float. If only digits, a simple sscanf
 *  is done. If the position is of the form n/m, each part is decoded and the
 * results of the division is returned.
 */

    float result;
    int i, n, m;
    char * pdest;
    char sub_string1[10], sub_string2[10];

    if (!(pdest = strstr(string,"/"))) {
        sscanf(string,"%f",&result);             /* no / in the input */
    } else {
        i = pdest - string +1;                   /* number of chars to copy */
        strncpy(sub_string1, string, i);
        sub_string1[i - 1] = '\0';
        strcpy(sub_string2, pdest + 1);          /* copy part after / */
        sscanf(sub_string1,"%d", &n);
        sscanf(sub_string2,"%d", &m);
        result = (float)n/(float)m;
    }
    return result;
}

/* ************************************************************** */
/* ************************************************************** */

int end_flip(int value)
{
/* routine to flip the endian nature of the input integer
 * This routine was copied from a Web article:
 *   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas
 */

   return (((value&0x000000FF)<<24)+((value&0x0000FF00)<<8)+
   ((value&0x00FF0000)>>8)+((value&0xFF000000)>>24));
}

/* ************************************************************** */
/* ************************************************************** */

float end_flip_real(float value)
{
/* routine to flip the endian nature of the input float 
 * This routine was copied from a Web article:
 *   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas
 */
   union u {float vi; unsigned char c[4];}; 
   union v {float ni; unsigned char d[4];};
   union u un; 
   union v vn; 
   un.vi = value;
   vn.d[0]=un.c[3]; 
   vn.d[1]=un.c[2]; 
   vn.d[2]=un.c[1]; 
   vn.d[3]=un.c[0]; 
   return (vn.ni); 

}

/* ************************************************************** */
/* ************************************************************** */

void Error_Box(char * string)
{
    fprintf(stderr,"%s\n",string);
}

/* ************************************************************** */
/* ************************************************************** */

int get_next_token( char *p, int max_len, FILE * fpin)
{
/* routine to return the next token in a stream. A token is the string
 * characters to the next white space consisting of a space, a tab, or
 * a newline character. Carriage returns are ignored. The returned value
 * is 1 (true) if a token is found and 0 (false) if the end of file is
 * reached */

    char *pp = p;
    int inchar;
    int started = 0;
    int openquote = -1;
    int len = max_len - 2;

    for (;;) {        /* start an infinite loop */
        if((inchar = getc(fpin)) == EOF)
            break;             /* here for end-of-file */
        if (inchar == '\t' || inchar == ' ') {
            if (started && openquote != 1) {
                *pp = '\0';       /* add null terminator */
                return 1;
            }
        } else if (inchar == '\n') {
            if (started) {
                *pp = '\0';
                return 1;
            }
        } else if (inchar == '#' && !started) { /* comment - skip to end of line */
           char comment[256];
           char *c = comment;
                for (;;) {
                  if((inchar = getc(fpin)) == EOF)
                    return 0;             /* here for end-of-file */
                  *c++=(char)inchar;
                  if (inchar=='\n') {
                    *c='\0';
                    if (!strncmp(comment," End of data for", 16))
                       return 0;
                    break;
                  }
                }
        } else if (inchar != '\r') {
            *pp++ = (char)inchar;
            started = 1;
            if (inchar == (char)39)
              openquote *= -1;
            if (!--len)
                return 0;
        }
    }
    return 0;
}

/* ************************************************************** */
/* ************************************************************** */

void find_lattice_type( void)
{
/* routine to scan symmetry elements, find the lattice type, and fill in the coorcdinates
   of the lattice points */

    int i;

    drvui->nbr=1;

    for (i = 1; i <= drvui->ng; ++i) {
        if (drvui->ss[i-1][0][0] == 1 && drvui->ss[i-1][1][1] == 1 &&
            drvui->ss[i-1][2][2] == 1) {  /* rotational part of symmetry is x,y,z - centering? */
            if (drvui->ts[i-1][0]== 0. && drvui->ts[i-1][1] == 0.5 &&
                drvui->ts[i-1][2] == 0.5) { /* found x,1/2+y,1/2+z */
                if (drvui->nbr == 3 || drvui->nbr == 4)
                    drvui->nbr = 5;  /* x,1/2+y,1/2+z && B || C ==> F */
                else
                    drvui->nbr = 2;  /* x,1/2+y,1/2+z ==> A */
            }
            if (drvui->ts[i-1][0]== 0.5 && drvui->ts[i-1][1] == 0. &&
                drvui->ts[i-1][2] == 0.5) { /* found 1/2+x,y,1/2+z */
                if (drvui->nbr == 2 || drvui->nbr == 4)
                    drvui->nbr = 5;  /* 1/2+x,y,1/2+z && A || C ==> F */
                else
                    drvui->nbr = 3;  /* 1/2+x,y,1/2+z ==> B */
            }
            if (drvui->ts[i-1][0]== 0.5 && drvui->ts[i-1][1] == 0.5 &&
                drvui->ts[i-1][2] == 0.) { /* found 1/2+x,1/2+y,z */
                if (drvui->nbr == 2 || drvui->nbr == 3)
                    drvui->nbr = 5;  /* 1/2+x,1/2+y,z && A || B ==> F */
                else
                    drvui->nbr = 4;  /* 1/2+x,1/2+y,z ==> C */
            }
            if (drvui->nbr==1 && drvui->ts[i-1][0]==0.5 && drvui->ts[i-1][1]==0.5 &&
                drvui->ts[i-1][2]==0.5)
                   drvui->nbr=6;  /* 0.5+X,0.5+Y,0.5+Z ==> I */

            if (drvui->ts[i-1][0]<0.5 &&
                drvui->ts[i-1][0]>0.25 )
                drvui->nbr=7;  /* 0.33+X ==> R */
        }

    }

    switch (drvui->nbr) {
    case 1:
        drvui->nlat = 1;
        break;
    case 2:
        drvui->nlat = 2;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
    case 3:
        drvui->nlat = 2;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.0;
        drvui->lat_pos[1][2] = 0.5f;
        break;
    case 4:
        drvui->nlat = 2;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.0;
        break;
    case 5:
        drvui->nlat = 4;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        drvui->lat_pos[2][0] = 0.5f;
        drvui->lat_pos[2][1] = 0.0;
        drvui->lat_pos[2][2] = 0.5f;
        drvui->lat_pos[3][0] = 0.5f;
        drvui->lat_pos[3][1] = 0.5f;
        drvui->lat_pos[3][2] = 0.0;
        break;
    case 6:
        drvui->nlat = 2;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
    case 7:
        drvui->nlat = 3;
        drvui->lat_pos[1][0] = (1.f / 3.f);
        drvui->lat_pos[1][1] = (2.f / 3.f);
        drvui->lat_pos[1][2] = (2.f / 3.f);
        drvui->lat_pos[2][0] = (2.f / 3.f);
        drvui->lat_pos[2][1] = (1.f / 3.f);
        drvui->lat_pos[2][2] = (1.f / 3.f);
    }
}

/* ************************************************************** */
/* ************************************************************** */

void read_inp (int Quick)

/* routine to read and process input file */
{
  char input[256];				 /* input buffer for data read from file */
  char input1[256];                              /* temporary buffer used for printing */
  char input2[256];                              /* input buffer for data read from file */
  char string[100];				 /* temporary string */
  char filename[256];                            /* used for map file name */
  int intype;					 /* type of datum read */
  int i, j, k;					 /* temporary */
  char t_name[4];				 /* temporary atom name storage */
  char t_color[40];				 /* temporary ellipsoid color storage */
  int in_line;					 /* flag for inline import of foreign files */
  float t_rad;
  int n_el_tmp = 0;
  int done;
  int average = 0;
  char modl[5];
  int modnum;
  float avg_occ,min_occ;
  int tmp_frame_no = 1;
  char atom_pos[3][30];
  char filetype[10];
  char string2[20], res[20];

/* ************************************************************** */
/* ************************************************************** */
/*
    Commands Implemented

    arrow
    atom
    average
    axislines
    background
    betaij
    bestplane
    bij,Bij
    bond
    bounds
    box
    cell
    clip
    cutout
    dash
    depthcue
    edges
    ellipcolor
    ellipsoids
    finish
    frame
    import
    inline
    labelscale
    labeltext
    list
    lonepair
    lookat
    magtrans
    magnification
    mapcalclimits
    mapcontour
    mapcontour2d
    mapread
    mapregion
    mapslice
    molcomp
    nolabels
    noshadow
    occupancy
    origin
    orthographic
    pack
    phaseshift
    phong
    plane
    polyedge
    polysz
    polytolerance
    polylimit
    polyfudge
    polyvert
    rem, REM
    shell
    slab
    special
    sgrp, spgp, spgr
    sphere
    title, titl
    uij,Uij
    values
    vectors    
    view
    vrml1
    vrml2
    vrml97
    x3d
    xyzoff
    end, END
                                                                  */
/* ************************************************************** */
/* ************************************************************** */

  Map_Info.xlim[0] = 0.0f;
  Map_Info.xlim[1] = 1.0f;
  Map_Info.ylim[0] = 0.0f;
  Map_Info.ylim[1] = 1.0f;
  Map_Info.zlim[0] = 0.0f;
  Map_Info.zlim[1] = 1.0f;
  drvui->origin1_flag = 0;
  drvui->numOfFourierContours = 0;
  drvui->frames[drvui->frame_no].slice = 0;
        
  for (i = 0; i < 3; i++)
       drvui->xyzoff[i] = 0.0f;
  if (drvui->frame_no > 1) {
    if (packflag)
       for (i=0; i<6; i++) {
           drvui->frames[drvui->frame_no].cryst_lim[i] = drvui->frames[drvui->frame_no-1].cryst_lim[i];
           drvui->frames[drvui->frame_no].clip_lim[i] = drvui->frames[drvui->frame_no-1].clip_lim[i];
       }
    else
       for (i=3; i<6; i++) {
           drvui->frames[drvui->frame_no].cryst_lim[i] = 1.;
       }
  }
  
  while (!feof (drvui->fpin)) {
    memset(input,0,255);
    memset(input1,0,255);
    memset(string,0,100);
    memset(t_color,0,40);
    if (fgets (input, 255, drvui->fpin)) {		 /* read a line */
      trim_string(input, 256);
      Blank_Strip (input);			 /* suppress any input blanks */
      strcpy(input2,input);                      /* make a copy for listing file */
      input[255] = 0;                   /* force a terminator */
    } else {
      fprintf (stderr, "Error reading file.");
      delay (5);
      exit (1);
    }
    intype = 0;
    strcpy(input1,input);
    if (strncmp (input, "title", 5) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
        else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "titl", 4) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "TITL", 4) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "rem", 3) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore rem command if not current frame */
    }
    if (strncmp (input, "REM", 3) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore rem command if not current frame */
    }
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "sgrp", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore sgrp command if not current frame */
    }
    if (strncmp (input, "spgp", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore spgp command if not current frame */
    }
    if (strncmp (input, "spgr", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore spgr command if not current frame */
    }
    if (strncmp (input, "ellipsoids", 10) == 0)
      intype = 4;
    if (strncmp (input, "lookat", 6) == 0)
      intype = 5;
    if (strncmp (input, "atom", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 6;
      else
          intype = 0;                     /* ignore atom command if not current frame */
    }
    if (strncmp (input, "polysz", 6) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 7;
      else
          intype = 0;                       /* ignore polysz command if not current frame */
    }
    if (strncmp (input, "sphere", 6) == 0) {
      if(tmp_frame_no == drvui->frame_no) 
          intype = 8;
      else
          intype = 0;                       /* ignore sphere command if not current frame */
    }
    if (strncmp (input, "end", 3) == 0)
      intype = 9;
    if (strncmp (input, "END", 3) == 0)
      intype = 9;
    if (strncmp (input, "xyzoff", 6) == 0)
      intype = 10;
    if (strncmp (input, "bond", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 11;
      else
          intype = 0;                       /* ignore bond command if not current frame */
    }
    if (strncmp (input, "box", 3) == 0)
      intype = 12;
    if (strncmp (input, "edges", 5) == 0)
      intype = 13;
    if (strncmp (input, "view", 4) == 0)
      intype = 14;
    if (strncmp (input, "plane", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 15;
      else
          intype = 0;                       /* ignore plane command if not current frame */
    }
    if (strncmp (input, "bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "Bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "betaij", 6) == 0)
      intype = 17;
    if (strncmp (input, "uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "Uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "bounds", 6) == 0)
      intype = 19;
    if (strncmp (input, "pack", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 20;
      else
          intype = 0;                       /* ignore pack command if not current frame */
    }
    if (strncmp (input, "magnification", 13) == 0)
      intype = 21;
    if (strncmp (input, "origin", 6) == 0)
      intype = 22;
    if (strncmp (input, "ellipcolor", 10) == 0)
      intype = 23;
    if (strncmp (input, "import", 6) == 0) {
      if(tmp_frame_no == drvui->frame_no)
        intype = 24;
      else
        intype = 0;                     /* ignore import command if not current frame */
    }
    if (strncmp (input, "inline", 6) == 0) {
     if(tmp_frame_no == drvui->frame_no) {
       intype = 25;
     } else {    /* skip ahead to next frame or end */
       int end_seen=0; /* some formats have their own END card */
       while(!feof(drvui->fpin) && intype!=0 ) {
         if (fgets (input, 255, drvui->fpin)) {         /* read a line */
            trim_string(input,256);          /* kill any ^M or ^J at end of line */
            Blank_Strip (input);             /* suppress any leading blanks */
            input[255] = 0;                   /* force a terminator */
            if (strncmp (input, "frame", 5) == 0) {
              intype=37;
              break;
            }
            if (strncmp (input, "end", 3) == 0) {
             if (strstr(input2,"schakal") && end_seen==0 ) {
                end_seen=1;
              } else {
                intype=9;
                break;
              }
            }
          } else {
            Error_Box("Error reading file.");
            return;
          }
        } /* advance to next frame or end command */
      } /* if not in current frame */
    }
    if (strncmp (input, "orthographic", 12) == 0)
      intype = 26;
    if (strncmp (input, "vectors", 7) == 0)
      intype = 27;
    if (strncmp (input, "phong", 5) == 0)
      intype = 28;
    if (strncmp (input, "molcomp", 7) == 0)
      intype = 29;
    if (strncmp (input, "cutout", 6) == 0)
      intype = 30;
    if (strncmp (input, "depthcue", 8) == 0)
      intype = 31;
    if (strncmp (input, "vrml1", 5) == 0)
      intype = 32;
    if (strncmp (input, "vrml2", 5) == 0)
      intype = 1;
    if (strncmp (input, "vrml97", 6) == 0)
      intype = 1;
    if (strncmp (input, "polytol", 7) == 0)
      intype = 33;
    if (strncmp (input, "polylimit", 9) == 0)
      intype = 33;
    if (strncmp (input, "polyfudge", 9) == 0)
      intype = 33;
    if (strncmp (input, "clip", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 34;
      else
          intype = 0;                       /* ignore clip command if not current frame */
    }
    if (strncmp (input, "list", 4) == 0)
      intype = 35;
    if (strncmp (input, "polyedge", 8) == 0)
      intype = 36;
    if (strncmp (input, "frame", 5) == 0)
      intype = 37;
    if (strncmp (input, "axislines", 9) == 0)
      intype = 38;
    if (strncmp (input, "shell", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 39;
      else
          intype = 0;                       /* ignore shell command if not current frame */
    }
    if (strncmp (input, "nolabels", 8) == 0)
      intype = 40;
    if (strncmp (input, "lonepair", 8) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 41;
      else
          intype = 0;                       /* ignore lonepair command if not current frame */
    }
    if (strncmp (input, "labeltext", 9) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 42;
      else
          intype = 0;                     /* ignore labeltext command if not current frame */
    }
    if (strncmp (input, "polyvert", 8) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 43;
      else
          intype = 0;                       /* ignore polyvert command if not current frame */
    }
    if (strncmp (input, "background", 10) == 0)
      intype = 44;
    if (strncmp (input, "slab", 4) == 0)
      intype = 45;
    if (strncmp (input, "mag_trans", 9) == 0)
      intype = 46;
    if (strncmp (input, "special", 7) == 0)
      intype = 47;
    if (strncmp (input, "noshadow", 8) == 0)
      intype = 48;
    if (strncmp (input, "finish", 6) == 0)
      intype = 49;
    if (strncmp (input, "arrow", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 50;
        else
          intype = 0;                     /* ignore arrow command if not current frame */
    }
    if (strncmp (input, "dash", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 51;
      else
          intype = 0;                     /* ignore dash command if not current frame */
    }
    if (strncmp (input, "mapcontour", 10) == 0)
      intype = 52;
    if (strncmp (input, "mapcontour2d", 12) == 0)
      intype = 61;
    if (strncmp (input, "mapread", 7) == 0) {
      if (tmp_frame_no == drvui->frame_no)
          intype = 53;
      else
          intype = 0;
    }
    if (strncmp (input, "mapregion", 9) == 0) {
      if (tmp_frame_no == drvui->frame_no)
          intype = 54;
      else
          intype = 0;
    }
    if (strncmp (input, "mapcalclimits", 13) == 0)
      intype = 55;
    if (strncmp (input, "labelscale", 10) == 0)
      intype = 56;
    if (strncmp (input, "bestplane", 9) == 0)
      intype = 57;
    if (strncmp (input, "average", 7) == 0)
          intype = 58;
    if (strncmp (input, "occupancy", 9) == 0)
          intype = 59;
    if (strncmp (input, "phaseshift", 10) == 0) {
          if (drvui->automation == 1) 
              intype = 0;
          else
          intype = 60;
    }      
    if (strncmp (input, "aimsurf", 7) == 0)
          intype = 62;
    if (strncmp (input, "voids", 5) == 0)
          intype = 63;
    if (strncmp (input, "values", 6) == 0)
          intype = 64;         
    if (strncmp (input, "mapslice", 8) == 0) {
          if (tmp_frame_no == drvui->frame_no)
              intype = 65;
          else
              intype = 0;
    }
    if (strncmp (input, "qvector", 7) == 0)
          intype = 66;
    if (strncmp (input, "maplegend", 9) == 0)
          intype = 67;
    if (strncmp (input, "x3d", 3) == 0)
          intype = 68;
             
    switch (intype) {
    case 0:
      break;
    case 1:
      break;					 /* ignore title except for listing */
    case 2:                      /* cell command */
      for (i = 3; i <= 5; ++i)
          drvui->lat_con[i] = 0.0f;
      (void) sscanf (input, "%*s %f %f %f %f %f %f", &drvui->lat_con[0]
		     , &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4]
		     , &drvui->lat_con[5]);
      break;
    case 3:
      symop (input);				 /* process space group information */
      break;
    case 4:					 /* ellipsoids command */
      do_ellipsoids = 1;
      (void) sscanf (input, "%*s %f", &drvui->Ellipsoid_Prob);
      if (drvui->Ellipsoid_Prob > 1)
        drvui->Ellipsoid_Prob *= 0.01f;		 /* Probability to fraction */
      break;
    case 5:					 /* read vectors for lookat command */
      (void) sscanf (input, "%*s %f %f %f %f %f %f", &drvui->lookat_v1[0]
		     , &drvui->lookat_v1[1], &drvui->lookat_v1[2], &drvui->lookat_v2[0],
             &drvui->lookat_v2[1], &drvui->lookat_v2[2]);
      Options = Options | L_OPT;
      break;
    case 6:					 /* atom command */
      get_label (input, &drvui->atoms[natom].atom_l[0], &drvui->atoms[natom].atom_l[1],
		 &drvui->atoms[natom].atom_l[2], &drvui->atoms[natom].atom_l[3], 1);
      (void)sscanf (input, "%d %s %s %s",&drvui->atoms[natom].atom_n,atom_pos[0],atom_pos[1],atom_pos[2]);
      for (i=0; i<3; i++)
          drvui->atoms[natom].atom_xyz[i] = convert_pos(atom_pos[i]);
      drvui->atoms[natom].atom_fn = tmp_frame_no;
      natom++;
      check_dynamic_storage();
      break;
    case 7:				 /* polysz command */
    case 39:				 /* shell command */
    case 43:				 /* polyvert command */
      get_label (input, &drvui->polyhedra[drvui->npoly].poly_l[0],
                 &drvui->polyhedra[drvui->npoly].poly_l[1],
                 &drvui->polyhedra[drvui->npoly].poly_l[2],
                 &drvui->polyhedra[drvui->npoly].poly_l[3], 1);
      if (intype == 7) {
          (void) sscanf (input, "%f %39c", &drvui->polyhedra[drvui->npoly].poly_size, t_color);
      } else if (intype == 39) {
          (void) sscanf (input, "%f %f %39c",
	                 &drvui->polyhedra[drvui->npoly].poly_min,
                         &drvui->polyhedra[drvui->npoly].poly_size, t_color);
      } else if (intype == 43) {
          get_label (input, &drvui->polyhedra[drvui->npoly].poly_t[0],
                     &drvui->polyhedra[drvui->npoly].poly_t[1],
                     &drvui->polyhedra[drvui->npoly].poly_t[2],
                     &drvui->polyhedra[drvui->npoly].poly_t[3], 0);
          (void) sscanf (input, "%f %39c", &drvui->polyhedra[drvui->npoly].poly_size, t_color);
      }
      trim_string(t_color, 40);
      strcpy(drvui->polyhedra[drvui->npoly].poly_col,t_color);
      if (intype == 7 || intype == 43)
          drvui->polyhedra[drvui->npoly].poly_min = 0.0f;
      drvui->polyhedra[drvui->npoly].poly_fn = tmp_frame_no;
      if (intype == 7 || intype == 39)
          strcpy(drvui->polyhedra[drvui->npoly].poly_t,"");
      drvui->npoly++;
      check_dynamic_storage();
      break;
    case 8:					 /* sphere command */
    { char str[9][40], *pnt;
      double inpval[9];
      memset(str,0,360*sizeof(char));
      get_label (input, &drvui->spheres[drvui->nsphere].sphere_l[0],
                 &drvui->spheres[drvui->nsphere].sphere_l[1],
                 &drvui->spheres[drvui->nsphere].sphere_l[2],
                 &drvui->spheres[drvui->nsphere].sphere_l[3], 1);
      i = sscanf(input,"%s %s %s %s %s %s %s %s %s", str[0], str[1], str[2], str[3], str[4],
                 str[5], str[6], str[7], str[8]);
      for (j=0; j<i; j++)
          inpval[j] = strtod(str[j],&pnt);
      drvui->spheres[drvui->nsphere].sphere_n = -1;
      if (i == 2 || i == 4 || i == 6) {
          drvui->spheres[drvui->nsphere].sphere_size = (float)inpval[0];
          j = 1;
      } else if (i == 3 || i == 5 || i == 7) {
          drvui->spheres[drvui->nsphere].sphere_n = (int)(inpval[0] + 0.01);
          drvui->spheres[drvui->nsphere].sphere_size = (float)inpval[1];
          j = 2;
      } else {
          Error_Box("Improperly formed sphere command.");
      }
      strcpy(string,str[j]);
      for (k=j+1; k<i; k++) {
          strcat(string," ");
          strcat(string,str[k]);
      }
      trim_string(string,40);
      strcpy(drvui->spheres[drvui->nsphere].sphere_col, string);
      drvui->spheres[drvui->nsphere].sphere_fn = drvui->frame_no;
      drvui->nsphere++;
      check_dynamic_storage();

      break; }
    case 9:					 /* end or END */
      if (!drvui->lat_con[0]) {
        fprintf(stderr,"There is no 'cell' line in your input file.\n");
        exit(0);
      }
      for (i = 0; i < natom; i++) {		 /* shift atom coordinates for origin offset */
          for (j = 0; j < 3; j++) {
            drvui->atoms[i].atom_xyz[j] -= drvui->xyzoff[j];
          }
      }
      if (average == 1) drvui->modulated = -1;
      drvui->max_frame = tmp_frame_no;
      if (!Quick)
          fprintf(drvui->flout,"End of Frame %d\n",drvui->frame_no);
      if (!Quick && drvui->frame_no == drvui->max_frame)
          fprintf (drvui->flout, "Final Frame done\n");
      if (n_el_tmp > 0) {
          for (i = 0; i < n_el_tmp; i++) {
            get_label (drvui->ellips[i].el_color_tmp, &t_name[0], &t_name[1],
		       &t_name[2], &t_name[3], 0);
            if (drvui->ellips[i].el_color_tmp[0] == '*') {
              (void) sscanf (drvui->ellips[i].el_color_tmp, "%*s %s", t_color);
              j = -1;
            } else {
              (void) sscanf (drvui->ellips[i].el_color_tmp, "%d %s", &j, t_color);
            }
            for (k = 0; k < drvui->n_ellips; k++) {	 /* find this atom in ellipsoid list */
              if ((t_name[0] == drvui->ellips[k].ellips_l[0]) &&
                  (t_name[1] == drvui->ellips[k].ellips_l[1]) &&
                  (t_name[2] == drvui->ellips[k].ellips_l[2]) &&
                  (t_name[3] == drvui->ellips[k].ellips_l[3])) {
                if ((j == -1) || (j == drvui->ellips[k].ellips_n)) {
                  strcpy (drvui->ellips[k].ellips_col, t_color);	/* copy color */
                  drvui->ellips[k].ell_type += 1000;                  /* >= 1000 means ellipcolor read */
                  drvui->ellips[k].save_el_number=j;                  
                }
              }
           }
          }
      }
      if (drvui->natprop > 0) {
          for (i = 0; i < drvui->natprop; i++) {
            j = drvui->atprops[i].atprop_n;
              for (k = 0; k < natom; k++) {       /* find this atom in list */
                if ((j == -1) || (j == drvui->atoms[k].atom_n)) {
                  if (check_atom_name (drvui->atprops[i].atprop_l, drvui->atoms[k].atom_l)) {
                    drvui->atoms[k].radius = drvui->atprops[i].radius;
                  }
                }
              }
          }
      }                                                                        
      for (i=1; i<drvui->nsphere; i++) {        /* remove any auto-generated ellipsoids that */
          for (j=1; j<drvui->n_ellips; j++) {          /*     interfere with spheres */
           if (drvui->ellips[j].ell_type==1001) {
              if (check_atom_name(drvui->spheres[i].sphere_l,drvui->ellips[j].ellips_l)) {
                  if (drvui->spheres[i].sphere_n == -1
                      || drvui->spheres[i].sphere_n == drvui->ellips[j].save_el_number) {
                      drvui->ellips[j].ell_type = 1;
                  }
              }
            }
          }
      }
      switch (FourierMapType) {
      case 1:
          read_grd( FourierFileName, Quick);
          break;
      case 2:
          read_stf( FourierFileName, Quick);
          break;
      case 3:
          read_w2k( FourierFileName, Quick);
          break;
      case 4:
          read_vasp( FourierFileName, Quick);
          break;
      case 5:
          read_flp( FourierFileName, Quick);
          break;
      case 6:
          read_fcf( FourierFileName, Quick);
          break;
      case 7:
          read_dn6( FourierFileName, Quick);
          break;
      case 8:
          read_m80( FourierFileName, Quick);
          break;
      case 9:
          read_exc( FourierFileName, Quick);
          break;
      case 10:
          read_m81( FourierFileName, Quick);
          break;
      case 11:
          read_xsf( FourierFileName, Quick);
          break;
      default:
          break;
      }
      for (i = 1; i <= drvui->max_frame; i++) {
        if (drvui->frames[i].map_lim_set == 0) {    /* mapregion line NOT read */
          xMin = Map_Info.xlim[0];
          xMax = Map_Info.xlim[1];
          yMin = Map_Info.ylim[0];
          yMax = Map_Info.ylim[1];
          zMin = Map_Info.zlim[0];
          zMax = Map_Info.zlim[1];
          x4Val = x5Val = x6Val = 0.0f;
          drvui->frames[i].map_lim[0]=xMin;
          drvui->frames[i].map_lim[3]=xMax;
          drvui->frames[i].map_lim[1]=yMin;
          drvui->frames[i].map_lim[4]=yMax;
          drvui->frames[i].map_lim[2]=zMin;
          drvui->frames[i].map_lim[5]=zMax;
          drvui->frames[i].map_lim[6]=x4Val;
          drvui->frames[i].map_lim[7]=x5Val;
          drvui->frames[i].map_lim[8]=x6Val;
        }
        if (drvui->frames[i].slice >0 ) { /* calculate 2d plane in cartesian space */
          float p[3];
          vnormalize(drvui->frames[i].mapnorm);
          drvui->frames[i].planeeq[0] = (float)drvui->frames[i].mapnorm[0];
          drvui->frames[i].planeeq[1] = (float)drvui->frames[i].mapnorm[1];
          drvui->frames[i].planeeq[2] = (float)drvui->frames[i].mapnorm[2];
          make_bmat (drvui->sys, drvui->lat_con, drvui->b_mat, drvui->ginv,drvui->rec_lat_con);
          Convert_Cryst_Cart(drvui->b_mat, drvui->frames[i].mapslice, p, origin);
          drvui->frames[i].planeeq[3] = -(float) (drvui->frames[i].mapnorm[0] * p[0]
                                            + drvui->frames[i].mapnorm[1] * p[1]
                                            + drvui->frames[i].mapnorm[2] * p[2]);
        }
      }
      if (drvui->subsys_vol[0] == 1.0f)
          drvui->subsys_vol[0] = drvui->subsys_ref_volume;
      return;					 /* thru with end */
    case 10:					 /* xyzoff command */
      (void)sscanf (input, "%*s %s %s %s",atom_pos[0],atom_pos[1],atom_pos[2]);
      for (i=0; i<3; i++)
          drvui->xyzoff[i] = convert_pos(atom_pos[i]);
      break;
    case 51:                     /* dash command */
    case 11:					 /* bond command */
      if (intype == 51) {
        k = sscanf (input, "%*s %d", &j);
        if (k == 2)     /* number of segments if given */
          Token_Strip (input, 1);
        else
          j = 5;
      }
      get_label (input, &drvui->bonds[drvui->nbond].bond_l1[0],
                 &drvui->bonds[drvui->nbond].bond_l1[1],
		 &drvui->bonds[drvui->nbond].bond_l1[2],
                 &drvui->bonds[drvui->nbond].bond_l1[3], 1);
      get_label (input, &drvui->bonds[drvui->nbond].bond_l2[0],
                 &drvui->bonds[drvui->nbond].bond_l2[1],
		 &drvui->bonds[drvui->nbond].bond_l2[2],
                 &drvui->bonds[drvui->nbond].bond_l2[3], 0);
      (void) sscanf (input, "%f %f %f %39c",
		     &drvui->bonds[drvui->nbond].bond_size,
                     &drvui->bonds[drvui->nbond].bond_min,
                     &drvui->bonds[drvui->nbond].bond_max,t_color);
      trim_string(t_color, 40);
      strcpy(drvui->bonds[drvui->nbond].col_bond,t_color);
      if (strlen (drvui->bonds[drvui->nbond].col_bond) == 0)
          strcpy (drvui->bonds[drvui->nbond].col_bond, "Grey");
      drvui->bonds[drvui->nbond].bond_fn = tmp_frame_no;
      if (intype == 11)
          drvui->bonds[drvui->nbond].bond_style = 0;
      else 
          drvui->bonds[drvui->nbond].bond_style = j;
      drvui->nbond++;
      check_dynamic_storage();
      break;
    case 12:					 /* box command */
      (void) sscanf (input, "%*s %f %39c", &rad_cell, t_color);
      trim_string(t_color, 40);
      strcpy(drvui->col_cell,t_color);
      if (strlen (drvui->col_cell) == 0)
          strcpy (drvui->col_cell, "Black");
      if (rad_cell == 0.0f) {
          docell = 0;
      }
      break;
    case 13:					 /* edges command */
      memset(drvui->col_edge,0,40);
      (void) sscanf (input, "%*s %f %39c", &rad_edge, drvui->col_edge);
      trim_string(drvui->col_edge, 40);
      edges = 1;
      if (rad_edge < 0.01f)
          rad_edge = 0.01f;
      if (strlen (drvui->col_edge) == 0)
          strcpy (drvui->col_edge, "Black");
      break;
    case 14:					 /* view command */
      if ((Options & V_OPT) == 0)		 /* only if no -v switch */
          (void) sscanf (input, "%*s %f %f %f", &xrot, &yrot, &zrot);
      xrot_i = xrot;
      yrot_i = yrot;
      zrot_i = zrot;
      break;
    case 15:					 /* plane command */
      get_label (input, &drvui->planes[drvui->nplane].plane_l[0],
                 &drvui->planes[drvui->nplane].plane_l[1],
		 &drvui->planes[drvui->nplane].plane_l[2],
                 &drvui->planes[drvui->nplane].plane_l[3], 1);
      (void) sscanf (input, "%f %39c", &drvui->planes[drvui->nplane].plane_size, t_color);
      trim_string(t_color, 40);
      strcpy(drvui->planes[drvui->nplane].plane_col, t_color);
      drvui->planes[drvui->nplane].plane_fn = drvui->frame_no;
      drvui->nplane++;
      check_dynamic_storage();
      break;
    case 16:					 /* bij or Bij command */
      drvui->ellips[drvui->n_ellips].ell_type = -1;
    case 17:					 /* betaij command */
      if (intype == 17)
          drvui->ellips[drvui->n_ellips].ell_type = 0;
    case 18:					 /* uij or Uij command */
      if (intype == 18)
          drvui->ellips[drvui->n_ellips].ell_type = 1;
      get_label (input, &drvui->ellips[drvui->n_ellips].ellips_l[0],
                 &drvui->ellips[drvui->n_ellips].ellips_l[1],
		 &drvui->ellips[drvui->n_ellips].ellips_l[2],
                 &drvui->ellips[drvui->n_ellips].ellips_l[3], 1);
      (void) sscanf (input, "%d %f %f %f %f %f %f %39c", &drvui->ellips[drvui->n_ellips].ellips_n, 
                     &drvui->ellips[drvui->n_ellips].ellips[0],
                     &drvui->ellips[drvui->n_ellips].ellips[1],
                     &drvui->ellips[drvui->n_ellips].ellips[2], 
                     &drvui->ellips[drvui->n_ellips].ellips[3],
                     &drvui->ellips[drvui->n_ellips].ellips[4],
                     &drvui->ellips[drvui->n_ellips].ellips[5], t_color);
      trim_string(t_color, 40);
      strcpy(drvui->ellips[drvui->n_ellips].ellips_col, t_color);	/* read unique coefficients and color */
      if (!strlen (drvui->ellips[drvui->n_ellips].ellips_col))
          strcpy (drvui->ellips[drvui->n_ellips].ellips_col, "Gray80");
      drvui->n_ellips++;
      check_dynamic_storage();
      break;
    case 19:					 /* bounds command */
      if ((Options & B_OPT) == 0) {		 /* only do this if no -b switch */
          (void) sscanf (input, "%*s %f %f %f", &boxlim[0], &boxlim[1],
		       &boxlim[2]);
          boxflag = 1;
      }
      break;
    case 20:					 /* pack command */
      if ((Options & P_OPT) == 0) {		 /* only if no -p switch */
          (void) sscanf (input, "%*s %f %f %f %f %f %f", 
              &drvui->frames[drvui->frame_no].cryst_lim[0], &drvui->frames[drvui->frame_no].cryst_lim[3],
              &drvui->frames[drvui->frame_no].cryst_lim[1], &drvui->frames[drvui->frame_no].cryst_lim[4],
              &drvui->frames[drvui->frame_no].cryst_lim[2], &drvui->frames[drvui->frame_no].cryst_lim[5]);
          if (!clipflag) {
              for (i=0; i<6; i++) {
                  drvui->frames[drvui->frame_no].clip_lim[i] = drvui->frames[drvui->frame_no].cryst_lim[i];
              }
          }
          packflag = 1;
      }
      break;
    case 21:					 /* magnification command */
      if ((Options & M_OPT) == 0)		 /* only if no -m switch */
          (void) sscanf (input, "%*s %f", &Magnification);
      break;
    case 22:					 /* origin command */
      if ((Options & O_OPT) == 0) {		 /* only if no -o switch */
          (void) sscanf (input, "%*s %f %f %f", &origin[0], &origin[1],
		       &origin[2]);
      }
      break;
    case 23:					 /* ellipcolor command */
      Token_Strip (input, 1);
      trim_string(input, 256);
      strcpy (drvui->ellips[n_el_tmp++].el_color_tmp, input);
      break;
    case 24:					 /* inline command */
    case 25:					 /* import command */
      in_line = 0;
      done = 0;
      if (average == 1) drvui->modulated=-1;
      if (!Quick)
        fprintf (drvui->flout, "*         %s\n======================= Start of import/inline commands\n",
            input2);     /* echo to listing file */
      if (strncmp (input, "inline", 6) == 0)
          in_line = 1;
      intype = 0;
      Token_Strip (input, 1);			 /* strip command */
      sscanf(input,"%s",filetype); 
      Token_Strip (input, 1);
      trim_string(input,256);
      if (strncmp (filetype, "gsas", 4) == 0) {
          import_gsas (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "pcr", 3) == 0) {
          import_pcr (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      Block_CIF = 1;
      if (strncmp (filetype, "cif", 3) == 0) {
          int ii,jj;
          ii=strlen(input)-1;
          if (isdigit(input[ii])) { /* if number at end */
              for (jj=ii;jj>ii-6;jj--) {
                  if (isspace(input[jj])) { /* separate number */
                      input[jj]='\0';
                      Block_CIF=atoi(&input[jj+1]);
                      break;
                  }
                  if (isalpha(input[jj]))
                      break; /* unless it is attached to filename */
              }
          } else {
            Block_CIF = 0;
          }
          import_cif (input, in_line, Quick, &Block_CIF, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "schakal", 7) == 0) {
          import_schakal (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "shakal", 6) == 0) {
          import_schakal (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "shelx", 5) == 0) {
          import_shelx (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "fdat", 4) == 0) {
          import_fdat (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "csd", 3) == 0) {
          import_fdat (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "ccdf", 4) == 0) {
          import_fdat (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "wien2k", 6) == 0) {
          import_wien (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "discus", 6) == 0) {
          import_discus (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (strncmp (filetype, "exciting", 8) == 0 || strncmp(filetype, "elk", 3) ==0) {
          import_exc (input, in_line, Quick, tmp_frame_no);
          done = 1;
      }
      if (done == 0) {
          fprintf (stderr,
              "Import/inline source type '%s' not recognized.\n", input);
      }
      if (!Quick)
        fprintf (drvui->flout, "======================= End of import/inline commands\n");
      break;
    case 26:					 /* orthographic */
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
      break;
    case 27: {                   /* unit cell direction vectors */
      float v[3];
      Display_axes = 1;
      offset[0] = offset[1] = offset[2] = 0.0f;
      i = sscanf( input, "%*s %f %f %f", &v[0], &v[1], &v[2]);
      if (i == 4)
          for (i=0; i<3; i++)
              offset[i] = v[i];
      }
      break;
    case 28:					 /* phong */
      (void) sscanf (input, "%*s %f %f", &drvui->Phong_Value, &drvui->Phong_Size);
      break;
    case 29:					 /* molcomp */
      (void) sscanf (input, "%*s %f", &drvui->mol_d);
      domolcomp = 1;
      break;
    case 30:					 /* cutout */
      El_Cutout = 1;
      memset(drvui->Cutout_color,0,40);
      (void) sscanf (input, "%*s %39c", drvui->Cutout_color);
      trim_string(drvui->Cutout_color, 40);
      break;
    case 31:					 /* depthcue - thickness of edges scaled by Z coordinate */
      (void) sscanf (input, "%*s %f", &DepthCue);
      break;
    case 32:					 /* vrml1 - turn on VRML1 output */
      Vrml2 = 0;
      break;
    case 33:					 /* polylimit (formerly polyfudge) */
      (void) sscanf (input, "%*s %f", &drvui->polylimit);
      if (!Quick)
          fprintf (drvui->flout,"Tolerance factor for non-planarity of polyhedra faces "
              "set to %f. This option should be used with care\n",drvui->polylimit);
      break;
    case 34:					 /* clip command */
      (void) sscanf (input, "%*s %f %f %f %f %f %f", &drvui->frames[drvui->frame_no].clip_lim[0],
		     &drvui->frames[drvui->frame_no].clip_lim[3], &drvui->frames[drvui->frame_no].clip_lim[1],
                     &drvui->frames[drvui->frame_no].clip_lim[4], &drvui->frames[drvui->frame_no].clip_lim[2],
                     &drvui->frames[drvui->frame_no].clip_lim[5]);
      clipflag = 1;
      break;
    case 35:					 /* list - limit for printed distance table */
      (void) sscanf (input, "%*s %f", &printdist);
      if (!Quick)
          fprintf (drvui->flout,"Interatomic distances up to %f Angstroms will be "
               "tabulated in the logfile\n",printdist);
      break;
    case 36:					 /* polyedge - enables individual coloring of polyhedral edges */
      get_label (input, &t_name[0], &t_name[1], &t_name[2], &t_name[3], 1);
      (void) sscanf (input, "%f %39c", &t_rad, t_color);
      trim_string(t_color, 40);
      for (i = 1; i < drvui->npoly; i++) {		 /* find this atom in poly list */
          if ((t_name[0] == drvui->polyhedra[i].poly_l[0]) &&
              (t_name[1] == drvui->polyhedra[i].poly_l[1]) &&
              (t_name[2] == drvui->polyhedra[i].poly_l[2]) && (t_name[3] == drvui->polyhedra[i].poly_l[3])) {
              edges = 1;
              drvui->polyhedra[i].poly_rad_edge = t_rad;
              strcpy (drvui->polyhedra[i].poly_col_edge, t_color);
          } 
      }
      break;
    case 37:					 /* frame command */
      /* if the current frame inherited a clip flag, default clip limit to pack range */
      if (clipflag==1 && drvui->frames[tmp_frame_no].clip_lim[0]== drvui->frames[tmp_frame_no].clip_lim[3]){
      for (i=0; i<6; i++)
          drvui->frames[tmp_frame_no].clip_lim[i] = drvui->frames[tmp_frame_no].cryst_lim[i];
      }
      tmp_frame_no++;
      break;
    case 38:					 /* axislines */
      memset(drvui->Ellipaxis_color,0,40);
      i = sscanf (input, "%*s %f %39c", &drvui->Ellipaxis_width,drvui->Ellipaxis_color);
      trim_string(drvui->Ellipaxis_color, 40);
      if (i < 3)
          strcpy (drvui->Ellipaxis_color, "Gray20");
      if (drvui->Ellipaxis_color[strlen (drvui->Ellipaxis_color) - 1] == '\n')
          drvui->Ellipaxis_color[strlen (drvui->Ellipaxis_color) - 1] = '\0';
      break;
    case 40:					 /* nolabels command */
      Labels = 0;				 /* set value false */
      break;
    case 41:					 /* lonepair command */
      get_label (input, &drvui->cones[drvui->ncone].cone_l1[0],
                 &drvui->cones[drvui->ncone].cone_l1[1],
                 &drvui->cones[drvui->ncone].cone_l1[2],
                 &drvui->cones[drvui->ncone].cone_l1[3], 1);
      (void) sscanf (input, "%d %f %f %f %39c",
		     &drvui->cones[drvui->ncone].numlonepairs,
                     &drvui->cones[drvui->ncone].cone_height,
		     &drvui->cones[drvui->ncone].cone_min,
                     &drvui->cones[drvui->ncone].cone_max, t_color);
      trim_string(t_color, 40);
      strcpy(drvui->cones[drvui->ncone].col_cone, t_color);
      if (strlen (drvui->cones[drvui->ncone].col_cone) == 0)
          strcpy (drvui->cones[drvui->ncone].col_cone, "Grey");
      drvui->cones[drvui->ncone].cone_fn = drvui->frame_no;
      drvui->ncone++;
      check_dynamic_storage();
      break;
    case 42:					 /* labeltext command */
      memset (drvui->labels[drvui->nlabel].label_label, 0, 64);
      Token_Strip (input, 1);
      (void) sscanf (input, "%f %f %f %63c",
		     &(drvui->labels[drvui->nlabel].label_x[0]),
		     &(drvui->labels[drvui->nlabel].label_x[1]),
                     &(drvui->labels[drvui->nlabel].label_x[2]),
		     drvui->labels[drvui->nlabel].label_label);
      trim_string(drvui->labels[drvui->nlabel].label_label, 64);
       drvui->labels[drvui->nlabel].label_fn = drvui->frame_no;
      drvui->nlabel++;
      check_dynamic_storage();
      break;
    case 44:					 /* background (color) command */
      Token_Strip (input, 1);
      trim_string(input, 256);
      strcpy (drvui->col_bg, input);
      break;
    case 45:
      (void)sscanf (input,"%*s %f %f %f %f %f %f %f %f %f %f %f %f %d",
          &drvui->slab_con[0],&drvui->slab_con[1],&drvui->slab_con[2],&drvui->slab_con[3],
          &drvui->slab_con[4],&drvui->slab_con[5],
          &drvui->slab_off[0],&drvui->slab_off[1],&drvui->slab_off[2],
          &drvui->slab_rot[0],&drvui->slab_rot[1],&drvui->slab_rot[2],&slabmode);

      break;
    case 46:					 /* magnetic command */
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %f",&drvui->mag_matrix[0][0],
                    &drvui->mag_matrix[0][1],&drvui->mag_matrix[0][2],
                    &drvui->mag_matrix[1][0],&drvui->mag_matrix[1][1],
                    &drvui->mag_matrix[1][2],&drvui->mag_matrix[2][0],
                    &drvui->mag_matrix[2][1],&drvui->mag_matrix[2][2]);
      break;
    case 47:
      (void) sscanf (input, "%*7c %d %d %d",
       	&omit1[nomits],&omit2[nomits],&omit3[nomits]);	
      nomits++;
      break;
    case 48:            /* noshadow */
      drvui->noshadow = 1;
      break;
    case 49:			/* finish command */
      (void) sscanf (input, "%*s %f %f %f %f", &drvui->ambient,
            &drvui->diffuse, &drvui->specular, &drvui->roughness);	
      break;
    case 50:            /* arrow command */
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %39c",&drvui->arrows[drvui->nmag].mag_xp[0],
                    &drvui->arrows[drvui->nmag].mag_xp[1],&drvui->arrows[drvui->nmag].mag_xp[2],
                    &drvui->arrows[drvui->nmag].mag_xc[0],&drvui->arrows[drvui->nmag].mag_xc[1],
                    &drvui->arrows[drvui->nmag].mag_xc[2],&drvui->arrows[drvui->nmag].arrow_length,
                    &drvui->arrows[drvui->nmag].arrow_diam,t_color);
      trim_string(t_color,40);
      strcpy(drvui->arrows[drvui->nmag].col_arrow,t_color);
      drvui->arrows[drvui->nmag].arrow_fn = tmp_frame_no;
      drvui->nmag++;
      break;
    case 52:            /* mapcontour command */
	{
	  char *amp;
      drvui->numOfFourierContours++;
      check_dynamic_storage();
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %s %69c",&drvui->fourier[drvui->numOfFourierContours].FourierContourLevel,
          string,t_color);
      trim_string (t_color, 70);
      if (!strlen (t_color))
        strcpy (t_color, "White");
      amp  = strchr(t_color,'&');
      if (amp) {
        *amp='\0';
        amp++;
        strcpy(drvui->fourier[drvui->numOfFourierContours].FourierBackColor,amp);
      }
      strcpy (drvui->fourier[drvui->numOfFourierContours].FourierContourColor, t_color);
      if(strncmp(string, "mesh", 4) == 0)
          drvui->fourier[drvui->numOfFourierContours].FourierContourSolid = 0;
      else
          drvui->fourier[drvui->numOfFourierContours].FourierContourSolid = 1;
      break;
	}
    case 53:              /* mapread command */
      Token_Strip (input, 1);
      strcpy(Map_Info.title," Title not given in map file");
      if (Quick) break;
      memset(filename,0,sizeof(filename));
      memset(t_name,0,sizeof(t_name));
      i = sscanf(input, "%*s %s %s %s", filename, string2, res);
      strcpy (FourierFileName, filename);
      if (i < 4)
        Map_Info.res = 4; /* No resolution specified - set 0.25 A */
      else
        sscanf(res, "%d", &Map_Info.res);
      if (Map_Info.res <= 0) Map_Info.res = 1;
      if (i < 3)
        strcpy(string2, "Fo"); /* Map type not specified - default to Fo */
      t_name[0] = '\0';
      if (strncmp(string, "grd", 3) == 0) {
          FourierMapType=1;
      } else if (strncmp(string, "stf", 3) == 0) {
          FourierMapType=2;
      } else if (strncmp(string, "w2k", 3) == 0) {
          FourierMapType=3;
      } else if (strncmp(string, "vsp", 4) == 0) {
          FourierMapType=4;
      } else if (strncmp(string, "flp", 4) == 0) {
          FourierMapType=5;
      } else if (strncmp(string, "fcf", 4) == 0) {
          FourierMapType=6;
          Map_Info.map_type = 0;
          if (!strcmp(string2,"Fc")) 
            Map_Info.map_type = 1;
          else if (!strcmp(string2,"Fo-Fc")) 
            Map_Info.map_type = 2;
          else if (!strcmp(string2,"2Fo-Fc")) 
            Map_Info.map_type = 3;
          else if (!strcmp(string2,"Fo2")) 
            Map_Info.map_type = 4;
      } else if (strncmp(string, "dn6", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=7;
      } else if (strncmp(string, "m80", 4) == 0) {
          FourierMapType=8;
          Map_Info.map_type = 0;
          if (!strcmp(string2,"Fc")) 
            Map_Info.map_type = 1;
          else if (!strcmp(string2,"Fo-Fc")) 
            Map_Info.map_type = 2;
          else if (!strcmp(string2,"2Fo-Fc")) 
            Map_Info.map_type = 3;
          else if (!strcmp(string2,"Fo2")) 
            Map_Info.map_type = 4;
      } else if (strncmp(string, "m81", 3) == 0) {
          FourierMapType=10;
          Map_Info.map_type = 0;
          if (!strcmp(string2,"Fc")) 
            Map_Info.map_type = 1;
          else if (!strcmp(string2,"Fo-Fc")) 
            Map_Info.map_type = 2;
          else if (!strcmp(string2,"2Fo-Fc")) 
            Map_Info.map_type = 3;
          else if (!strcmp(string2,"Fo2")) 
            Map_Info.map_type = 4;
      } else if (strncmp(string, "exc", 3) == 0) {
          FourierMapType=9;
      } else if (strncmp(string, "xsf", 3) == 0) {
          FourierMapType=11;    
      } else {
          fprintf(drvui->flout,"mapread command does not specify a legal file type.\n");
      }
      break;
    case 54:              /* mapregion command */
      Token_Strip(input, 1);
      (void) sscanf (input, "%f %f %f %f %f %f %f %f %f", &xMin, &xMax, &yMin,
                            &yMax, &zMin, &zMax, &x4Val, &x5Val, &x6Val);
      drvui->frames[tmp_frame_no].map_lim_set= 1;
      drvui->frames[tmp_frame_no].map_lim[0]=xMin;
      drvui->frames[tmp_frame_no].map_lim[3]=xMax;
      drvui->frames[tmp_frame_no].map_lim[1]=yMin;
      drvui->frames[tmp_frame_no].map_lim[4]=yMax;
      drvui->frames[tmp_frame_no].map_lim[2]=zMin;
      drvui->frames[tmp_frame_no].map_lim[5]=zMax;
      drvui->frames[tmp_frame_no].map_lim[6]=x4Val;
      drvui->frames[tmp_frame_no].map_lim[7]=x5Val;
      drvui->frames[tmp_frame_no].map_lim[8]=x6Val;
      i = 0;
      if (fabs(xMax - xMin) < 0.001f) i++;
      if (fabs(yMax - yMin) < 0.001f) i++;
      if (fabs(zMax - zMin) < 0.001f) i++;
      if (i == 1) drvui->Fourier2d = 1;
      break;
    case 55:              /* mapcalclimits command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f %f %f %f %f %f",&Map_Info.xlim[0], &Map_Info.xlim[1],
          &Map_Info.ylim[0], &Map_Info.ylim[1], &Map_Info.zlim[0], &Map_Info.zlim[1]);
      break;
    case 56:              /* labelscale command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f", &drvui->label_scale);
      break;
    case 57: /* bestplane command */
      {
        char bplane[15][5];
        char *p;
        sscanf(input,"%*s %d",&drvui->bplanes[drvui->nbplane].nbatoms);
        if (drvui->bplanes[drvui->nbplane].nbatoms > 15) {
            Error_Box("Cannot handle more than 15 atoms in a best plane.");
            return;
        }
        Token_Strip(input, 2);
        for (i=0;i<drvui->bplanes[drvui->nbplane].nbatoms;i++) {
         sscanf(input,"%s",bplane[i]);
          Token_Strip(input,1);
        }
        memset(drvui->bplanes[drvui->nbplane].bplane_col,0,40);
        sscanf(input,"%f %f %39c",&drvui->bplanes[drvui->nbplane].bplane_d1,
               &drvui->bplanes[drvui->nbplane].bplane_d2,t_color);
        trim_string(t_color,40);
        if (!strlen(t_color))strcpy(t_color,"White");
        strcpy(drvui->bplanes[drvui->nbplane].bplane_col,t_color);
        for (i=0;i<drvui->bplanes[drvui->nbplane].nbatoms;i++) {
          memset(drvui->bplanes[drvui->nbplane].bplane_t[i],0,4);
          strcpy(drvui->bplanes[drvui->nbplane].bplane_t[i],bplane[i]);
          p = strstr(bplane[i],"*");
          if (p!=NULL) {
            *p='\0';
            drvui->bplanes[drvui->nbplane].bplane_n[i]=-1;
          } else {
            for (j=1;j<4;j++)
              if (bplane[i][j]<= '9') {
                char *q=drvui->bplanes[drvui->nbplane].bplane_t[i]+j;
                drvui->bplanes[drvui->nbplane].bplane_n[i]=strtol(q,NULL,10);
                for (k=j;k<4;k++)
                  drvui->bplanes[drvui->nbplane].bplane_t[i][k]=' ';
                break;
              }
            }
          }
        drvui->nbplane++;
        check_dynamic_storage();
        break;
      }
    case 58:
        average=1;
      break;
    case 59:
      Token_Strip (input, 1);
      memset(modl,0,4);
      sscanf (input,"%s %d %f %f",modl,&modnum, &avg_occ, &min_occ);
      if (modl[1]=='\0')
        modl[1]=modl[2]=modl[3]=' ';
      if (modl[2]=='\0')
        modl[2]=modl[3]=' ';
      if (modl[3]==0)
        modl[3]=' ';
      modl[4]='\0';
          for (j=0;j< natom;j++) {
            if (check_atom_name(modl,drvui->atoms[j].atom_l) && modnum == drvui->atoms[j].atom_n ) {
              drvui->atoms[j].occupancy=avg_occ;
              drvui->atoms[j].min_occ=min_occ;
              break;
            }
          }
      break;
    case 60:
      sscanf(input,"%*s %f %f %f",&drvui->phaseshift[0], &drvui->phaseshift[1],
                                      &drvui->phaseshift[2]);
      for (j=0; j<3; j++)
          drvui->phaseshift[j] = (float)fmod (drvui->phaseshift[j], 1.);
      break;
 
    case 61:            /* mapcontour2d command */
      drvui->numOfFourierContours++;
      check_dynamic_storage();
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %39c",&drvui->fourier[drvui->numOfFourierContours].FourierContourLevel,
                    &drvui->fourier[drvui->numOfFourierContours].FourierContourStep,
                    &drvui->fourier[drvui->numOfFourierContours].FourierContourTop,
                    t_color);
      trim_string(t_color,40);
      if (!strlen(t_color))strcpy(t_color,"White");
      strcpy(drvui->fourier[drvui->numOfFourierContours].FourierContourColor,t_color);
      break;
    case 62:
      memset(modl,0,4);
      sscanf(input,"%*s %s %d %s %s %39c", modl,&modnum,filename,string,t_color);
      if (modl[1]=='\0')
        modl[1]=modl[2]=modl[3]=' ';
      if (modl[2]=='\0')
        modl[2]=modl[3]=' ';
      if (modl[3]==0)
        modl[3]=' ';
      modl[4]='\0';
      if (drvui->nsurf==MAX_SURF) {
        if (!Quick)
          fprintf(drvui->flout, "Maximum number of surfaces exceeded, skipping\n");
      break;
      }
      strcpy(drvui->surfatom[drvui->nsurf],modl);
      drvui->surfnum[drvui->nsurf]=modnum;
      if (!strncmp(string,"mesh",4)){
        drvui->surftype[drvui->nsurf]=0;
      } else if (!strncmp(string,"solid",5)) {
        drvui->surftype[drvui->nsurf]=1;
      } else 
        drvui->surftype[drvui->nsurf]=2;
      trim_string(t_color,40);
      if (!strlen(t_color))strcpy(t_color,"White");
      strcpy(drvui->surfcolor[drvui->nsurf],t_color);
      read_aim(filename,Quick);
      drvui->nsurf++;
      break;

    case 63: /* voids command */
      sscanf(input,"%*s %d %f %d %d %d %39c", &j, &drvui->probesize, 
                    &drvui->voidgrid[0],&drvui->voidgrid[1],&drvui->voidgrid[2],t_color);
      trim_string(t_color,40);
      if (!strlen(t_color))strcpy(t_color,"White");
      strcpy(drvui->voidcolor,t_color);
      if (drvui->voidflag>=0) { 
        drvui->voidflag=j;
        if (j==1) {
          drvui->voidmap=(char***)zalloc(drvui->voidgrid[0]*sizeof(*drvui->voidmap));
          for (j=0;j<drvui->voidgrid[0];j++) {
            drvui->voidmap[j]=(char**)zalloc(drvui->voidgrid[1]*sizeof(*drvui->voidmap[j]));
            for (k=0;k<drvui->voidgrid[1];k++)
              drvui->voidmap[j][k]=(char*)zalloc(drvui->voidgrid[2]*sizeof(*drvui->voidmap[j][k]));
          }
        } 
      }
      break;
    case 64:                     /* values command */
      get_label (input, &drvui->atprops[drvui->natprop].atprop_l[0],
		       &drvui->atprops[drvui->natprop].atprop_l[1],
		       &drvui->atprops[drvui->natprop].atprop_l[2],
		       &drvui->atprops[drvui->natprop].atprop_l[3], 1);

      i = sscanf (input, "%s %s", string, string2);
      if (atoi (string) == 0)
        drvui->atprops[drvui->natprop].atprop_n = -1;
      else
        drvui->atprops[drvui->natprop].atprop_n = atoi(string);
      drvui->atprops[drvui->natprop].radius = (float)atof (string2);
      drvui->atprops[drvui->natprop].atprop_fn = tmp_frame_no;
      drvui->natprop++;
      check_dynamic_storage ();
      break;
    case 65: /* mapslice command */
      Token_Strip (input,1);
      j = sscanf (input,"%f %f %f %f %f %f %d",&drvui->frames[tmp_frame_no].mapslice[0],
                 &drvui->frames[tmp_frame_no].mapslice[1],&drvui->frames[tmp_frame_no].mapslice[2],
                 &drvui->frames[tmp_frame_no].mapnorm[0],&drvui->frames[tmp_frame_no].mapnorm[1],
                 &drvui->frames[tmp_frame_no].mapnorm[2],&drvui->frames[tmp_frame_no].slice);
      if (j <7) drvui->frames[tmp_frame_no].slice = 1;
      break;
    case 66: /* qvector command */
      Token_Strip (input,1);
      j = drvui->no_cell_vec;
      sscanf (input,"%f %f %f",  &drvui->cell_vec[j][0],
              &drvui->cell_vec[j][1], &drvui->cell_vec[j][2]);
      drvui->no_cell_vec++;
      drvui->modulated++;
      break;
    case 67:
      ShowMapLegend = 1;
      break;
    case 68:
      Vrml2 = 1;
      X3D = 1;
      break;
                                                                                
    default:
      if (!Quick) {
          fprintf (drvui->flout,
		     "Next line does not start with a recognizable keyword - ignored\n");
          fprintf (drvui->flout, "*         %s\n", input1);	 /* echo to listing file */
      }
      break;
    }
    if (!Quick && intype)
      fprintf (drvui->flout, "*         %s\n", input1);	 /* echo to listing file */
  }
  if (drvui->n_ellips > 1)
      do_ellipsoids = 1;
  if (packflag != 0) {
      boxlim[0] = boxlim[1] = boxlim[2] = 100.0f;
  }
}						 /* end of read_inp */

/* ************************************************************** */
/* ************************************************************** */

int vec_dif(int n1, float v1[3], int n2, float v2[3], int n3, float v3[3], float v[3])
{
/* routine to determine if vector v == n1*v1+n2*v2+n3*v3, returns 1 if equal, 0 if not */
    int i;
    float d[3];

    for (i=0; i<3; i++)
        d[i] = v[i] - n1 * v1[i] - n2 * v2[i] - n3 * v3[i];
    return d[0] * d[0] + d[1] * d[1] + d[2] * d[2] < 1.0e-4 ? 1 : 0;
}


