/* $Id: DRAWxtl1.c 1115 2011-02-23 20:33:07Z martin $ */

/* ===================================================================== */
/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 1                                                                */
/*                                                                       */
/* Coded by:  L.W. Finger                                                */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            Larry.Finger@lwfinger.net                                  */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/* This program is copyrighted by Larry W. Finger, Martin Kroeker and Brian Toby
 * and is distributed under the GNU General Public License - see the
 * accompanying COPYING file for more details.
 */

/* Routines included in this file:
 *
 *   main - entry point
 *   add_vert - add a vertex to the list with checking
 *   add_vert_nc - add a vertex without checking
 *   add_to_list - add an atom to the unit-cell list
 *   analyze_bonds - print the bond table
 *   axeqb - solve the matrix equation AX = B
 *   Blank_Strip - strip leading blanks
 *   build_box_contents - get all the atoms in the display box
 *   Calc_Rot - calculate the rotation matrix from lookat vectors
 *   Convert_ellipsoids - change all ellipsoids to betaij notation
 *   Conv_Sym_Mat - change hexagonal and trigonal symmetry rotations to cartesian
 *   delay - wait for a time
 *   dist - calculate a bond distance
 *   do_drawing - calls all the individual generation routines
 *   expand_atoms - uses symmetry to expand an atom in the unit cell
 *   eigen - calculate eigen values and vectors
 *   find_all_in_box - get all the atoms in the display box
 *   generate_bonds - draw the appropriate bonds
 *   generate_cones - find and draw the lone-pair cones
 *   generate_arrows - draw the arrows
 *   get_atom_id - encode drawing info into atom number
 *   get_input - preliminary routine for reading input file
 *   get_label - get an atom label
 *   label_cell - convert cell axis labels to labeltext commands
 *   Locate_Triple - find the "best" location for the vector triple
 *   make_bmat - form the unit-cell matrices
 *   modulate_parameters - perform modulation on occupancies and positions
 *   modulate_uij - perform modulation on thermal parameters
 *   Output_Spheres - generate the spheres
 *   P_to_C - convert ellipsoid probability to appropriate constant
 *   plot_vrml_poly - output VRML polyhedra
 *   print_sym - format the symmetry matrices
 *   Token_Strip - remove an item from the beginning of the input line
 *   usage - print out the minimal help information
 *   resizeWindow - GLUT callback when openGL window is resized
 *   display_crystal - GLUT callback for openGL display
 *   Dump_View - prepare EPS file
 *   Dump_Gif - prepare GIF file
 *   key - GLUT callback to interpret keyboard commands
 *   specialkey - GLUT callback to interpret various control keys
 *   mouse - GLUT callback for mouse
 *   mousemotion - GLUT callback function for motion
 *   start_picking - used in dynamic edit of screen contents
 *   Process_hits - process the omit list
 *   pick_label - used in dynamic positioning of text labels
 *   polygon_normal_3d - calculates normal vector of a polygon
 *   polygon_solid_angle_3d - calculates projected solid angle of a 3d plane polygon
 *   not_in_slab - return info regarding whether an atom is inside the slab
 *   dot0_3d - perform a dot product
 *   tb_init - initialize trackball
 *   trackball - implements the virtual trackball
 *   window_to_tb - converts from window coordinates to trackball
 *   tb_project_to_sphere - does the projection from trackball to sphere
 *   axis_to_quaternion - changes a single axial rotation to quaternion form
 *   qmult - multiplies two quaternions
 *   qnormalize - normalizes a quaternion
 *   quaternion_to_rotmatrix - generates the equivalent rotation matrix
 *   vnew - creates a new vector and sets its components
 *   vset - loads the components into an existing vector
 *   vcopy - copies a vector
 *   vzero - zeros a vector
 *   vnormalize - normalizes a vector to unit length
 *   vlength - returns the length of a vector
 *   vscale - multiplies a vector by a scalar
 *   vmult - multiplies two vectors
 *   vadd - adds two vectors
 *   vsub - subtracts a vector from another
 *   vdot - takes the dot product of two vectors
 *   vcross - calculates the cross product
 *   XYZ_Rot_to_Q - forms a quaternion from rotation matrices about X, Y, and Z
 *   print_map_info - prints the information about a Fourier map
 *   Progress_Window - simulates a progress meter on stderr
 *   zalloc - combines malloc and memset
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_gbl.h"
#include "gl2ps.h"
#ifndef WIN32
  #include <unistd.h>
#else
  #define cbrt(val) pow((val), 1.0 / 3.0)
#endif

#ifdef OPENGL
#ifdef WIN32
  #define AMBLEV 0.2f
#else
  #define AMBLEV 0.2f
#endif
  static void display_crystal (void);
  static void key (unsigned char, int, int);
  static void mouse (int, int, int, int);
  static void mousemotion (int, int);
  static void process_hits(GLint, GLuint*);
  static void resizeWindow (int, int);
  static void specialkey (int, int, int);
  static void start_picking(int,int);
  static int pick_label(int,int,int,int,float*);
  static void close_and_exit(void);
  void Dump_Gif(void);
#endif

/* ************************************************************** */
/* ************************************************************** */
int main (int argc, char *argv[])
{
  int i, j, Quick = 0;
  char temp[200];

#ifdef DELAY
  time_t start, end;
#endif
  float t_start,t_end,t_step;
  int nstep,no_steps;
  char string[200],string2[10];
#ifdef OPENGL
  GLfloat mat_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat light_ambient[] =	{ AMBLEV, AMBLEV, AMBLEV, 1.0f };
  GLfloat light_specular[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat mat_shininess[] = { 0.8f };
#endif

#ifdef DELAY
  start = time (NULL);
#endif

/* Initialize */

  drvui = zalloc(sizeof(CRYSTAL));
  init_dynamic_storage();
  drvui->X_Origin = drvui->Y_Origin = drvui->Z_Origin = 0.5f;
  drvui->X_Boxlim = drvui->Y_Boxlim = drvui->Z_Boxlim = 100.0f;
  memset(drvui->slab_con,0,sizeof(drvui->slab_con));
  offset[0] = offset[1] = offset[2] = 0.0f;
  drvui->glback[0] = drvui->glback[1] = drvui->glback[2] = 1.0f;
  xrot = yrot = zrot = 0.0f;
  boxflag = packflag = domolcomp = 0;
  do_ellipsoids = edges = 0;
  drvui->Ellipsoid_Prob = 0.5f;
  drvui->label_scale = 1.0f;
  drvui->labels_moved = 0;
  drvui->autolabel = 0;
  drvui->automation = 0;
  drvui->acentric = 1;
  Unit_Cell = M_cameras = no_comment = Labels = Vrml2 = 1;
  X3D = 0;
  gl_view = 1;
  Display_axes = El_Cutout = 0;
  strcpy (drvui->Cutout_color, "Gray20");
  strcpy (drvui->Ellipaxis_color, "Gray20");
  SpMult = drvui->BndMult = Magnification = 1.0f;
  strcpy (drvui->col_cell, "Black");	/* preset unit-cell colors to Black */
  drvui->Ellipsoid_Scale = P_to_C (0.5f);
  Options = 0;
  drvui->polylimit = 0.1f;
  drvui->Phong_Value = 0.2f;
  drvui->Phong_Size = 1.0f;
  DepthCue = 0.0f;
  printdist = 3.5f;
  drvui->numOfFourierContours = 0;
  drvui->Fourier2d = 0;
  drvui->no_subsys = 1;
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++)
      drvui->subsys_fact[0][i][j] = 0.0f;
    drvui->subsys_fact[0][i][i] = 1.0f; /* set to identity */
  }
  memset(omit1,0,1000);
  memset(omit2,0,1000);
  memset(omit3,0,1000);
/* Parse arguments */

  if (argc < 2) {
    usage ();
    delay (3);
  }
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-n"))
      Unit_Cell = 0;
    else if (!strcmp (argv[i], "-h"))
      usage ();
    else if (!strcmp (argv[i], "-?"))
      usage ();
    else if (!strcmp (argv[i], "-o")) {
      drvui->X_Origin = (float) atof (argv[++i]);
      drvui->Y_Origin = (float) atof (argv[++i]);
      drvui->Z_Origin = (float) atof (argv[++i]);
      Options = Options | O_OPT;
    }
    else if (!strcmp (argv[i], "-O")) {
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
    }
    else if (!strcmp (argv[i], "-C")) {
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
    }
    else if (!strcmp (argv[i], "-c")) {
      no_comment = 0;
    }
    else if (!strcmp (argv[i], "-b")) {
      drvui->X_Boxlim = (float) atof (argv[++i]);
      drvui->Y_Boxlim = (float) atof (argv[++i]);
      drvui->Z_Boxlim = (float) atof (argv[++i]);
      boxflag = 1;
      Options = Options | B_OPT;
    } else if (!strcmp (argv[i], "-p")) {
      drvui->frames[1].cryst_lim[0] = (float) atof (argv[++i]);
      drvui->frames[1].cryst_lim[3] = (float) atof (argv[++i]);
      drvui->frames[1].cryst_lim[1] = (float) atof (argv[++i]);
      drvui->frames[1].cryst_lim[4] = (float) atof (argv[++i]);
      drvui->frames[1].cryst_lim[2] = (float) atof (argv[++i]);
      drvui->frames[1].cryst_lim[5] = (float) atof (argv[++i]);
      Options = Options | P_OPT;
      packflag = 1;
    } else if (!strcmp (argv[i], "-v")) {
      xrot = (float) atof (argv[++i]);
      yrot = (float) atof (argv[++i]);
      zrot = (float) atof (argv[++i]);
      Options = Options | V_OPT;
    } else if (!strcmp (argv[i], "-M")) {
      strncpy(drvui->MSMS_Path,argv[++i],1023);
      drvui->MSMS_Path[1023]='\0';
    } else if (!strcmp (argv[i], "-m")) {
      Magnification = (float) atof (argv[++i]);
      Options = Options | M_OPT;
    } else if (!strcmp (argv[i], "-u")) {
      Display_axes = 1;
    } else if (!strcmp (argv[i], "-Q")) {
      Quick = 1;
    } else if (!strcmp (argv[i], "-H")) {
      view_inst = 0;
    } else if (!strcmp (argv[i], "-G")) {
      gl_view = 0;
    } else if (!strcmp (argv[i], "-a")) {
        drvui->autolabel = 2;
    } else if (!strcmp (argv[i], "-S")) {
        drvui->Stereo = 1;
        sscanf(argv[++i],"%d",&drvui->cross_eyed);
        sscanf(argv[++i],"%f",&drvui->stereo_base);
    } else if (!strcmp (argv[i], "-s")) {
        drvui->Stereo = 2;
        sscanf(argv[++i],"%d",&drvui->cross_eyed);
        sscanf(argv[++i],"%f",&drvui->stereo_base);
    } else if (!strcmp (argv[i], "-A")) {
        drvui->automation = 1;
        sscanf(argv[++i],"%f",&t_start);
        sscanf(argv[++i],"%f",&t_end);
        sscanf(argv[++i],"%f",&t_step);
    } else
      drvui->Str_File = argv[i];
  }
 
/* next section will change working directory to that of the data file */
  (void)getcwd(drvui->savedpath,100);           /* save current working directory */
  strcpy(temp,drvui->Str_File);
  for (i=strlen(temp); i>0; --i) {
    if(temp[i] == '/' || temp[i] == '\\') {
        temp[i] = '\0';
        break;\
    }
  }
  if (i > 0) chdir(temp);
  (void)getcwd(drvui->savedpath,100);           /* save current working directory */
  for (i=strlen(drvui->Str_File); i>0; --i) {
    if (drvui->Str_File[i] == '/' || drvui->Str_File[i] == '\\') {
      i++;
      break;
    }
  }
  if (i > 0) {
    strcpy(temp,drvui->Str_File +i);
    strcpy(drvui->Str_File,temp);
  }
  if (!(drvui->fpin = fopen(drvui->Str_File,"r"))) {
      printf("Input file '%s' cannot be opened. Run aborted!\n",drvui->Str_File);
      exit(0);
  }
  fclose(drvui->fpin);

  if (strstr(drvui->Str_File,".cif")
  && strlen(strstr(drvui->Str_File,".cif"))==4  ) { /* filename ends in .cif */
    FILE *newstr;
    strcpy(temp,drvui->Str_File);     /* copy original cif filename */
    strcat(temp,".str");              /* and add extension */
    newstr=fopen(temp,"w");
    if (!newstr) {
       printf("Implied str file '%s' could not be created. Run aborted!\n",temp);
       exit(0);
    }   
    fprintf(newstr,"titl imported\n");
    fprintf(newstr,"import cif %s 1\n",drvui->Str_File);
    fprintf(newstr,"box 0.02 Black\n");
    fprintf(newstr,"background White\n");
    fprintf(newstr,"view 0. 0. 0.\n");
    fprintf(newstr,"pack  -0.05 1.05 -0.05 1.05 -0.05 1.05\n");
    fprintf(newstr,"end\n");
    fclose(newstr);
    drvui->auto_ellipse=1;
    strcpy(drvui->Str_File,temp);           
  }
  if (packflag != 0) {
    drvui->X_Boxlim = drvui->Y_Boxlim = drvui->Z_Boxlim = 500.0f;
  }
  if ((drvui->atom_no = zalloc ((unsigned) ((long) (MAX_VERTS * sizeof (int))))) == NULL) {
    printf ("Unable to get initial atom_no allocation\n");
    return 0;
  }
  if ((drvui->atom_so = zalloc ((unsigned) ((long) (MAX_VERTS * sizeof (int))))) == NULL) {
    printf ("Unable to get initial atom_so allocation\n");
    return 0;
  }
  if ((xypos = zalloc ((unsigned) (3 * (long) (MAX_VERTS * sizeof (float))))) == NULL) {
    printf ("Unable to get initial xypos allocation\n");
    return 0;
  }
  if ((xypos_nm = zalloc ((unsigned) (3 * (long) (MAX_VERTS * sizeof (float))))) == NULL) {
    printf ("Unable to get initial xypos_nm allocation\n");
    return 0;
  }
#ifdef OPENGL
  if (gl_view) {
    glutInit (&argc, argv);
    glutInitDisplayMode ((GLbitfield) (GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB));
    glutInitWindowPosition (100, 50);
    glutInitWindowSize (gl_width, gl_height);
    (void) glutCreateWindow ("DRAWxtl V4.5");
    glutDisplayFunc (display_crystal);
    glutReshapeFunc (resizeWindow);

/*************************************************************************************
 *
 *   The glutWMCloseFunc call below makes closing the openGL window by clicking on the X
 *   in the upper right corner the same as pressing the <esc>, 'q', or 'Q' keys. This
 *   function is not part of the standard GLUT API and may not compile on all systems.
 *   It is OK to comment it out. Without it, the structure orientation at time of exit
 *   is not written to the POV file, nor is the appropriate view statement appended to 
 *   the 'str' file, otherwise the program operates normally.
 *
 *   LWF 08/08/05
 **************************************************************************************/

#if (defined (WIN32) && !defined(CYGWIN)) || defined(FREEGLUT) || (defined(GLUT_MACOSX_IMPLEMENTATION) && GLUT_MACOSX_IMPLEMENTATION>1)
/*    glutWMCloseFunc (close_and_exit); */
#endif
    glutKeyboardFunc (key);
    glutMouseFunc (mouse);
    glutMotionFunc (mousemotion);
    glutSpecialFunc (specialkey);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMaterialfv (GL_FRONT, GL_SPECULAR, mat_specular);

    glMaterialfv (GL_FRONT, GL_SHININESS, mat_shininess);

    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    glEnable (GL_COLOR_MATERIAL);

    glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable (GL_DEPTH_TEST);
  }
#endif
  if (drvui->automation) {
       char *q;
       drvui->flout=fopen("automation.log","w");
    no_steps = (int)(((t_end - t_start) / t_step) + 0.5f) + 1;
    for (nstep = 1; nstep < no_steps; nstep++) {
    drvui->phaseshift[0] = t_start + (float)nstep * t_step;
    drvui->automation=nstep;
       strcpy(string, drvui->Str_File);
       q=strrchr(string,'.');
       if (q) *q='\0';
      sprintf(string2, "%.3d.pov", nstep);
       strcat(string, string2);
       drvui->fpoutp=fopen(string,"w");
       drvui->fpoutv=fopen("automation.wrl","w");
       fprintf (drvui->flout,"Generating file %d of %d, t=%f\n",nstep,no_steps-1,drvui->phaseshift[0]);
       Quick=1;
       crystal_compute(Quick,0);
       fclose(drvui->fpoutp);
       fclose(drvui->fpoutv);
       unlink("automation.wrl");
    }
   if (FourierPt) free(FourierPt);
   return 1;
  } else  
  crystal_compute (Quick,0);			 /* generate POV and VRML outputs */
#ifdef DELAY
  end = time (NULL);
  printf ("Elapsed time: %10.1f secs.\n", difftime (end, start));
  start = time (NULL);
  delay (2);
#endif
#ifdef OPENGL
  if (gl_view) {
    glutMainLoop ();
  }
#endif
  if (FourierPt) free(FourierPt);
  return 1;
}						 /* end main for POV and VRML versions */

/* ************************************************************** */
/* ************************************************************** */

void add_vert (float vert[3], int no, int check, int modulate, int sign)

/* procedure to add the vertex in 'vert' to the current list if the
   vertex is contained within the limits of the box, within the limits
   of a polyhedral center, or needed to complete a molecule. If modulate
   is true (!= 0), the coordinates in vert will be adjusted for modulation
   before testing */
{
  float c_vert[3];               /* vertex in cartesian coords */
  float r_vert[3];               /* vertex in rotated coordinates */
  float vert_saved[3];           /* save the unmodulated vetex */
  int i, j;                      /* loop variables */
  float temp;                    /* temporary storage */
  int outside;                   /* temporary for molecular completion */
  int number;                    /* decoded atom number */
  int poly_center;               /* true if this atom is center of a polyhedron */
  int consider_extra;
  float d = 0.0f;
  double occupancy;
  int sym_no;

  number = no / 1000;
  sym_no = no - 1000*number;
  for (i=0; i<3; i++)
      vert_saved[i] = vert[i];
  if (modulate & (drvui->modulated > 0)) {
      modulate_parameters(vert, &occupancy, sym_no, number);
      if (occupancy < 0.01) return;
  }
  for (i = 0; i < 3; ++i) {      /* convert vertex coordinates to Cartesian */
    c_vert[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      c_vert[i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  if (number >= 0 && number < drvui->atom_alloc)
      poly_center = (((drvui->atoms[number].atom_n >> 8) & 255) > 0);
  else
      poly_center = 0;
  consider_extra = (domolcomp != 0 || (drvui->npoly > 1 && !poly_center));
  outside = 0;
  if (check == 0) {
      for (i = 0; i <= 2; i++) {  /* check packing range */
        if ((vert[i] < drvui->frames[drvui->frame_no].cryst_lim[i]) ||
            (vert[i] > drvui->frames[drvui->frame_no].cryst_lim[i + 3]))
          outside = 1;
      }
      if (outside == 0 && slabmode == 1)
        outside = not_in_slab(c_vert[0], c_vert[1],c_vert[2]);

    if (((outside == 0) && (fabs (c_vert[0]) > boxlim[0])) ||
         (fabs (c_vert[1]) > boxlim[1]) || (fabs (c_vert[2]) > boxlim[2]))
        outside = 1;                     /* check box limits */
      if (outside != 0 && !consider_extra)
        return;
  }
  d = 0.0f;
  if (outside == 1) {            /* molecule or polyhedral completion is in progress */
    float dmin, dmax;
    int polyno;                          /* number of the polyhedron */

    for (i = 1; i < NvertM; i++) {
      if (drvui->atoms[drvui->atom_no[i]/1000].atom_fn != drvui->frame_no) continue;

      dmax = drvui->mol_d * drvui->mol_d;
      dmin = 0.;
      d = 0.0f;
      for (j = 0; j < 3; j++) {
        temp = (s_vert[3 * i + j] - c_vert[j]);
        d += temp * temp;
      }
      j = drvui->atom_no[i] / 1000;
      poly_center = (((drvui->atoms[j].atom_n >> 8) & 255) > 0);
      if (poly_center) {
        polyno = (drvui->atoms[j].atom_n >> 8 ) & 255;
        dmax = drvui->polyhedra[polyno].poly_size * drvui->polyhedra[polyno].poly_size;
        dmin = drvui->polyhedra[polyno].poly_min * drvui->polyhedra[polyno].poly_min;
      }
      if ((domolcomp != 0 || poly_center) && (d <= dmax && d > 0.02)) {
        outside = 2;
        if (poly_center && d < dmin)
          outside = 1;
        if ((fabs (c_vert[0]) > 10. * boxlim[0]) || (fabs (c_vert[1]) > 10. * boxlim[1])
                 || (fabs (c_vert[2]) > 10. * boxlim[2])) {
          fprintf (stderr, "mol'completing beyond 10*lattice constant - aborting run\n");
          outside = 1;                   /* far out of check box limits - combinatorial explosion */
          break;
        }
      }
    }
  }
  if (outside == 1)
    return;
  i = (int) (no / 1000);
  for (j = 1; j < NvertM; ++j) {
    if (fabs (s_vert[3 * j] - c_vert[0]) < 0.01 && fabs (s_vert[3 * j + 1] - c_vert[1]) < 0.01 &&
        fabs (s_vert[3 * j + 2] - c_vert[2]) < 0.01 &&
        check_atom_name(drvui->atoms[i].atom_l,drvui->atoms[drvui->atom_no[j]/1000].atom_l)  && (number == i))
      return;                            /* eliminate duplicates */
  }
  for (i = 0; i <= 2; ++i) {             /* add atom to Master Lists */
    xypos[3 * NvertM + i] = vert[i];
    xypos_nm[3 * NvertM + i] = vert_saved[i];
    s_vert[3 * NvertM + i] = c_vert[i];
/*    if (offset[i] > c_vert[i])
        offset[i] = c_vert[i];
*/        
  }
  i = no - 1000*(no/1000);

/*if (occupancy != 1.) fprintf (stderr,"occ=%f\n",occupancy);*/
  drvui->vert_occ[NvertM] = (float)occupancy;
  drvui->atom_so[NvertM] = sign;
  drvui->atom_no[NvertM++] = no;
  if (NvertM > MAX_VERTS) {
    if (domolcomp != 0) {
      Error_Box ("Too many vertices for dimensions:\n"
          " Did you use a 'molcomp' command incorrectly?\n"
          " If not, increase MAX_VERTS.");
    } else {
      Error_Box("Too many vertices for dimensions - increase MAX_VERTS.");
    }
    return;
  }
  for (i = 0; i <= 2; ++i) {             /* calculate position of point after POV rotation */
    r_vert[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      r_vert[i] += (float) G_Rot[j][i] * c_vert[j];
  }
  if (drvui->automation>1) return;
  for (i = 0; i <= 2; ++i) {             /* Update Min and Max of output */
    if (POV_Max[i] < r_vert[i])
      POV_Max[i] = r_vert[i];
    if (POV_Min[i] > r_vert[i])
      POV_Min[i] = r_vert[i];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void add_vert_nc (float vert[3])

/* procedure to add the vertices in 'vert' to the current working list
   (NOT Master List ) without checking */
{

  int i, j;					 /* loop variables */

  if (!s_vert) {
      if ((s_vert = zalloc ((unsigned) (3 * (long) 
          (MAX_VERTS * sizeof (float))))) ==	NULL) {
        printf ("Unable to get initial s_vert allocation\n");
        delay (5);
        return;
      }
  }  
  for (i = 0; i <= 2; ++i) {			 /* convert vertex coordinates to Cartesian */
    s_vert[3 * nvert + i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      s_vert[3 * nvert + i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  nvert++;
}

/* ************************************************************** */
/* ************************************************************** */

void add_to_list (float xp[3], int no, int nos)

/* add atom position to unit-cell list if unique */
/* no - symmetry operator number */
{
  int i, ind;

  for (i = 0; i <= 2; ++i) {			 /* get coordinates between 0 and 1 */
    if (xp[i] >= 1.0)
      xp[i] -= 1.0f;
    if (xp[i] < 0.0)
      xp[i] += 1.0f;
  }
  ind = 0;
  for (i = 0; i < ncell; ++i) {
    if ((fabs (drvui->cell_xyz[i][0] - xp[0]) <= 0.00002) &&
	(fabs (drvui->cell_xyz[i][1] - xp[1]) <= 0.00002) &&
	(fabs (drvui->cell_xyz[i][2] - xp[2]) <= 0.00002))
      ind = 1;					 /*set if same */
  }
  if (ind == 0) {				 /* true for new position */
    for (i = 0; i <= 2; ++i)			 /*copy to list */
      drvui->cell_xyz[ncell][i] = xp[i];
    drvui->sym_op_signed[ncell] = nos;
    drvui->sym_op_no[ncell++] = no;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void analyze_bonds (void)

/* routine to print bond distances up to a maximum of 'printdist' Angstrom */
{
  int i, j, k, l, nvert1, *itype;
  float *d, dmin;
  double occ;

  if ((itype = zalloc ((unsigned) (MAX_VERTS * sizeof (int)))) == NULL)
    return;
  fprintf (drvui->flout, "\n\n Bond Distance Analysis\n");
/* Find all atoms of this type in master list */
  l = nvert = nvert1 = 1;
  for (j = 0; j < natom; ++j) {
    if (drvui->atoms[j].atom_fn != drvui->frame_no) continue;
    l = nvert;
    find_all_in_box (j);
    for (k = l; k < nvert; ++k)
      itype[k] = j;                      /* itype[k] has atom number j */
  }
  for (i = 0; i < natom; ++i) {		 /* loop through atom types */
    if (drvui->atoms[i].atom_fn != drvui->frame_no) continue;
/* add the base entry to the beginning of the atom list  - picked from vertex list
   because position originating from x,y,z symmetry operator may be absent due to
   crenel exclusion */
    k = -1;
    for (l=1; l<nvert; l++) {    /* find an entry for atom i */
      if (itype[l] == i) {
        for (j = 0; j < 3; ++j) {
          o_vert[j] = o_vert[3*l+j];
          o_vert_nm[j] = o_vert_nm[3*l+j];
          s_vert[j] = s_vert[3*l+j];
        }
        k = l;
        break;
      }
    }
    if (k < 0) {
/* There is no entry for atom i in the list due to limits, etc. Add the base one. */
      for (j = 0; j < 3; j++) {
        s_vert[j] = 0;
        o_vert[j] = drvui->atoms[i].atom_xyz[j];
        o_vert_nm[j] = o_vert[j];
        for (l = 0; l < 3; l++)
          s_vert[j] += (float) drvui->b_mat[l][j] * (drvui->atoms[i].atom_xyz[l] - origin[l]);
      }
    }
    if ((d = zalloc ((unsigned) (nvert * sizeof (float)))) == NULL) {
      free (itype);
      printf ("Unable to get bond distance allocation\n");
      return;					 /* get space for distances */
    }
    for (j = 1; j < nvert; ++j)			 /* calculate distances */
      d[j] = dist (0, j);
    fprintf (drvui->flout, "\nDistances from %c%c%c%c%3d at %8.5f%8.5f%8.5f    to\n\n",
             drvui->atoms[i].atom_l[0], drvui->atoms[i].atom_l[1], drvui->atoms[i].atom_l[2],
             drvui->atoms[i].atom_l[3], drvui->atoms[i].sv_atom_n, o_vert[0], o_vert[1], o_vert[2]);
    for (k = nvert1; k < nvert; ++k) {
      dmin = 999999.9f;
      for (j = nvert1; j < nvert; ++j) {     /* loop through distances */
        if (d[j] < dmin) {
          l = j;
          dmin = d[j];
        }
      }
      if (dmin == 0.0 && itype[l] == i) {
        d[l] = 999999.9f;
      } else {
        if (dmin <= printdist) {
          fprintf (drvui->flout," %c%c%c%c%3d at %9.5f%9.5f%9.5f  %8.3f",
                   drvui->atoms[itype[l]].atom_l[0], drvui->atoms[itype[l]].atom_l[1],
                   drvui->atoms[itype[l]].atom_l[2], drvui->atoms[itype[l]].atom_l[3],
                   drvui->atoms[itype[l]].sv_atom_n, o_vert[3 * l], o_vert[3 * l + 1],
                   o_vert[3 * l + 2], dmin);
          d[l] = 999999.9f;
          if (drvui->atoms[i].atom_ismod != 0 || drvui->atoms[itype[l]].atom_ismod != 0 ) { /* one or both ends are modulated */
              float temp_pa[3];
              float vert1[3], vert2[3], vert3[3];
              float dsum = 0.0f, dmax = 0.0f, thisd;
              int m, mx[3], m1, m2, number = 0;
              dmin = 999999.0f;
              for (m=0; m<3; m++) {
                  temp_pa[m] = drvui->phaseshift[m];
                  mx[m] = 50;
                  if (drvui->modulated <= m) mx[m] = 1;
              }
              for (m=0; m<mx[0]; m++) {   /* sum over values of initial phaseshift for first modulation*/
                  drvui->phaseshift[0] = m * 0.02f;
                  for (m1=0; m1<mx[1]; m1++) {
                      drvui->phaseshift[1] = m1 * 0.02f; /* second modulation */
                      for (m2=0; m2<mx[2]; m2++) {
                          drvui->phaseshift[2] = m2 * 0.02f;   /* third modulation */
                          for (j=0; j<3; j++) {
                              vert1[j] = o_vert_nm[j];
                              vert2[j] = o_vert_nm[3 * l + j];
                          }
                          modulate_parameters(vert1, &occ, 0, i);
                          modulate_parameters(vert2, &occ, vert_sym_no[l], itype[l]);
                          for (j = 0; j < 3; ++j) {   /* convert vertex coordinates to Cartesian */
                              vert3[j] = 0.0f;
                              for (k = 0; k <= 2; ++k)
                                  vert3[j] += (float) drvui->b_mat[j][k] * (vert2[k] - vert1[k]);
                          }
                          thisd = (float)sqrt(vert3[0]*vert3[0] + vert3[1]*vert3[1] + vert3[2]*vert3[2]);
                          dsum += thisd;
                          dmin = min(thisd, dmin);
                          dmax = max(thisd, dmax);
                          number++;
                      }
                  }
              }
              dsum /= (float)number;          /* get mean distance */
              for (m=0; m<3; m++)
                  drvui->phaseshift[m] = temp_pa[m];
              fprintf(drvui->flout,"%8.3f %8.3f %8.3f", dsum, dmin, dmax);
          }
          fprintf(drvui->flout,"\n");
        } else
          break;
      }
    }
    free (d);
  }
  free (itype);
  fprintf (drvui->flout, "\n\n");
}

/* ************************************************************** */
/* ************************************************************** */

void axeqb (double a1[3][3], double x[3], double b[3])

/* solve matrix equation ax = b for x, where a is 3x3 and x and b are
   vectors */
{
  int i, j;
  double a[3][3], det;

  det = determinant (a1);
  if (fabs (det) < 1.0e-6)
    det = 1.0e-6;
  a[0][0] = (a1[1][1] * a1[2][2] - a1[1][2] * a1[2][1]) / det;
  a[1][0] = -(a1[1][0] * a1[2][2] - a1[1][2] * a1[2][0]) / det;
  a[2][0] = (a1[1][0] * a1[2][1] - a1[1][1] * a1[2][0]) / det;
  a[0][1] = -(a1[0][1] * a1[2][2] - a1[0][2] * a1[2][1]) / det;
  a[1][1] = (a1[0][0] * a1[2][2] - a1[0][2] * a1[2][0]) / det;
  a[2][1] = -(a1[0][0] * a1[2][1] - a1[0][1] * a1[2][0]) / det;
  a[0][2] = (a1[0][1] * a1[1][2] - a1[0][2] * a1[1][1]) / det;
  a[1][2] = -(a1[0][0] * a1[1][2] - a1[0][2] * a1[1][0]) / det;
  a[2][2] = (a1[0][0] * a1[1][1] - a1[0][1] * a1[1][0]) / det;
  for (i = 0; i < 3; i++) {
    x[i] = 0.0;
    for (j = 0; j < 3; j++)
      x[i] += a[i][j] * b[j];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void Blank_Strip (char input[])

/* routine to suppress any leading spaces in the input string */
{
  int i, j;

  for (i = 0; i < (int)strlen(input); i++)
    if (input[i] != ' ')
      break;
  for (j = 0; j < (int)strlen(input); j++) {
    input[j] = input[i];
    if (input[i++] == '\0')
      break;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void build_box_contents (void)

/* routine to locate all atoms in the display parallelopiped plus
   any atoms needed to complete molecules or polyhedra */
{
  float vert[3];				 /* storage for center */
  int ix[3];					 /* array for lattice offset */
  int j, k, saved_nvert;
  int atoms, loop;

  loop = 0;
  while (1) {                        /* infinite loop here */
    saved_nvert = NvertM;      /* save number currently in list */
    for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
      for (j=0;j<3;j++) {            /* put the base atom in the range 0-1 */
        drvui->atoms[atoms].saved_xyz[j] = drvui->atoms[atoms].atom_xyz[j];
        if (drvui->atoms[atoms].atom_xyz[j] < 0.0f) drvui->atoms[atoms].atom_xyz[j] += 1.0f;
        if (drvui->atoms[atoms].atom_xyz[j] > 1.0f) drvui->atoms[atoms].atom_xyz[j] -= 1.0f;
      }
      expand_atom (atoms);          /* generate all atoms of this flavor in unit cell */
      for (j = 0; j < ncell; ++j) { /*atoms in cell */
        for (k = 0; k <= 2; ++k)
          vert[k] = drvui->cell_xyz[j][k];
        add_vert (vert, 1000 * atoms + drvui->sym_op_no[j], 0, 1, drvui->sym_op_signed[j]);	/* add point to cell if inside box */
      }
    }
    if (saved_nvert == NvertM) {
      break;
    }
  }
  while (1) {                        /* infinite loop here */
    saved_nvert = NvertM;			 /* save number currently in list */
    for (atoms = 0; atoms < natom; atoms++) {	 /* loop through atoms */
      expand_atom (atoms);			 /* generate all atoms of this flavor in unit cell */
      for (j=0;j<ncell;++j) {        /*atoms in cell*/
        for (k=0;k<=2;++k) 
          vert[k] = drvui->cell_xyz[j][k];
        add_vert(vert, 1000*atoms+drvui->sym_op_no[j], 0, 1, drvui->sym_op_signed[j]);  /* add point to cell if inside box*/
      }
      for (ix[0] = (int) drvui->frames[drvui->frame_no].cryst_lim[0] - 1; 
               ix[0] <= (int) drvui->frames[drvui->frame_no].cryst_lim[3] + 1; ++ix[0]) {/* step through cells parallel a */
        for (ix[1] = (int) drvui->frames[drvui->frame_no].cryst_lim[1] - 1; 
               ix[1] <= (int) drvui->frames[drvui->frame_no].cryst_lim[4] + 1; ++ix[1]) {	/* parallel b */
          for (ix[2] = (int) drvui->frames[drvui->frame_no].cryst_lim[2] - 1; 
               ix[2] <= (int) drvui->frames[drvui->frame_no].cryst_lim[5] + 1; ++ix[2]) {	/* parallel c */
            if (abs (ix[0]) + abs (ix[1]) + abs (ix[2]) != 0) {
              for (j = 0; j < ncell; ++j) {	 /*atoms in cell */
                for (k = 0; k <= 2; ++k)
                  vert[k] = ix[k] + drvui->cell_xyz[j][k];
                add_vert (vert, 1000 * atoms + drvui->sym_op_no[j], 0, 1, drvui->sym_op_signed[j]);	/* add point to cell if inside box */
              }
            }
          }
        }
      }
    }
    if ((domolcomp == 0 && drvui->npoly == 1) || saved_nvert == NvertM) {
      for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
          for (j=0;j<3;j++) {            /* restore base atom position */
              drvui->atoms[atoms].atom_xyz[j] = drvui->atoms[atoms].saved_xyz[j];
          }
      }
      return;
    }
    if (loop == 0)
      printf ("\n");
    loop++;
    printf ("Recursive Atom Addition, Loop No. %3d, %4d Atoms in List.\n",
	    loop, NvertM - 1);
  }
}						 /* end of build_box_contents */

/* ************************************************************** */
/* ************************************************************** */

void Calc_Rot (float v1[3], float v2[3])

/* routine to calculate rotation axes needed to put v1 parallel to Z
   and the projection of v2 parallel to Y */
{
  float u1[3], u2[3], u3[3], temp;
  double U[3][3];
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      U[i][j] = 0.0f;			 /* clear U */
  temp = 0.0f;
  for (i = 0; i < 3; i++) {			 /* convert v1 to Cartesian */
    u1[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u1[i] += (float) drvui->b_mat[i][j] * v1[j];
    temp += u1[i] * u1[i];
  }
  if (temp <= 0.0) {
    u1[2] = 1.0f;			 /* if null input, make u1 = 0,0,1 */
    U[2][2] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {			 /* make a unit vector */
      u1[i] /= temp;
      U[i][2] = u1[i];				 /* put column in U */
    }
  }
  for (i = 0; i < 3; i++) {			 /* convert v2 to Cartesian */
    u2[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u2[i] += (float) drvui->b_mat[i][j] * v2[j];
  }
  u3[0] = u2[1] * u1[2] - u1[1] * u2[2];	 /* calculate u2 X u1 */
  u3[1] = u2[2] * u1[0] - u2[0] * u1[2];
  u3[2] = u2[0] * u1[1] - u1[0] * u2[1];
  temp = 0.0f;
  for (i = 0; i < 3; i++)
    temp += u3[i] * u3[i];
  if (temp <= 0.0) {
    u3[0] = 1.0f;			 /* if null input, make u3 = 1,0,0 */
    U[0][0] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {			 /* make a unit vector */
      u3[i] /= temp;
      U[i][0] = u3[i];				 /* add column to U */
    }
  }
  u2[0] = u1[1] * u3[2] - u3[1] * u1[2];	 /* calculate u2 = u1 X u3 */
  u2[1] = u1[2] * u3[0] - u1[0] * u3[2];
  u2[2] = u1[0] * u3[1] - u3[0] * u1[1];
  for (i = 0; i < 3; i++)
    U[i][1] = u2[i];
/* make determinant of U > 0 */
  if (determinant (U) < 0.0) {
    for (i = 0; i < 3; i++)
      U[i][0] = -U[i][0];
  }

/* The rotation required to get the desired orientation, which is described
    by G_Rot, has been set up so G_Rot * U = I. Thus G_Rot = U(inv),
    U is unitary and U(inv) = U(transpose). */

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      G_Rot[i][j] = (float) U[j][i];
/* calculate rotations */
  temp = (float) (G_Rot[1][2] * G_Rot[1][2] + G_Rot[2][2] * G_Rot[2][2]);
  if (temp > 0.001) {
    float temp2;

    yrot = (float) (atan (G_Rot[0][2] / sqrt (temp)) * RAD);
    temp2 = (float) (G_Rot[1][2] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    xrot = -(float) asin (temp2) * (float) RAD;
    temp2 = -(float) (G_Rot[0][1] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    zrot = (float) asin (temp2) * (float) RAD;
  } else {
    yrot = -(float) asin (G_Rot[0][2]) * (float) RAD;
    xrot = 0.0f;				 /* XXXXXXXXXXX  - May not be right */
    zrot = 0.0f;				 /* XXXXXXXXXXX  - May not be right */
  }
}

void Convert_Cart_Cryst(float bmat_inv[3][3], float inp[3], float oup[3], float origin[3])
{
	int i, j;

	for (i = 0; i < 3; i++) {
		oup[i] = origin[i];
		for (j = 0; j < 3; j++)
			oup[i] += bmat_inv[i][j] * inp[j];
	}
}

void Convert_Cryst_Cart(double b_mat[3][3], float inp[3], float oup[3], float origin[3])
{
	int i, j;

	for (i = 0; i < 3; i++) {
		oup[i] = 0.0f;
		for (j = 0; j < 3; j++)
			oup[i] += (float)b_mat[i][j] * (inp[j] - origin[j]);
	}
}
/* ************************************************************** */
/* ************************************************************** */


void convert_ellipsoid (void)

/* routine to convert ellipsoid coefficients Bij or Uij to betaij form */
{
  int i, j, save;

  for (i=1; i<drvui->n_ellips; ++i) {
    if (drvui->ellips[i].ellips_ismod != 0 )
      continue;
    save = drvui->ellips[i].ell_type/1000;
    if (drvui->ellips[i].ell_type % 1000 != 0) {     /* multiply Uij and Bij by a*(i)*a*(j)/4 */
      for (j = 0; j < 3; ++j)
        drvui->ellips[i].ellips[j] *= drvui->rec_lat_con[j] * drvui->rec_lat_con[j] * 0.25f;
      drvui->ellips[i].ellips[3] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[1] * 0.25f;
      drvui->ellips[i].ellips[4] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[2] * 0.25f;
      drvui->ellips[i].ellips[5] *= drvui->rec_lat_con[1] * drvui->rec_lat_con[2] * 0.25f;
      if (drvui->ellips[i].ell_type % 1000 == 1)        /* multiply Uij by 8pi^2 */
        for (j = 0; j < 6; ++j)
          drvui->ellips[i].ellips[j] *= 78.9568f;
    }
    drvui->ellips[i].ell_type= 1000 * save;
  }
}						 /* end of convert_ellipsoid */

/* ************************************************************** */
/* ************************************************************** */

void Conv_Sym_Mat (void)

/* routine to convert the rotational part of a symmetry matrix to a Cartesian rotation */
{
  int i, j, k, l;
  double mat[3][3];
  float B[3][3], Binv[3][3];
  float tmp[3][3];

  if (drvui->sys == 5) {     /* hexagonal systems are special */
      for (i = 0; i < 3; i++)
         for (j = 0; j < 3; j++) {
            B[i][j] = (float)drvui->b_mat[i][j];
            Binv[i][j] = B[i][j];
         }
      matinv(Binv);
  }

  for (l = 0; l < drvui->ng; l++) {
    for (j = 0; j < 3; j++) {
      for (k = 0; k < 3; k++) {
        mat[j][k] = drvui->rss[l][j][k];
      }
    }
    if (determinant (mat) < 0.0) {		 /* if improper rotation, negate all elements */
      for (j = 0; j < 3; j++) {
        for (k = 0; k < 3; k++) {
          mat[j][k] *= -1.0f;
        }
      }
    }
    for (j = 0; j < 3; j++)
      for (k = 0; k < 3; k++)
        drvui->rssC[l][j][k] = (float)mat[j][k];  /* save a copy of the original */

    if (drvui->sys == 5) {     /* hexagonal systems are special */
        float R[3][3];
        for (j = 0; j < 3; j++)
            for (k = 0; k < 3; k++)
                R[j][k] = (float)mat[j][k];
        matmul(B, R, tmp);
        matmul(tmp, Binv, R);              /* The Cartesian matrix is B R Binv */
        for (j = 0; j < 3; j++)
            for (k = 0; k < 3; k++)
                mat[j][k] = R[j][k];
    }
     
    for (j = 0; j < 3; j++) {			 /* restore matrix */
      for (k = 0; k < 3; k++) {
        drvui->rss[l][j][k] = (float) mat[j][k];
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void delay (int i)

/* delays for 'i' seconds if DELAY is defined, otherwise it is a noop */
{
#ifdef DELAY
  time_t start, end;

  end = time (NULL);
  start = time (NULL);
  while (difftime (start, end) < (float) i)
    start = time (NULL);
#endif
}

/* ************************************************************** */
/* ************************************************************** */

float dist (int j, int k)

/* calculates distance between vertices at s_vert[j] and s_vert[k] */
{
  float d, t;
  int i;

  d = 0.0f;
  for (i = 0; i <= 2; ++i) {
    t = s_vert[3 * j + i] - s_vert[3 * k + i];
    d += t * t;
  }
  return ((float) sqrt (d));
}

/* ************************************************************** */
/* ************************************************************** */

void do_drawing (void)

/* routine to put together the elements of the drawing */
{
    int i;

    label_cell();
    if (drvui->frame_no == 1) 
        draw_cell (docell);	         /* generate unit-cell corners and draw it */
    nvert = 0;                       /* empty vertex list */
    generate_spheres ();             /* generate all spheres to be drawn within box */
    if (do_ellipsoids && drvui->n_ellips > 1 && drvui->frame_no == 1) {
        generate_ellipsoids ();		 /* now the ellipsoids */
    }
    if (drvui->nmag >0)
        generate_arrows();
    generate_bonds ();				 /* generate all bonds to be drawn within box */
    generate_cones ();				 /* generate all lonepairs to be drawn within box */
    generate_poly ();                /* generate all polyhedra to be drawn */
    generate_planes ();				 /* now the planes */
    generate_texts ();               /* any textual elements */
    generate_lsq_planes ();          /* any least squares planes */
    generate_aimsurf ();             /* any Bader surfaces */
    
    calculate_voids();
    generate_voids();		     /* any cavities or molecular surfaces */
    
    if (slabmode > 1 && drvui->frame_no == 1)
        show_slab();                 /* show cutout section */
    if (ReadFourMap) {
      if (drvui->frames[drvui->frame_no].slice > 1) {
                       generate_slice();
      } else if (!drvui->Fourier2d) { /* 3d map */
          for (i = 1; i <= drvui->numOfFourierContours; ++i) {
/* generate & display a Fourier map */
              generate_map (drvui->fourier[i].FourierContourLevel, drvui->fourier[i].FourierContourSolid,
                            drvui->fourier[i].FourierContourColor,drvui->fourier[i].FourierBackColor);
          }
      } else {                /* generate 2d map contours */
          for (i = 1; i <= drvui->numOfFourierContours; ++i) {
              float level = drvui->fourier[i].FourierContourLevel;
              for (;;) {
                  generate_map (level, 0, drvui->fourier[i].FourierContourColor,"");
                  level += drvui->fourier[i].FourierContourStep;
                  if (level > drvui->fourier[i].FourierContourTop) break;
              }
          }
      }
    }
}


/* ************************************************************** */
/* ************************************************************** */

void expand_atom (int natom)

/* routine to expand an atom position and obtain all equipoints within the unit cell */
/* natom - number of atom in list to expand */
{
  float xp[3], xpp[3];           /* user in generating position */
  int i, j, k, l;                /* loop variables */

  ncell = 0;                     /* clear list */
  for (i = 0; i < drvui->ng; ++i) {     /* symmetry operations */
    for (j = 0; j <= 2; ++j) {
      xp[j] = drvui->ts[i][j];
      for (k = 0; k <= 2; ++k)
        xp[j] = xp[j] + drvui->ss[i][j][k] * drvui->atoms[natom].atom_xyz[k];
    }
    for (k = 0; k < drvui->nlat; ++k) { /*lattice points */
      for (l = 0; l <= 2; ++l)
        xpp[l] = xp[l] + drvui->lat_pos[k][l];
      add_to_list (xpp, i, i);
      if (!drvui->acentric) {
        for (l = 0; l <= 2; ++l)
          xpp[l] = -xpp[l];
        if (i==0) 
          add_to_list (xpp, i, -1000);
        else
          add_to_list (xpp, i, -i);  
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

int
eigen (float *biso, float beta[6], float valu[3], float vect[3][3])

/* routine to calculate eigenvalues 'valu' and vectors 'v' for
 * array of anisotropic thermal parameters 'beta', 'biso' is the equivalent
 * isotropic B
 *
 * some parts of this routine were copied from ORTEP
 */
{
    static double errnd = 1.0E-9;

    double b0, b1, b2;

    double a1, a2, phi;

    double a[3][3], b[3][3], w[3][3], u[3];

    float v1[3], v2[3], v3[3];

    double tem, sigma, smax;

    double Vect[3][3];

    int i, j, l, ii, iii, i1, imax = 0;

    int neq;

    static double tpi2 = 2.0 * M_PI * M_PI;	/* 2 pi^2 */

/* put beta into square symmetric matrix b (called M in ORTEP) */

    for (i = 0; i < 3; i++)
	b[i][i] = beta[i];
    b[0][1] = b[1][0] = beta[3];
    b[0][2] = b[2][0] = beta[4];
    b[1][2] = b[2][1] = beta[5];
/* multiply b * ginv to get w */
    for (i = 0; i < 3; i++) {
	for (j = 0; j < 3; j++) {
	    w[i][j] = 0.0;
	    for (l = 0; l < 3; l++) {
		w[i][j] += b[l][i] * drvui->ginv[j][l];
	    }
	}
    }
    sigma = 0.0;
    for (i = 0; i < 3; i++) {
	for (j = 0; j < 3; j++) {
	    a[i][j] = w[i][j];
	    sigma += a[i][j] * a[i][j];
	}
    }
    if (sigma <= 0.0)
	return (0);		/* error on null 'w' */
    sigma = sqrt (sigma);
/* get coefficients of third-order characteristic equation 
 *        Equation is -L^3 + b2 L^2 + b1 L + b0 = 0
 */
    b2 = a[0][0] + a[1][1] + a[2][2];
    b1 = -a[0][0] * a[1][1] - a[0][0] * a[2][2] - a[1][1] * a[2][2]
	+ a[0][2] * a[2][0] + a[0][1] * a[1][0] + a[1][2] * a[2][1];
    b0 = a[0][0] * a[1][1] * a[2][2] + a[1][0] * a[2][1] * a[0][2]
	+ a[2][0] * a[1][2] * a[0][1] - a[2][0] * a[0][2] * a[1][1]
	- a[0][0] * a[2][1] * a[1][2] - a[1][0] * a[0][1] * a[2][2];

    *biso = (float) (4.0 * b2 / 3.0);
/* transform to get to form x^3 + a1 x + a2  == 0 */
    a1 = (-3.0 * b1 - b2 * b2) / 3.0;
    a2 = (-2.0 * b2 * b2 * b2 - 9.0 * b2 * b1 - 27.0 * b0) / 27.0;
    tem = (a2 * a2) / 4.0 + (a1 * a1 * a1) / 27.0;
    if (tem > errnd) {
	/* two complex roots */
	fprintf (drvui->flout, "Complex Roots! tem = %f\n", tem);
	return (0);		/* error if complex roots */
    } else if (tem < -errnd) {
	/* 3 real and unequal roots */
	phi = acos (-(a2 / 2.0) / sqrt (-(a1 * a1 * a1) / 27.0));
	u[0] = 2.0 * sqrt (-a1 / 3.0) * cos (phi / 3.0) + b2 / 3.0;
	u[1] = 2.0 * sqrt (-a1 / 3.0) * cos (phi / 3.0 + 120.0 / RAD) + b2 / 3.0;
	u[2] = 2.0 * sqrt (-a1 / 3.0) * cos (phi / 3.0 + 240.0 / RAD) + b2 / 3.0;
    } else {
	double sign_a2 = 1.0;
	/* Workaround brain-dead cbrt routine in Windows */
	if (a2 < 0.0) {
	    sign_a2 = -1.0;
	    a2 = fabs(a2);
	}
	/* 3 real roots with at least 2 being equal */
	u[0] = -2.0 * sign_a2 * cbrt (a2 / 2.0) + b2 / 3.0;
	u[1] = sign_a2 * cbrt (a2 / 2.0) + b2 / 3.0;
	u[2] = u[1];
    }
    for (j = 0; j < 2; j++) {	/* sort roots in increasing order */
	if (u[j] > u[j + 1]) {
	    tem = u[j + 1];
	    u[j + 1] = u[j];
	    u[j] = tem;
	}
    }
    for (i = 0; i < 3; i++)
	if (valu[i] < 0.0)
	    valu[i] = 0.001f;
	else
	    valu[i] = (float) sqrt (u[i] / tpi2);

/* count multiple roots */
    neq = 1;
    for (j = 0; j < 2; j++)
	if (fabs (u[j] - u[j + 1]) < errnd)
	    neq++;
    if (neq > 2) {
/* 3 equal roots */
	for (ii = 0; ii < 3; ii++) {
	    for (i = 0; i < 3; i++)
		vect[i][ii] = 0.0;
	    vect[ii][ii] = 1.0;
	}
	return (1);
    } else if (neq == 2) {
	for (ii = 0; ii < 3; ii++) {
	    v1[ii] = v2[ii] = v3[ii] = 0.0;
	}
	if (drvui->sys == 5) {	/* hexagonal */ 
/* With 2 equal eigenvalues, we have degenerate vectors */
	    if (fabs (u[0] - u[1]) < errnd) {
		v3[2] = 1.0f;	/* 3rd vector is unique */
		v1[0] = 1.0f;
	    } else {
		v3[0] = 1.0f;	/* 1st one is */
		v1[2] = 1.0f;
	    }
	} else if (drvui->sys == 6) {		/* cubic with atom at xxx */
	    if (fabs (u[0] - u[1]) < errnd) {
		v3[0] = v3[1] = v3[2] = (float) sqrt (3.0) / 3.0f;
		v1[0] = -(float) sqrt (2.0) / 2.0f;
		v1[1] = -v1[0];
	    } else {
		v1[0] = v1[1] = v1[2] = (float) sqrt (3.0) / 3.0f;
		v3[0] = -(float) sqrt (2.0) / 2.0f;
		v3[1] = -v3[0];
	    }
	} else {		/* tetragonal */
	    if (fabs (u[0] - u[1]) < errnd) {
		v3[2] = 1.0f;	/* 3rd vector is unique */
		v2[1] = 1.0f;
		v1[0] = 1.0f;
	    } else {
		v3[0] = 1.0f;	/* 1st one is */
		v1[1] = 1.0f;
		v2[2] = 1.0f;
	    }
	}
	vcross (v3, v1, v2);	/* v1, v2 and v3 are mutually perp. unit vectors */
	vnormalize (v2);
	for (j = 0; j < 3; j++) {
	    vect[0][j] = v1[j];
	    vect[1][j] = v2[j];
	    vect[2][j] = v3[j];
	}
	return (1);
    }
    for (ii = 0; ii < 3; ii++) {
	for (iii = 0; iii < 3; iii++) {
	    for (i = 0; i < 3; i++)
		a[i][iii] = w[i][iii];
	    a[iii][iii] = w[iii][iii] - u[ii];
	}
	smax = 0.0;
	for (i = 0; i < 3; i++) {
	    i1 = (i < 2) ? i + 1 : 0;
	    b[0][i] = a[1][i] * a[2][i1] - a[2][i] * a[1][i1];
	    b[1][i] = a[2][i] * a[0][i1] - a[0][i] * a[2][i1];
	    b[2][i] = a[0][i] * a[1][i1] - a[1][i] * a[0][i1];
	    tem = b[0][i] * b[0][i] + b[1][i] * b[1][i] + b[2][i] * b[2][i];
	    if (tem > smax) {
		smax = tem;
		imax = i;
	    }
	}
	if (smax <= errnd)
	    return (0);
	smax = (float) sqrt (smax);
	for (i = 0; i < 3; i++)
	    v2[i] = (float) (b[i][imax] / smax);
	for (i = 0; i < 3; i++) {	/* Convert v2 to cartesian system */
	    v1[i] = 0.0;
	    for (j = 0; j < 3; j++)
		v1[i] += (float) drvui->b_mat[i][j] * v2[j];
	}
	vnormalize (v1);
	for (i = 0; i < 3; i++)
	    Vect[ii][i] = v1[i];	/* unit vector */
    }
    if (determinant (Vect) < 0.97) {
	for (i = 0; i < 3; i++) {
	    v2[i] = (float) Vect[1][i];
	    v3[i] = (float) Vect[2][i];
	}
	vcross (v2, v3, v1);	/* Set is not orthogonal -the smallest root is probably wrong */
	for (i = 0; i < 3; i++)
	    Vect[0][i] = v1[i];
    }
    for (i = 0; i < 3; i++)
	for (j = 0; j < 3; j++)
	    vect[i][j] = (float) Vect[i][j];
    return (1);
}

/* ************************************************************** */
/* ************************************************************** */

void find_all_in_box (int i)

/* routine to locate all atoms of type 'i' in the display box */
/* i - atom number */
{
  int n, j, k, l;                        /* loop indices */
  int sym_no;                            /* symmetry number */

  for (n = 1; n < NvertM; n++) {         /* loop through the Master list */
    j = (int) drvui->atom_no[n] / 1000;  /* extract atom number */
    if (j == i) {
      drvui->orig_atom_no[nvert] = i;
      sym_no = drvui->atom_no[n] - 1000 * j; /* symmetry operator number */
      for (k = 0; k < 3; k++) {
          o_vert[3 * nvert + k] = xypos[3 * n + k];
          o_vert_nm[3 * nvert + k] = xypos_nm[3 * n + k];
      }
      for (k = 0; k < 3; k++) {
          s_vert[3 * nvert + k] = 0.0f;
          for (l = 0; l < 3; l++) {         /* calculate cartesian coordinates */
            s_vert[3 * nvert + k] +=
               (float) drvui->b_mat[k][l] * (o_vert[3 * nvert + l] - origin[l]);
          }
      }
      vert_sym_nos[nvert] = drvui->atom_so[n];
      vert_sym_no[nvert++] = sym_no;
      if (nvert > MAX_VERTS) {
          Error_Box("Too many vertices for dimensions - increase MAX_VERTS.");
          return;
      }
      if (nvert > 2 * NvertM) {
#ifdef WIN32
          Error_Box("Overrun of vert_sym_no. Please send 'str' file\n"
                   "to Larry.Finger@@lwfinger.net.");
          exit(0);
#else
          vert_sym_no = (int *)realloc(vert_sym_no,  (nvert + 2) * sizeof(int));
          drvui->orig_atom_no = (int *)realloc(drvui->orig_atom_no, (nvert+2) * sizeof(int));
          o_vert = (float *)realloc (o_vert, (6 * (nvert+2) * sizeof (float)));
          o_vert_nm = (float *)realloc (o_vert_nm, (6 * (nvert+2) * sizeof (float)));
#endif
      }
    }
  }
}      /* end of find_all_in_box */

/* ************************************************************** */
/* ************************************************************** */
void generate_bonds (void)

/* routine to generate bond descriptions and add then to edit list */
{
  int start, same_type;
  int i, j, k;        			 /* loop counters */
  int N_Bond;        			 /* Counter for number of bonds of a type */
  int l1;        			 /* loop counter for matrix operations */
  int nvert1;        			 /* vertex counter */
  int Bond_Count;        		 /* Counter for bonds output */
  char bnd1[4], bnd2[4];        	 /* temporary storage of bond labels */
  float d;        			 /* bond distance */
  float df[3];        			 /* vector difference */
  float beta;        			 /* rotation angle */
  float gamma;        			 /* another rotation angle */
  float at1[3], at2[3];        		 /*initial coordinates of bond endpoints */
  float clip0, clip1;        		 /* clipping scale factors */
  int m, l2, l3, out = 0;
  int ell[2] = {0,0};  			 /* save ellipsoid numbers */

  int is_ellipsoid[2] = {0,0}; 		 /* ellipsoid markers - for moving bond endpoint */
  int o1, p, p1, q, r;        		 /* loop variables for rotation matrix calculation */
  float elrot0[3][3];        		 /* rotation matrices of ellipsoids */
  float Z0[3], Z1[3];        		 /* components of rms displacement along bond */
  int *ellips_id;
  float factor, rms0, rms1;
  float glr, glg, glb;
#ifdef OPENGL
  GLUquadricObj *glu_quadObj = NULL;
#endif
  int omit;
  char col_bond_p[40];
  char col_bond_v[40];
  int numdashes;
  float dashes = 5.0f;
  float mf;
  
  if (drvui->nbond == 1)
    return;        			 /* exit if no work to be done */

  Bond_Count = 0;
  if (!(ellips_id = zalloc ((unsigned) ((long) (MAX_VERTS * sizeof (int)))))) {
    fprintf(drvui->flout,"Unable to get ellips_id allocation!");
    exit(0);
  }

#ifdef OPENGL
              if (gl_view) 
                glu_quadObj = gluNewQuadric ();
#endif

  for (i = 1; i < drvui->nbond; ++i) {        	 /* loop through bond types */
    if (drvui->bonds[i].bond_fn != drvui->frame_no) continue;  /* skip if not in this frame */
    strcpy(col_bond_p,drvui->bonds[i].col_bond);
    strcpy(col_bond_v,drvui->bonds[i].col_bond);
    Transform_POV_Color(col_bond_p);
    Transform_VRML_Color(col_bond_v);
    nvert = 0;        			         /* empty the vertex list */
    for (j = 0; j < natom; j++) {        /* loop through atom types */
      int o = drvui->atoms[j].atom_n;
      if (((o & 255) > 0) | (((o >> 24) & 255) > 0)) { /* sphere or ellipsoid */

        o = (drvui->atoms[j].atom_n >> 24) & 255;
        if (o > 0) {                      /* ellipsoid */
          if ((drvui->bonds[i].bond_l1[0] == drvui->ellips[o].ellips_l[0]) &&
              (drvui->bonds[i].bond_l1[1] == drvui->ellips[o].ellips_l[1]) &&
              (drvui->bonds[i].bond_l1[2] == drvui->ellips[o].ellips_l[2]) &&
              (drvui->bonds[i].bond_l1[3] == drvui->ellips[o].ellips_l[3])) {
            if (do_ellipsoids == 1 && El_Cutout == 1 && drvui->ellips[o].ell_type>100) {    /* skip if B iso */
              is_ellipsoid[0] = 1;         /* or no cutout */
              ell[0] = o;
            }
            o = nvert;
            find_all_in_box (j);          /* get all atoms of type j */
            for (p = o; p < nvert; p++)
              ellips_id[p] = ell[0];      /* save ellipsoid id for these entries */
          }
        } else {                          /* sphere */
          o = drvui->atoms[j].atom_n & 255;     /* number of sphere */
          is_ellipsoid[0] = 0;
          if ((drvui->bonds[i].bond_l1[0] == drvui->spheres[o].sphere_l[0]) &&
              (drvui->bonds[i].bond_l1[1] == drvui->spheres[o].sphere_l[1]) &&
              (drvui->bonds[i].bond_l1[2] == drvui->spheres[o].sphere_l[2]) &&
              (drvui->bonds[i].bond_l1[3] == drvui->spheres[o].sphere_l[3])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 /* o > 0 */
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->bonds[i].bond_l1[0];
    bnd1[1] = drvui->bonds[i].bond_l1[1];
    bnd1[2] = drvui->bonds[i].bond_l1[2];
    bnd1[3] = drvui->bonds[i].bond_l1[3];
    nvert1 = nvert;
    if ((drvui->bonds[i].bond_l1[0] == drvui->bonds[i].bond_l2[0]) &&
        (drvui->bonds[i].bond_l1[1] == drvui->bonds[i].bond_l2[1]) &&
        (drvui->bonds[i].bond_l1[2] == drvui->bonds[i].bond_l2[2]) &&
        (drvui->bonds[i].bond_l1[3] == drvui->bonds[i].bond_l2[3])) {
      bnd2[0] = drvui->bonds[i].bond_l1[0];        	 /* atoms are same type */
      bnd2[1] = drvui->bonds[i].bond_l1[1];
      bnd2[2] = drvui->bonds[i].bond_l1[2];
      bnd2[3] = drvui->bonds[i].bond_l1[3];
      is_ellipsoid[1] = is_ellipsoid[0];
      ell[1] = ell[0];
      same_type = 1;
    } else {
      bnd2[0] = drvui->bonds[i].bond_l2[0];
      bnd2[1] = drvui->bonds[i].bond_l2[1];
      bnd2[2] = drvui->bonds[i].bond_l2[2];
      bnd2[3] = drvui->bonds[i].bond_l2[3];
      same_type = 0;
/* atoms i and j are not the same type */
      for (j = 0; j < natom; ++j) {         /* loop through atom types */
        if (((drvui->atoms[j].atom_n & 255) > 0) | (((drvui->atoms[j].atom_n >> 24) & 255) > 0)) { /* sphere or ellipsoid */
          int o;

          o = (drvui->atoms[j].atom_n >> 24) & 255;
          if (o > 0) {                      /* ellipsoid */
            if ((drvui->bonds[i].bond_l2[0] == drvui->ellips[o].ellips_l[0]) &&
                (drvui->bonds[i].bond_l2[1] == drvui->ellips[o].ellips_l[1]) &&
                (drvui->bonds[i].bond_l2[2] == drvui->ellips[o].ellips_l[2]) &&
                (drvui->bonds[i].bond_l2[3] == drvui->ellips[o].ellips_l[3])) {
              if (do_ellipsoids == 1 && El_Cutout == 1 && drvui->ellips[o].ell_type > 100) {    /* skip if B iso */
                is_ellipsoid[1] = 1;         /* or no cutout */
                ell[1] = o;
              }
              o = nvert;
              find_all_in_box (j);         /* get all atoms of type j */
              for (p = o; p < nvert; p++)
                ellips_id[p] = ell[1];         /* save ellipsoid type */
            }
          } else {
            o = drvui->atoms[j].atom_n & 255;               /* sphere number */
            if ((drvui->bonds[i].bond_l2[0] == drvui->spheres[o].sphere_l[0]) &&
                (drvui->bonds[i].bond_l2[1] == drvui->spheres[o].sphere_l[1]) &&
                (drvui->bonds[i].bond_l2[2] == drvui->spheres[o].sphere_l[2]) &&
                (drvui->bonds[i].bond_l2[3] == drvui->spheres[o].sphere_l[3])) {
              find_all_in_box (j);         /* get all atoms of type j */
              is_ellipsoid[1] = 0;
            }
          }        			 /* o > 0 */
        }
      }        				 /* for (j=0;j<natom;++j) */
    }        				 /* if atoms are same type */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        start = nvert1;
        if (same_type == 1)
          start = j;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if ((d >= drvui->bonds[i].bond_min) && (d <= drvui->bonds[i].bond_max)) {    /* here if distance within range */
            df[0] = df[1] = drvui->Bond_Mult * drvui->bonds[i].bond_size;    /* scaling */
            df[2] = 0.5f *d;

            Bond_Count++;        	 /* update count */
            for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
              Z0[p] = s_vert[3 * j + p];
              Z1[p] = s_vert[3 * k + p];
            }

            if (do_ellipsoids == 1 && El_Cutout == 1) {

/* Elipsoid has octant cutout, calculate rms displacement along bond  */

/* calculate multiplier - the 0.9 forces the end of the bond closer to the ellipsoid */
              factor = (float) (drvui->Ellipsoid_Scale * 0.9 / (PI * d * 1.41421356));

/* Mean-square displacement of a scaled thermal ellipsoid along the direction of a bond is given by: */
/*                                                                                      */
/*      D = factor^2 * (X'G'RBR'GX)                                                     */
/*                                                                                      */
/*  where X is the interatomic vector given in triclinic fractional coordinates,        */
/*        G is the lattice metric with the ij'th element given by a[i].a[j],            */
/*        R is the rotational part of the symmetry operation that transform atom i,     */
/*        B is a square matrix of the betaij thermal factors,                           */
/*    and a prime indicates the transpose of the matrix.                                */
/*                                                                                      */
              if (is_ellipsoid[0] == 1) {     /* at one end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[j];
                p1 = ellips_id[j];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1].ellips[p];   /* extract betaij's into square matrix */
                B[0][1] = B[1][0] = drvui->ellips[p1].ellips[3];
                B[0][2] = B[2][0] = drvui->ellips[p1].ellips[4];
                B[1][2] = B[2][1] = drvui->ellips[p1].ellips[5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; ++q) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; r++)
                      elrot0[p][q] += B[p][r] * drvui->rssC[o1][q][r];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) drvui->rssC[o1][p][r] * elrot0[r][q];    /* BP is rotated betij matrix */
                  }
                }
                for (p = 0; p < 3; p++) {
                  Z0[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z0[p] += df[q] * BP[q][p];
                }
/* rms0 is the scaled rms displacement value of the ellipsoid in the direction of the bond */
                rms0 = factor * (float) sqrt (Z0[0] * df[0] +
                          Z0[1] * df[1] + Z0[2] * df[2]);
/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z0[q] = s_vert[3 * j + q] + (s_vert[3 * k + q] - s_vert[3 * j + q]) * rms0 / d;
              }        			 /* if is_ellipsoid[0] */

              if (is_ellipsoid[1] == 1) {     /* ellipsoid at other end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[k];
                p1 = ellips_id[k];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1].ellips[p];     /* extract betaij's into square matrix */
                B[0][1] = B[1][0] = drvui->ellips[p1].ellips[3];
                B[0][2] = B[2][0] = drvui->ellips[p1].ellips[4];
                B[1][2] = B[2][1] = drvui->ellips[p1].ellips[5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; q++) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; ++r)
                      elrot0[p][q] += B[p][r] * drvui->rssC[o1][q][r];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) drvui->rssC[o1][p][r] * elrot0[r][q];    /* BP is rotated betij matrix */
                  }
                }
                for (p = 0; p < 3; p++) {
                  Z1[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z1[p] += df[q] * BP[q][p];
                }
/* rms1 is the scaled rms displacement value of the ellipsoid in the direction of the bond */

                rms1 = factor * (float) sqrt (Z1[0] * df[0] +
        			      Z1[1] * df[1] + Z1[2] * df[2]);

/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z1[q] = s_vert[3 * k + q] - (s_vert[3 * k + q]
                      - s_vert[3 * j + q]) * rms1 / d;
              }        			 /* if is_ellipsoid[1] */
            }        			 /* if any segmented ellipsoids present */

            for (l1 = 0; l1 < 3; l1++) {
              at1[l1] = (float) Z0[l1];
              at2[l1] = (float) Z1[l1];
              df[l1] = (float) (Z1[l1] - Z0[l1]);
            }

            if (clipflag != 0) {
              out = 0;
              l2 = 0;
              l3 = 0;
              clip0 = 0.0f;
              clip1 = 1.0f;
              for (l1 = 0; l1 < 3; l1++) {
                if (o_vert[3 * j + l1] < drvui->frames[drvui->frame_no].clip_lim[l1] - 0.01
                      || o_vert[3 * j + l1] > drvui->frames[drvui->frame_no].clip_lim[l1 + 3] + 0.01)
                  l2 = 1;
                if (o_vert[3 * k + l1] < drvui->frames[drvui->frame_no].clip_lim[l1] - 0.01
                      || o_vert[3 * k + l1] > drvui->frames[drvui->frame_no].clip_lim[l1 + 3] + 0.01)
                  l3 = 1;
              }
              if (l2 && l3)
                out = 1;        	 /* both atoms outside */
              if (l3)
                clip1 = 0.5f;
              if (l2)
                clip0 = 0.5f;

              at2[0] = at1[0] + clip1 * df[0];
              at2[1] = at1[1] + clip1 * df[1];
              at2[2] = at1[2] + clip1 * df[2];

              at1[0] = at1[0] + clip0 * df[0];
              at1[1] = at1[1] + clip0 * df[1];
              at1[2] = at1[2] + clip0 * df[2];
            }        			 /* if clipflag */

            omit = 0;
            for (l1 = 0; l1 < nomits; l1++) {
              if (omit1[l1]== i*1000 && omit2[l1] == Bond_Count) omit = 1;
            }

            if ((clipflag == 0 || out == 0) && omit == 0) {
              fprintf (drvui->fpoutp," /* Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f*/\n",
                       bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
                       bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              if (drvui->bonds[i].bond_style == 0) {
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                         at1[0], at1[1], at1[2], at2[0], at2[1],at2[2],
                         drvui->Bond_Mult * drvui->bonds[i].bond_size);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", col_bond_p);
                fprintf(drvui->fpoutp," }\n");
              } else {
                numdashes = drvui->bonds[i].bond_style;
                dashes = (float) numdashes;

                for (m = 0; m < 3; m++)
                  df[m] = (float) (at2[m] - at1[m]);
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                         at1[0], at1[1], at1[2], at1[0]+df[0]/dashes,at1[1]+df[1]/dashes,at1[2]+df[2]/dashes,
                         drvui->Bond_Mult * drvui->bonds[i].bond_size);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", col_bond_p);
                fprintf (drvui->fpoutp," }\n");
                for (m = 2; m < numdashes - 1; m += 2 ) {
                  mf = (float) m;
                  fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                         at1[0]+df[0]*mf/dashes, at1[1]+df[1]*mf/dashes, at1[2]+df[2]*mf/dashes, 
                         at1[0]+df[0]*(mf+1.)/dashes,at1[1]+df[1]*(mf+1.)/dashes,at1[2]+df[2]*(mf+1.)/dashes, 
                         drvui->Bond_Mult * drvui->bonds[i].bond_size);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", 
                      col_bond_p);
                  fprintf(drvui->fpoutp," }\n");
                }
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                         at1[0]+df[0]*(dashes-1.)/dashes, at1[1]+df[1]*(dashes-1.)/dashes, at1[2]+df[2]*(dashes-1.)/dashes, 
                         at2[0],at2[1],at2[2], drvui->Bond_Mult * drvui->bonds[i].bond_size);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", col_bond_p);
                fprintf(drvui->fpoutp," }\n");
              }
              N_Bond++;
#if 0
  	      if (doAsy) {
#endif  	      
		(void) sscanf (col_bond_v, "%f %f %f", &glr, &glg, &glb);
		fprintf (drvui->fpouta,
			" // Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f\n",
			bnd1[0], bnd1[1], o_vert[3 * j],
			o_vert[3 * j + 1], o_vert[3 * j + 2], bnd2[0],
			bnd2[1], o_vert[3 * k], o_vert[3 * k + 1],
			o_vert[3 * k + 2]);
		if (drvui->bonds[i].bond_style == 0) {	/* solid bonds */
		    fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),",
    		            at1[0], at1[1], at1[2], at2[0], at2[1], at2[2]);
		    fprintf (drvui->fpouta, "rgb(%4.2f,%4.2f,%4.2f)+linewidth(%5.2f) );\n",
			    glr,glg,glb, 25. * drvui->Bond_Mult *
			    drvui->bonds[i].bond_size);
		} else {	/* dashed bonds */
		    numdashes = drvui->bonds[i].bond_style;
		    dashes = (float) numdashes;

		    for (m = 0; m < 3; m++)
			df[m] = (float) (at2[m] - at1[m]);
                    fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),",
                            at1[0], at1[1], at1[2], at1[0] + df[0] / dashes, 
                            at1[1] + df[1] / dashes, at1[2] + df[2] / dashes);
                    fprintf (drvui->fpouta, "rgb(%4.2f,%4.2f,%4.2f)+linewidth(%5.2f)+squarecap );\n",
                            glr,glg,glb, 25. * drvui->Bond_Mult *
                            drvui->bonds[i].bond_size);
                    for (m = 2; m < numdashes - 1; m += 2) {
                        mf = (float) m;
                        fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),",
                                at1[0] + df[0] * mf / dashes,
                                at1[1] + df[1] * mf / dashes,
                                at1[2] + df[2] * mf / dashes,
                                at1[0] + df[0] * (mf + 1.) / dashes,
                                at1[1] + df[1] * (mf + 1.) / dashes,
                                at1[2] + df[2] * (mf + 1.) / dashes);
                        fprintf (drvui->fpouta, "rgb(%4.2f,%4.2f,%4.2f)+linewidth(%5.2f)+squarecap );\n",
                                glr,glg,glb, 25. * drvui->Bond_Mult *
                                drvui->bonds[i].bond_size);
                    }
                    fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),",
                            at1[0] + df[0] * (dashes - 1.) / dashes,
                            at1[1] + df[1] * (dashes - 1.) / dashes,
                            at1[2] + df[2] * (dashes - 1.) / dashes, 
                            at2[0], at2[1], at2[2]);
                    fprintf (drvui->fpouta, "rgb(%4.2f,%4.2f,%4.2f)+linewidth(%5.2f)+squarecap );\n",
                            glr,glg,glb, 25. * drvui->Bond_Mult *
                            drvui->bonds[i].bond_size);
                }
#if 0
              }
#endif
              if (no_comment == 0)
                fprintf (drvui->fpoutv,"# Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f\n",
                         bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
                         bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              for (m = 0; m < 3; m++)
                df[m] = (float) (at2[m] - at1[m]);
              d = (float) sqrt (df[0] * df[0] + df[1] * df[1] + df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
              beta = (float) atan2 (df[0], df[1] + 0.0000001);    /* rotation angle about Z (in radians) */
              gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
              gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
#ifdef OPENGL
              if (gl_view) {
/*                glu_quadObj = gluNewQuadric ();*/
                glPushMatrix ();
                glLoadName(i*1000);
                glPushName(Bond_Count);
                (void)sscanf (col_bond_v, "%f %f %f", &glr, &glg, &glb);
                glColor3f (glr, glg, glb);
                if (drvui->bonds[i].bond_style == 0) {
                  glPushMatrix ();
                  glTranslatef (at1[0], at1[1], at1[2]);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bonds[i].bond_size,
                               drvui->Bond_Mult * drvui->bonds[i].bond_size, d, 10, 1);
                  glPopMatrix ();
                } else {
                  numdashes = drvui->bonds[i].bond_style;
                  dashes = (float) numdashes;
                  glPushMatrix ();
                  glTranslatef (at1[0], at1[1], at1[2]);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bonds[i].bond_size,
                               drvui->Bond_Mult * drvui->bonds[i].bond_size, d/dashes, 10, 1);
                  glPopMatrix ();
                  for (m = 2; m < numdashes; m+=2) {
                    mf = (float) m;
                    glPushMatrix ();
                    glTranslatef (at1[0]+df[0]*mf/dashes, at1[1]+df[1]*mf/dashes,at1[2]+df[2]*mf/dashes);
                    glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                    glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                    gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                    gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bonds[i].bond_size,
                                 drvui->Bond_Mult * drvui->bonds[i].bond_size, d/dashes, 10, 1);
                    glPopMatrix ();
                  }
                }
                glPopName();
                glPopMatrix ();   
              }
#endif                /* ifdef OPENGL */
              if (Vrml2) {
                if (drvui->bonds[i].bond_style == 0) {
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                           (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                           col_bond_v);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                  fprintf (drvui->fpoutv, " height %f", d);
                  fprintf (drvui->fpoutv, "}}]}}\n");
              } else {
                  numdashes = drvui->bonds[i].bond_style;
                  dashes = (float) numdashes;
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                           at1[0] + df[0] / dashes,at1[1] + df[1] / dashes, at1[2] + df[2] / dashes);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                           col_bond_v);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                  fprintf (drvui->fpoutv, " height %f", d/dashes);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                  for (m = 2; m <dashes-1; m += 2) { 
                    mf = (float) m;
                    fprintf (drvui->fpoutv, "  Transform{\n ");
                    fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      at1[0] + df[0] * mf / dashes, at1[1] + df[1] * mf/dashes, at1[2] + df[2] * mf / dashes);
                    fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                    fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                    fprintf (drvui->fpoutv, " children [ Shape {\n");
                    fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                           col_bond_v);
                    fprintf (drvui->fpoutv, "  geometry Cylinder {");
                    fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                    fprintf (drvui->fpoutv, " height %f", d/dashes);
                    fprintf (drvui->fpoutv, "}}]}}\n");
                  }  
                }
              } else {
                if (drvui->bonds[i].bond_style == 0) {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                           (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", col_bond_v);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                  fprintf (drvui->fpoutv, "   height %f\n", d);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                } else {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                           at1[0] + df[0] / dashes, at1[1] + df[1] / dashes, at1[2] + df[2] / dashes);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", col_bond_v);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                  fprintf (drvui->fpoutv, "   height %f\n", d/dashes);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                  for (m = 2; m < dashes - 1; m += 2) {
                    mf = (float) m;                                                 
                    fprintf (drvui->fpoutv, " Separator{\n");
                    fprintf (drvui->fpoutv, "  Transform{\n ");
                    fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                           at1[0] + df[0] *mf/dashes, at1[1] + df[1] *mf/dashes, at1[2] + df[2]*mf/dashes);
                    fprintf (drvui->fpoutv, "  Rotation {\n");
                    fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                    fprintf (drvui->fpoutv, "  Rotation {\n");
                    fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                    fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", col_bond_v);
                    fprintf (drvui->fpoutv, "  }\n");
                    fprintf (drvui->fpoutv, "  Cylinder {\n");
                    fprintf (drvui->fpoutv, "   parts SIDES\n");
                    fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bonds[i].bond_size);
                    fprintf (drvui->fpoutv, "   height %f\n", d/dashes);
                    fprintf (drvui->fpoutv, "  }\n }\n");
                  }
                }
              }
            }
          }        			 /* if distance OK */
        }        			 /* for k */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i (bond types) */
  free (ellips_id);
#ifdef OPENGL
#if !defined (WIN32) && !defined (__APPLE__)
      if (gl_view)      
          gluDeleteQuadric(glu_quadObj);
#endif
#endif
  printf ("%4d bonds.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d bonds.\n\n", Bond_Count);
}        				 /* end of generate_bonds */

/* ************************************************************** */
/* ************************************************************** */

void generate_cones (void)

/* routine to generate lone pair cone descriptions and add them to edit list */
{
  int start, same_type;
  int i, j, k, l;				 /* loop counters */
  int N_Bond;					 /* Counter for number of bonds of a type */
  int l1;					 /* loop counter for matrix operations */
  int nvert1;					 /* vertex counter */
  int Bond_Count;				 /* Counter for bonds output */
  char bnd1[4];					 /* temporary storage of bond labels */
  float d = 0;					 /* bond distance */
  float df[3];					 /* vector difference */
  float beta;					 /* rotation angle */
  float gamma;					 /* another rotation angle */
  float at1[3], at2[3];				 /*initial coordinates of bond endpoints */
  float clip0, clip1;				 /* clipping scale factors */
  int m, l2, l3, out = 0;

  int p;					 /* loop variables for rotation matrix calculation */
  float Z0[3], Z1[3];				 /* components of rms displacement along bond */
  float mybonds[20][20];			 /*temporary storage for atoms bonded to lonepair site */
  int nbnds = 0;
  float vecsum[3], vecpro[3];
  float glr, glg, glb;
#ifdef OPENGL
  GLUquadricObj *glu_quadObj = NULL;
#endif
  char col_cone_p[40];
  char col_cone_v[40];
  
  if (drvui->ncone == 1)
    return;					 /* exit if no work to be done */

#ifdef OPENGL
  if (gl_view)
    glu_quadObj = gluNewQuadric ();
#endif

  Bond_Count = 0;
  for (i = 1; i < drvui->ncone; ++i) {        	 /* loop through cone types */
    if (drvui->cones[i].cone_fn != drvui->frame_no) continue;  /* skip if wrong frame */
    strcpy(col_cone_p,drvui->cones[i].col_cone);
    strcpy(col_cone_v,drvui->cones[i].col_cone);
    Transform_POV_Color(col_cone_p);
    Transform_VRML_Color(col_cone_v);
    nvert = 0;        			         /* empty the vertex list */
    for (j = 0; j < natom; ++j) {        /* loop through atom types */
      if (((drvui->atoms[j].atom_n & 255) > 0) | (((drvui->atoms[j].atom_n >> 24) & 255) > 0)) { /* sphere or ellipsoid */
        int o;

        o = drvui->atoms[j].atom_n >> 24;
        if (o > 0) {                     /* ellipsoid */
          if ((drvui->cones[i].cone_l1[0] == drvui->ellips[o].ellips_l[0]) &&
              (drvui->cones[i].cone_l1[1] == drvui->ellips[o].ellips_l[1]) &&
              (drvui->cones[i].cone_l1[2] == drvui->ellips[o].ellips_l[2]) &&
              (drvui->cones[i].cone_l1[3] == drvui->ellips[o].ellips_l[3]))
            find_all_in_box (j);         /* get all atoms of type j */
        } else {
          o = drvui->atoms[j].atom_n & 255;
          if ((drvui->cones[i].cone_l1[0] == drvui->spheres[o].sphere_l[0]) &&
              (drvui->cones[i].cone_l1[1] == drvui->spheres[o].sphere_l[1]) &&
              (drvui->cones[i].cone_l1[2] == drvui->spheres[o].sphere_l[2]) &&
              (drvui->cones[i].cone_l1[3] == drvui->spheres[o].sphere_l[3])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 /* o > 0 */
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->cones[i].cone_l1[0];
    bnd1[1] = drvui->cones[i].cone_l1[1];
    bnd1[2] = drvui->cones[i].cone_l1[2];
    bnd1[3] = drvui->cones[i].cone_l1[3];
    nvert1 = nvert;
    same_type = 0;

/* atoms i and j are not the same type */
    for (j = 0; j < natom; ++j) {         /* loop through atom types */
      if (((drvui->atoms[j].atom_n & 255) > 0) | (((drvui->atoms[j].atom_n >> 24) & 255) > 0)) { /* sphere or ellipsoid */
        find_all_in_box (j);        	 /* get all atoms of type j */
      }
    }        				 /* for (j=0;j<natom;++j) */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      nbnds = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        nbnds = 0;
        start = nvert1;
        if (same_type)
          start = j;
        start = 0;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if (d > 0.2 && d <= printdist) {     /* here if distance within range */
            mybonds[0][nbnds] = s_vert[3 * k];
            mybonds[1][nbnds] = s_vert[3 * k + 1];
            mybonds[2][nbnds] = s_vert[3 * k + 2];
            nbnds++;
          }        			 /* if distance ok */
        }        			 /* for k */
    /*# */
        if (nbnds < 4 - abs(drvui->cones[i].numlonepairs)) {
          continue;
    }
        df[0] = df[1] = drvui->Bond_Mult * drvui->bonds[i].bond_size;    /* scaling */
        df[2] = 0.5f *d;

        Bond_Count++;        		 /* update count */
        for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
          Z0[p] = s_vert[3 * j + p];
        }

        for (l = 0; l < abs(drvui->cones[i].numlonepairs); l++) {
          for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
            Z1[p] = 0;

            if (drvui->cones[i].numlonepairs == 1) {
              for (k = 0; k < nbnds; k++)
                Z1[p] -= mybonds[p][k] - s_vert[3 * j + p];
              Z1[p] = Z0[p] + Z1[p];
            } else
              if (drvui->cones[i].numlonepairs <0) {
                vecpro[p] = 0.;
                vecpro[0] = (s_vert[3 * j + 1] - mybonds[1][0])
                     * (s_vert[3 * j + 2] - mybonds[2][1])
                     - (s_vert[3 * j + 1] - mybonds[1][1])
                     * (s_vert[3 * j + 2] - mybonds[2][0]);
                vecpro[1] = (s_vert[3 * j + 2] - mybonds[2][0])
                     * (s_vert[3 * j] - mybonds[0][1])
                     - (s_vert[3 * j + 2] - mybonds[2][1])
                     * (s_vert[3 * j] - mybonds[0][0]);
                vecpro[2] = (s_vert[3 * j] - mybonds[0][0])
                     * (s_vert[3 * j + 1] - mybonds[1][1])
                     - (s_vert[3 * j] - mybonds[0][1])
                     * (s_vert[3 * j + 1] - mybonds[1][0]);
               if (l>0)
                Z1[p] = Z0[p] + vecpro[p];
               else
               Z1[p] = Z0[p] - vecpro[p];
               }
             else
              if (drvui->cones[i].numlonepairs == 2) {
                vecsum[p] = 0.;
                vecpro[p] = 0.;
                for (k = 0; k < nbnds; k++) {
                  vecsum[p] += mybonds[p][k] - s_vert[3 * j + p];
                }

                vecpro[0] = (s_vert[3 * j + 1] - mybonds[1][0])
                     * (s_vert[3 * j + 2] - mybonds[2][1])
                     - (s_vert[3 * j + 1] - mybonds[1][1])
                     * (s_vert[3 * j + 2] - mybonds[2][0]);
                vecpro[1] = (s_vert[3 * j + 2] - mybonds[2][0])
                     * (s_vert[3 * j] - mybonds[0][1])
                     - (s_vert[3 * j + 2] - mybonds[2][1])
                     * (s_vert[3 * j] - mybonds[0][0]);
                vecpro[2] = (s_vert[3 * j] - mybonds[0][0])
                     * (s_vert[3 * j + 1] - mybonds[1][1])
                     - (s_vert[3 * j] - mybonds[0][1])
                     * (s_vert[3 * j + 1] - mybonds[1][0]);
                switch (l) {
                case 0:
                  Z1[p] = -vecsum[p] + vecpro[p];
                  break;
                case 1:
                  Z1[p] = -vecsum[p] - vecpro[p];
                  break;
                }
                Z1[p] = Z0[p] + Z1[p];
              }
          }        			 /* for p */

          d = (float) sqrt ((Z0[0] - Z1[0]) * (Z0[0] - Z1[0])
                          + (Z0[1] - Z1[1]) * (Z0[1] - Z1[1])
                          + (Z0[2] - Z1[2]) * (Z0[2] - Z1[2]));
          d = drvui->cones[i].cone_height / d;
          Z1[0] = Z0[0] + d * (Z1[0] - Z0[0]);
          Z1[1] = Z0[1] + d * (Z1[1] - Z0[1]);
          Z1[2] = Z0[2] + d * (Z1[2] - Z0[2]);

          for (l1 = 0; l1 < 3; l1++) {
            at1[l1] = (float) Z0[l1];
            at2[l1] = (float) Z1[l1];
            df[l1] = (float) (Z1[l1] - Z0[l1]);
          }

          if (clipflag != 0) {
            out = 0;
            l2 = 0;
            l3 = 0;
            clip0 = 0.0f;
            clip1 = 1.0f;
            for (l1 = 0; l1 < 3; l1++) {
              if (o_vert[3 * j + l1] < drvui->frames[drvui->frame_no].clip_lim[l1] - 0.01
                 || o_vert[3 * j + l1] > drvui->frames[drvui->frame_no].clip_lim[l1 + 3] + 0.01)
                l2 = 1;
              if (o_vert[3 * k + l1] < drvui->frames[drvui->frame_no].clip_lim[l1] - 0.01
                 || o_vert[3 * k + l1] > drvui->frames[drvui->frame_no].clip_lim[l1 + 3] + 0.01)
                l3 = 1;
            }
            if (l2 && l3)
              out = 1;        		 /* both atoms outside */
            if (l2)
              out = 1;
            if (l3)
              clip1 = 0.5f;
            if (l2)
              clip0 = 0.5f;

            at1[0] = at1[0] + clip0 * df[0];
            at1[1] = at1[1] + clip0 * df[1];
            at1[2] = at1[2] + clip0 * df[2];

            at2[0] = at1[0] + clip1 * df[0];
            at2[1] = at1[1] + clip1 * df[1];
            at2[2] = at1[2] + clip1 * df[2];

          }        			 /* if clipflag */

          if (clipflag == 0 || out == 0) {
            fprintf (drvui->fpoutp," /* LonePair on %c%c %8.5f %8.5f %8.5f */\n",
                     bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            fprintf (drvui->fpoutp," cone{<%8.5f, %8.5f, %8.5f>, %8.5f, <%8.5f, %8.5f, %8.5f>,  %8.5f \n",
                     at1[0], at1[1], at1[2],drvui->Bond_Mult * drvui->cones[i].cone_min, at2[0], at2[1],
                     at2[2], drvui->Bond_Mult * drvui->cones[i].cone_max);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n }\n", col_cone_p);
            fprintf (drvui->fpoutp," sphere{<%8.5f, %8.5f, %8.5f> %8.5f \n",
                     at2[0], at2[1], at2[2], drvui->Bond_Mult * drvui->cones[i].cone_max);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n", col_cone_p);
            N_Bond++;
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# LonePair on: %c%c %8.5f %8.5f %8.5f\n",
                 bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            for (m = 0; m < 3; m++)
                 df[m] = (float) (at2[m] - at1[m]);
            d = (float) sqrt (df[0] * df[0] + df[1] * df[1] + df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
            beta = (float) atan2 (df[0], df[1] + 0.0000001);    /* rotation angle about Z (in radians) */
            gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
            gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */

            sscanf (col_cone_v, "%f %f %f", &glr, &glg, &glb);
#ifdef OPENGL
            if (gl_view) {
              glPushMatrix ();
              glColor3f (glr, glg, glb);
              glTranslatef (at1[0], at1[1], at1[2]);
              glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.0f);
              glRotatef (-gamma * (float) RAD, 1.0f, 0.0f, 0.0f);
/*              glu_quadObj = gluNewQuadric ();*/
              gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
              gluCylinder (glu_quadObj, drvui->Bond_Mult * drvui->cones[i].cone_min,
                 drvui->Bond_Mult * drvui->cones[i].cone_max, d, 10, 1);
              glPopMatrix ();
              glPushMatrix ();
              glTranslatef (at2[0], at2[1], at2[2]);
              glutSolidSphere (drvui->cones[i].cone_max, 10, 10);
              glPopMatrix ();
            }
#endif
            if (Vrml2) {
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                       (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
              fprintf (drvui->fpoutv,"   children Transform { rotation 0 1 0 %f\n", 3.14);
              fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.57 + gamma);
              fprintf (drvui->fpoutv, " children [ Shape {\n");
              fprintf (drvui->fpoutv, " appearance Appearance { material Material {   diffuseColor %s} }\n",
                       col_cone_v);
              fprintf (drvui->fpoutv, "  geometry Cone {");
              fprintf (drvui->fpoutv, "bottomRadius %f", drvui->Bond_Mult * drvui->cones[i].cone_max);
              fprintf (drvui->fpoutv, " height %f", d);
              fprintf (drvui->fpoutv, "}}]}}}\n");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv," translation %8.5f %8.5f %8.5f\n",at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, " children [\n Shape {");
              fprintf (drvui->fpoutv,"appearance Appearance {material Material { diffuseColor %s} }\n",
                       col_cone_v);
              fprintf (drvui->fpoutv,"  geometry Sphere { radius %8.5f}", drvui->cones[i].cone_max);
              fprintf (drvui->fpoutv, "}]}");
            } else {
              fprintf (drvui->fpoutv, " Separator{\n");
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",(at1[0] + at2[0]) / 2.,
                       (at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 1 0  %f\n  }\n", 3.14);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.57 + gamma);
              fprintf (drvui->fpoutv, "  Material {\n   diffuseColor %s\n", col_cone_v);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  Cone {\n");
              fprintf (drvui->fpoutv, "   parts SIDES\n");
              fprintf (drvui->fpoutv, "   bottomRadius %f\n",drvui->Bond_Mult * drvui->cones[i].cone_max);
              fprintf (drvui->fpoutv, "   height %f\n", d);
              fprintf (drvui->fpoutv, "  }\n }\n");
              fprintf (drvui->fpoutv, " Separator {");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n", at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, "  Material {");
              fprintf (drvui->fpoutv, " shininess 0.3\n");
              fprintf (drvui->fpoutv, " diffuseColor %s}\n", col_cone_v);
              fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n", drvui->cones[i].cone_max);
              fprintf (drvui->fpoutv, " }\n");
            }        			 /* if (Vrml2) */
#if 0
  	    if (doAsy) {
#endif  	    
              fprintf (drvui->fpouta, "draw (pic, shift(%5.2f,%5.2f,%5.2f)*rotate(%.2f,Z)*rotate(%.2f,X)*\n",
                      at2[0],at2[1],at2[2],-beta*RAD,-gamma*RAD);
              fprintf (drvui->fpouta, "\tscale(%.2f,%.2f,%.2f)*unitcone,rgb(%.2f,%.2f,%.2f));\n",
                      drvui->Bond_Mult * drvui->cones[i].cone_max,drvui->Bond_Mult * drvui->cones[i].cone_max,-d,glr,glg,glb);
              fprintf (drvui->fpouta, "draw (pic, shift(%5.2f,%5.2f,%5.2f)*scale3(%.2f)*unitsphere,rgb(%.2f,%.2f,%.2f));\n",
                      at2[0],at2[1],at2[2],drvui->cones[i].cone_max,glr,glg,glb);
#if 0
            }
#endif
          }        			 /* if not clipped */

        }        			 /* for l */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i */
#ifdef OPENGL
#if !defined (WIN32) && !defined (__APPLE__)
      if (gl_view)      
          gluDeleteQuadric(glu_quadObj);
#endif
#endif
  fprintf (stderr, "%4d cones.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d cones.\n\n", Bond_Count);
}						 /* end of generate_bonds */

void generate_arrows (void)

/* routine to generate arrow objects for magnetic moments */
{
  int Arrow_Count;		         /* Counter for number output */
  int i, k, l;				 /* loop counters */
  float glr,glb,glg;
#ifdef OPENGL
  GLUquadricObj *glu_quadObj = NULL;
#endif
  float base;
  float x[3], xp[3], yp[3];
  int xi,xj,xk;
  float phi, chi;
  int savedNvertM = NvertM;
  char col_arrow_p[40];
  char col_arrow_v[40];

#ifdef OPENGL
    if (gl_view)
      glu_quadObj = gluNewQuadric ();
#endif

     
  Arrow_Count = 0;
  for (i = 0; i < drvui->nmag; ++i) {	   /* loop through magnetic arrows */
    if (drvui->arrows[i].arrow_fn != drvui->frame_no) continue; /* skip if not right frame */
    strcpy(col_arrow_p,drvui->arrows[i].col_arrow);
    strcpy(col_arrow_v,drvui->arrows[i].col_arrow);
    Transform_POV_Color(col_arrow_p);
    Transform_VRML_Color(col_arrow_v);
    for (k=0;k<3;k++) {
        x[k] = drvui->arrows[i].mag_xp[k] - drvui->xyzoff[k];
    }
    for (xi = -3; xi<=3; xi++) {
        xp[0] = (float)xi;
        for (xj = -3; xj<=3; xj++) {
            xp[1] = (float)xj;
            for (xk=-3; xk<=3; xk++) {
                xp[2] = (float)xk;         /* xp is a magnetic cell translation */
                for (k=0;k<3;k++) {
                    yp[k]=x[k];            /* yp will be position in nuclear cell  */
                    for (l=0;l<3;l++) {
                        yp[k] += drvui->mag_matrix[k][l] * xp[l];
                    }
                }
                add_vert(yp, 1000*i, 0, 0, 0);     /* add vector to vertex list - NOT modulated */
            }
        }
    }
  }                    /* end of loop through magnetic arrows */
  for (k=savedNvertM; k< NvertM; k++) {
    float tmp, tmp1;
    double t;
    float d1, d2, d3, alpha, temp;
    float cosalp;
    i = drvui->atom_no[k]/1000;         /* get number of magnetic item */
    for (l = 0; l<3; l++) {
        x[l] = drvui->arrows[i].mag_xc[l];   /* get magnetic components */
    }
    t = sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
    if (t < 0.002) continue;      /* skip this arrow if 0 length */
    for (l=0; l<3; l++) {
        x[l] /= (float)t;
    }
    t = sqrt(x[0] * x[0] + x[1] * x[1]);
    chi = (float) atan2(t, x[2]);
    if (t > 0.001)
        phi = (float) atan2(x[0]/t,x[1]/t);
    else
        phi = 0.0f;
      (void) sscanf (col_arrow_v, "%f %f %f", &glr, &glg, &glb);
#ifdef OPENGL
    if (gl_view) {
      glLoadName(200000+i);
      glPushName(k);
      glPushMatrix ();
      glColor3f (glr, glg, glb);
      glTranslatef (s_vert[3*k], s_vert[3*k+1], s_vert[3*k+2]);
      glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
      glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
      glTranslatef(0.0,0.0,-drvui->arrows[i].arrow_length/2.0f);
/*      glu_quadObj = gluNewQuadric ();*/
      gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
      gluCylinder (glu_quadObj, drvui->arrows[i].arrow_diam,
              drvui->arrows[i].arrow_diam, 0.9 * drvui->arrows[i].arrow_length, 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
/*      free(glu_quadObj);*/
#endif
      glPopMatrix();
      glPushMatrix();
      glTranslatef (s_vert[3*k], s_vert[3*k+1], s_vert[3*k+2]);
      glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
      glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
      glTranslatef(0.0f,0.0f, drvui->arrows[i].arrow_length * 0.27f);
	
/*      glu_quadObj = gluNewQuadric ();*/
      gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
      gluCylinder (glu_quadObj, drvui->arrows[i].arrow_diam * 2.0,
   	   	   0.01, 0.4*drvui->arrows[i].arrow_length, 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
/*      free(glu_quadObj);*/
#endif
      glPopMatrix ();
      glPopName();
    }
#endif

/* arrow in POV */

    tmp = 0.5f * drvui->arrows[i].arrow_length;
    base = 2.0f * drvui->arrows[i].arrow_diam;
    fprintf (drvui->fpoutp, " object{\n");
    fprintf (drvui->fpoutp, "   union{\n");
    fprintf (drvui->fpoutp,"     object{\n");
    fprintf (drvui->fpoutp,"      cylinder{ < 0,0,%8.5f > , < 0,0,%8.5f >,%8.5f\n",
             -0.5f * drvui->arrows[i].arrow_length, 0.27f * drvui->arrows[i].arrow_length,
             drvui->arrows[i].arrow_diam);
    fprintf (drvui->fpoutp,
             "         texture{pigment{color %s}}}\n", col_arrow_p);
    fprintf (drvui->fpoutp, "    finish{phong %5.2f phong_size %5.2f reflection 0.1}\n     }\n",
             drvui->Phong_Value, drvui->Phong_Size);
    tmp1 = 0.27f * drvui->arrows[i].arrow_length;
    tmp = 0.77f * drvui->arrows[i].arrow_length;
    fprintf (drvui->fpoutp, "     object{\n");
    fprintf (drvui->fpoutp, "      cone{<0, 0, %8.5f>, %8.5f, <0, 0, %8.5f>,  0 \n",
             tmp1, base, tmp); 
    fprintf (drvui->fpoutp,
	     "         texture{pigment{color %s}}}\n", col_arrow_p);
    fprintf (drvui->fpoutp, "     finish{phong %5.2f phong_size %5.2f reflection 0.1}\n     }\n   }\n",
             drvui->Phong_Value, drvui->Phong_Size);
    fprintf (drvui->fpoutp,"   rotate <%8.5f,0,0>\n", -chi * RAD);
    fprintf (drvui->fpoutp,"   rotate <0,0,%8.5f>\n", -phi * RAD);
    fprintf (drvui->fpoutp,"   translate <%8.5f, %8.5f, %8.5f>\n }\n",
             s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
#if 0
    if (doAsy) {
#endif
      fprintf (drvui->fpouta, "draw(pic, (%5.2f,%5.2f,%5.2f)--(%5.2f,%5.2f,%5.2f),\n",
              s_vert[3 * k]-0.8f*drvui->arrows[i].mag_xc[0], 
              s_vert[3 * k + 1]-0.8f*drvui->arrows[i].mag_xc[1], 
              s_vert[3 * k + 2]-0.8f*drvui->arrows[i].mag_xc[2],
              s_vert[3 * k]+1.2f*drvui->arrows[i].mag_xc[0], 
              s_vert[3 * k + 1]+1.2f*drvui->arrows[i].mag_xc[1], 
              s_vert[3 * k + 2]+1.2f*drvui->arrows[i].mag_xc[2]);
      fprintf (drvui->fpouta, "\tlinewidth(%.2f)+rgb(%.2f,%.2f,%.2f),EndArrow3(DefaultHead3,%.2f),currentlight);\n",
              20.f*drvui->arrows[i].arrow_diam,glr,glg,glb,4.f*20.f*drvui->arrows[i].arrow_diam); 
#if 0
      }
#endif

/* arrow in VRML - rotation vector is v cross z */
    d1 = x[1];
    d2 = -x[0];
    d3 = 0.0f;
    temp = (float) sqrt(d1*d1 + d2 * d2 + d3*d3 );
    if (temp < 0.001f) {        /* v is parallel to z - make cross be x */
        d1 = 1.0;
        temp = 1.0;
    }
    d1 /= temp;
    d2 /= temp;
    d3 /= temp;
    cosalp = x[2]/(float)sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);;
    alpha = -(float) acos (cosalp);
    if (Vrml2) {
/* cylinder for shaft of arrow */
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
               d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
               s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children [\n");
      fprintf (drvui->fpoutv, "      Transform{\n");
      fprintf (drvui->fpoutv, "        rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "        translation 0 0 %5.3f\n",
               -0.05f * drvui->arrows[i].arrow_length);
      fprintf (drvui->fpoutv, "        children [\n");
      fprintf (drvui->fpoutv, "          Shape {\n");
      fprintf (drvui->fpoutv, "            geometry Cylinder {\n");
      fprintf (drvui->fpoutv, "              radius %f\n", drvui->arrows[i].arrow_diam);
      fprintf (drvui->fpoutv, "              height %f\n", 0.9f * drvui->arrows[i].arrow_length);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "            appearance Appearance { material Material "
                                               "{ diffuseColor %s}}\n",
                                               col_arrow_v);
      fprintf (drvui->fpoutv, "          }\n");
      fprintf (drvui->fpoutv, "        ]\n");
      fprintf (drvui->fpoutv, "      }\n");
      fprintf (drvui->fpoutv, "    ]\n");
      fprintf (drvui->fpoutv, "  }\n");
/* Cone at end of arrow */
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
               d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
               s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children Transform {\n");
      fprintf (drvui->fpoutv, "      children [\n");
      fprintf (drvui->fpoutv, "        Transform{\n");
      fprintf (drvui->fpoutv, "          rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "          translation 0 0 %5.3f\n",
               0.5f * drvui->arrows[i].arrow_length);
      fprintf (drvui->fpoutv, "          children [\n");
      fprintf (drvui->fpoutv, "            Shape {\n");
      fprintf (drvui->fpoutv, "              geometry Cone {\n");
      fprintf (drvui->fpoutv, "                bottomRadius %f\n",
               2.0 * drvui->arrows[i].arrow_diam);
      fprintf (drvui->fpoutv, "                height %f\n",
               drvui->arrows[i].arrow_length*0.4);
      fprintf (drvui->fpoutv, "              }\n");
      fprintf (drvui->fpoutv, "              appearance Appearance { material Material "
                                               "{   diffuseColor %s}}\n",col_arrow_v);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "          ]\n");
      fprintf (drvui->fpoutv, "        }\n");
      fprintf (drvui->fpoutv, "      ]\n");
      fprintf (drvui->fpoutv, "    }\n");
      fprintf (drvui->fpoutv, "  }\n");
    } else { /* VRML 1 */
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
               s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
               d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", col_arrow_v);
      fprintf (drvui->fpoutv, "  Rotation{rotation 1 0 0 1.5708}\n");
      fprintf (drvui->fpoutv, "  Cylinder {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   radius %f\n", drvui->arrows[i].arrow_diam);
      fprintf (drvui->fpoutv, "   height %f\n", 0.9f * drvui->arrows[i].arrow_length);
      fprintf (drvui->fpoutv, "  }\n }\n");
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", col_arrow_v);
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "   translation 0 0 %5.3f\n  }\n",
               0.5f * drvui->arrows[i].arrow_length);
      fprintf (drvui->fpoutv, "  Cone {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   bottomRadius %f\n",2.0 * drvui->arrows[i].arrow_diam);
      fprintf (drvui->fpoutv, "   height %f\n", drvui->arrows[i].arrow_length*0.4);
      fprintf (drvui->fpoutv, "  }\n }\n");
    }
    Arrow_Count++;
  }
  NvertM = savedNvertM;

#if OPENGL
#if !defined (WIN32) && !defined (__APPLE__)
  if (gl_view)
      gluDeleteQuadric(glu_quadObj);
#endif
#endif
  if (Arrow_Count > 0) {
    fprintf (stderr,"%4d magnetic moment arrows.\n", Arrow_Count);
    fprintf (drvui->flout, "Generated %4d arrows.\n", Arrow_Count);
  }
}						 /* end of generate_arrows */

/* ************************************************************** */
/* ************************************************************** */

void get_atom_id (void)

/* routine to rework atom numbers to point at polyhedron, sphere
   plane, or ellipsoid characteristics list */
{
  int i, j;                              /* local variables */

/* lines commented out with a question mark would be needed for correctness, but
   would break the stishovite example that relies on polyhedral completion carrying
   over from the first frame into the second */


  for (i = 0; i < natom; ++i) {
    drvui->atoms[i].sv_atom_n = drvui->atoms[i].atom_n;    /* save original atom number */
    drvui->atoms[i].atom_n = 0;                     /* clear encoded location */
    for (j = 1; j < drvui->nsphere; ++j) {
      if (check_atom_name(drvui->atoms[i].atom_l, drvui->spheres[j].sphere_l)
      && (drvui->atoms[i].sv_atom_n == drvui->spheres[j].sphere_n || drvui->spheres[j].sphere_n == -1)) {
          drvui->atoms[i].atom_n = j;        	    /* atom is in sphere list */
      }
    }
    for (j = 1; j < drvui->npoly; ++j) {
        if (check_atom_name(drvui->atoms[i].atom_l, drvui->polyhedra[j].poly_l)) {
	      drvui->atoms[i].atom_n += (j << 8);			 /* atom is in poly. list */
        }
    }
    for (j = 1; j < drvui->nplane; ++j) {
        if (check_atom_name(drvui->atoms[i].atom_l,drvui->planes[j].plane_l)) {
	      drvui->atoms[i].atom_n += (j << 16);		 /* atom is in plane list */
        } 
    }
    if (do_ellipsoids) {
      for (j = 1; j < drvui->n_ellips; ++j) {
        if (check_atom_name(drvui->atoms[i].atom_l,drvui->ellips[j].ellips_l)
            && (drvui->atoms[i].sv_atom_n == drvui->ellips[j].ellips_n)) {
          drvui->atoms[i].atom_n += (j << 24);         /* atom is in ellipsoid list */
        }
      }
    }
  }
}						 /* end of get_atom_id  */

/* ************************************************************** */
/* ************************************************************** */

void get_input (int Quick)

/* Routine to process input file, make lattice metric and do preliminary
   ellipsoid processing */
{
  int i, j, k;
  float biso;

  if (!Quick)
    fprintf (drvui->flout, "\n********************* Input: ************************\n");
  read_inp (Quick);
  if (Quick)
    return;
  fprintf (drvui->flout, "*****************************************************\n\n");
  if (!Labels)
    Display_axes = 0;				 /* If no labels, no vectors */
  if (Vrml2)
    printf ("\nGenerating VRML97 Output.\n");
  else
    printf ("\nGenerating VRML1.0 Output.\n");
  print_sym ();                                                 /* print symmetry information */
  make_bmat (drvui->sys, drvui->lat_con, drvui->b_mat,
      drvui->ginv, drvui->rec_lat_con);  /* create the lattice metric */
/* set limits of search regions */
  if (packflag == 0 && boxflag == 0 && drvui->frame_no == 1) {		    /* neither given and frame 1 */
      for (i=0; i<3; i++) {
          drvui->frames[drvui->frame_no].cryst_lim[i] = origin[i] - 0.5f;     /* set to generate one cell */
          drvui->frames[drvui->frame_no].cryst_lim[i + 3] = origin[i] + 0.5f;
      }
      packflag = 1;
  }
  if (packflag != 0) {			 /* pack command given - calculate approximate bounds */
    float temp, temp1;

    for (i = 0; i < 3; i++) {
      temp = origin[i] - drvui->frames[drvui->frame_no].cryst_lim[i];
      temp1 = drvui->frames[drvui->frame_no].cryst_lim[i + 3] - origin[i];
      if (temp1 > temp)
          temp = temp1;
      boxlim[i] = (temp + 0.2f) * drvui->lat_con[i] / 0.5f;
    }
  } else {				 /* bounds command given - calculate approximate pack limits */
    for (i = 0; i < 3; i++) {
      drvui->frames[drvui->frame_no].cryst_lim[i] = origin[i] - boxlim[i] / drvui->lat_con[i] - (float) 0.2;
      drvui->frames[drvui->frame_no].cryst_lim[i + 3] = boxlim[i] / drvui->lat_con[i] - origin[i] + (float) 0.2;
    }
  }
  if ((Options & L_OPT) != 0)
    Calc_Rot (drvui->lookat_v1, drvui->lookat_v2);		 /* Get rotation angles for "lookat" */
  if (drvui->n_ellips > 1) {
    convert_ellipsoid ();                /* change Bij and Uij forms to betaij */
    fprintf (drvui->flout, "\nAnalysis of Thermal Ellipsoids:\n");
    for (i=1; i<drvui->n_ellips; i++) {
      if (drvui->ellips[i].ellips_ismod != 0) {
        fprintf(drvui->flout," (deferred for %c%c%c%c %d due to modulation)\n",
                drvui->ellips[i].ellips_l[0], drvui->ellips[i].ellips_l[1],
                drvui->ellips[i].ellips_l[2], drvui->ellips[i].ellips_l[3],
                drvui->ellips[i].ellips_n);
        continue; /* modulation must be applied to unchanged Uij values */
      }
      if (eigen (&biso, drvui->ellips[i].ellips, drvui->ellips[i].ellips_RMS, drvui->ellips[i].ellips_EV) == 0) {
        fprintf (drvui->flout,"Ellipsoid for %c%c%c%c %d is non-positive definite and has "
                 "been removed from the active list.\n", drvui->ellips[i].ellips_l[0],
                 drvui->ellips[i].ellips_l[1], drvui->ellips[i].ellips_l[2],
                 drvui->ellips[i].ellips_l[3], drvui->ellips[i].ellips_n);
        drvui->ellips[i].ell_type = -100;         /* disable this one */
        drvui->ellips[i].ellips_RMS[0] = (float) sqrt (biso / 78.957);    /* convert Biso to mu */
      } else {
        drvui->ellips[i].ell_type += 1;
        fprintf (drvui->flout,"\n%c%c%c%c %3d Equivalent Isotropic B: %6.2f, U: %6.3f\n",
                 drvui->ellips[i].ellips_l[0], drvui->ellips[i].ellips_l[1],
                 drvui->ellips[i].ellips_l[2], drvui->ellips[i].ellips_l[3],
                 drvui->ellips[i].ellips_n, biso, biso / 78.957);
        fprintf (drvui->flout, "      RMS Amplitudes: %6.3f %6.3f %6.3f\n",
                 drvui->ellips[i].ellips_RMS[0], drvui->ellips[i].ellips_RMS[1],
                 drvui->ellips[i].ellips_RMS[2]);
        fprintf (drvui->flout, "      Eigen Vectors (by row):\n");
        for (j = 0; j < 3; j++) {
            fprintf (drvui->flout, "     ");
            for (k = 0; k < 3; k++)
                fprintf (drvui->flout, "     %8.5f", drvui->ellips[i].ellips_EV[j][k]);
            fprintf (drvui->flout, "\n");
        }
      }
    }
    drvui->Ellipsoid_Scale = P_to_C (drvui->Ellipsoid_Prob);
  }
}						 /* end of get_input */

/* ************************************************************** */
/* ************************************************************** */

void get_label (char input[100], char *c1, char *c2, char *c3, char *c4,
		int strip)

/* Routine to get 1 to 4 character atom ID from input line, and shift input
   line to left  */
{

  int i, j;

  i = 0;
  if (strip != 0)
    while (input[i] != ' ')
      input[i++] = ' ';              /* skip over command */
  while (input[i] == ' ')
    i++;                             /* skip whitespace after command */
  *c1 = input[i];                    /* return first atom label character */
  input[i++] = ' ';
  *c2 = input[i];                    /* return 2nd atom label char */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 2nd not blank, rub it out */
  *c3 = input[i];                    /* return 3rd character */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 3rd not blank, rub it out */
  *c4 = input[i];                    /* return 4th character */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 4th not blank, rub it out */
  while (input[i] == ' ')
    i++;                             /* skip white space after atom label */
  for (j = 0; i < 100; i++) {
    input[j++] = input[i];           /* copy rest of command */
  }
}						 /* end of get_label */

/* ************************************************************** */
/* ************************************************************** */

void Locate_Triple (void)

/* routine to try to place the vector triple where it will be seen in the
POV diagram */
{
  int i, j;
  float temp, pos[3], cpos[3];

  for (i = 0; i < 3; ++i) {			 /* convert origin to Cartesian */
    pos[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      pos[i] -= (float) drvui->b_mat[i][j] * origin[j];
  }
  for (i = 0; i <= 2; ++i) {			 /* calculate position of point after POV rotation */
    cpos[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      cpos[i] += (float) G_Rot[j][i] * pos[j];
  }
  for (i = 0; i < 3; i++) {
    temp = 0.04f *(POV_Max[i] - POV_Min[i]);

    if (cpos[i] > 0)
      pos[i] = POV_Max[i] + temp;
    else
      pos[i] = POV_Min[i] - temp;
  }
/* calculate coordinates of pos before rotation */
  for (i = 0; i < 3; i++) {
    offset[i] = 0.0f;
    for (j = 0; j < 3; j++) {
      offset[i] += (float) G_Rot[i][j] * pos[j];
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void make_bmat (int sys, float lat_con[6], double b_mat[3][3], float ginv[3][3],
                float rec_lat_con[6])
{
  float snal, snbe, snga, csal, csbe, csga, vol;
  int i, j;

  for (i = 3; i <= 5; ++i)
    if (lat_con[i] <= 0)
      lat_con[i] = 90.0f;
  if (sys == 5) {
    lat_con[1] = lat_con[0];
    lat_con[5] = 120.0f;
  }
  if (sys == 4)
    lat_con[1] = lat_con[0];
  if (sys == 6) {
    lat_con[1] = lat_con[0];
    lat_con[2] = lat_con[0];
  }
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      b_mat[j][i] = 0.0f;         /* initialize matrix */
  b_mat[0][0] = 1.0f;
  csal = (float) cos (lat_con[3] / RAD);
  snal = (float) sin (lat_con[3] / RAD);
  csbe = (float) cos (lat_con[4] / RAD);
  snbe = (float) sin (lat_con[4] / RAD);
  csga = (float) cos (lat_con[5] / RAD);
  snga = (float) sin (lat_con[5] / RAD);
  for (i = 0; i < 3; i++)
    ginv[i][i] = lat_con[i] * lat_con[i];
  ginv[0][1] = ginv[1][0] = lat_con[0] * lat_con[1] * csga;
  ginv[0][2] = ginv[2][0] = lat_con[0] * lat_con[2] * csbe;
  ginv[1][2] = ginv[2][1] = lat_con[1] * lat_con[2] * csal;
  b_mat[0][1] = csga;
  b_mat[1][1] = snga;
  b_mat[0][2] = csbe;
  b_mat[1][2] = (csal - csbe * csga) / snga;
  b_mat[2][2] =    (float) sqrt (1.0 - b_mat[0][2] * b_mat[0][2] -
              b_mat[1][2] * b_mat[1][2]);
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      b_mat[j][i] *= lat_con[i];
/* calculate cell volume and reciprocal lengths */
  vol = (float) (b_mat[0][0] * (b_mat[1][1] * b_mat[2][2] - b_mat[1][2] * b_mat[2][1])
         - b_mat[0][1] * (b_mat[1][0] * b_mat[2][2] - b_mat[1][2] * b_mat[2][0])
         + b_mat[0][2] * (b_mat[1][0] * b_mat[2][1] - b_mat[1][1] * b_mat[2][0]));
  rec_lat_con[0] = lat_con[1] * lat_con[2] * snal / vol;
  rec_lat_con[1] = lat_con[0] * lat_con[2] * snbe / vol;
  rec_lat_con[2] = lat_con[1] * lat_con[0] * snga / vol;
  rec_lat_con[3] = (csbe * csga - csal) / (snbe * snga);
  rec_lat_con[4] = (csal * csga - csbe) / (snal * snga);
  rec_lat_con[5] = (csal * csbe - csga) / (snal * snbe);
  drvui->subsys_ref_volume = 1.0f/vol;    /* save the reciprocal volume */

}                                        /* end of make_bmat */

/* ************************************************************** */
/* ************************************************************** */

void modulate_parameters(float vert[3], double *occ, int sym_no, int atom_no)
{
/* routine to generate the occupancy and positional parameters for an atom with
   "average" positional parameters in 'vert'
   This routine relies very heavily on S. van Smaalen, Z. Kristallogr. 219 (2004) 681-691.
   A second reference used for 2 and 3D modulation is Jakubowicz et al., Phys. Rev. B 63 (2001). */

    double x4[3]={0.0,0.0,0.0}, rx4[3]={0.0,0.0,0.0};
    double arg;
    float xadd[3] = {0.0,0.0,0.0};
    float rxadd[3];
    float eps_inv[3][3];
    int id, i1, i2, i3, j, k, theatom;
    int axis;

    *occ = 1.0;
    if (drvui->modulated <= 0) return;
    for (j=0; j<3; j++)
        for (k=0; k<3; k++)
            eps_inv[j][k] = (float)drvui->ss_m[sym_no][j][k];
    (void)matinv(eps_inv);   /* the 'magical' epsilon^(-1) (van Smaalen eq 14) */
    for (k=0; k<drvui->modulated; k++) {
        x4[k] = drvui->phaseshift[k];
        for (j=0; j<3; j++)
            x4[k] += drvui->cell_vec[k][j] * vert[j];   /* contains x4=q1.r, x5=q2.r, x6=q3.r */
    }
    for (k=0; k<3; k++) {
        rx4[k] = -drvui->ts_m[sym_no][k];
        for (j=0; j<3; j++) {
            rx4[k] += eps_inv[k][j] * x4[j];  /* rx4 is x4 transformed by epsilon^(-1)  */
        }
        rx4[k] = fmod(rx4[k],1.0);
        if (rx4[k]<0) rx4[k] += 1.;
    }
    if (drvui->atoms[atom_no].occ_ismod & 2) { /* check for crenel occupancy modulation */
        float lower = drvui->modulate_x[atom_no].atom_occ_crenel[0];
        float upper = drvui->modulate_x[atom_no].atom_occ_crenel[1];
        if ((rx4[0] < lower) || (rx4[0] > upper))
            *occ = 0.0;
        if (((upper - 1.0f) > rx4[0]) || ((lower + 1.0f) < rx4[0]))
            *occ = 1.0;          /* handle the cases when lower < 0 or upper > 1 */
        if (*occ == 0.0)
            return; /* done if crenel occ is zero */
    }

    if (drvui->atoms[atom_no].occ_ismod & 1) {  /* check for Fourier occupancy modulation */
        *occ = drvui->atoms[atom_no].occupancy;
        for (j=0; j<3; j++)
            rx4[j] *= 2.0*PI;
        for (k=0; k<drvui->no_site_occ; k++) { /* loop through the fourier mod vectors */
            id = drvui->modulate_x[k].atom_occpar_id-1;
            theatom = drvui->modulate_x[k].atom_occpar_atom;
            if (theatom != atom_no) continue;
            i1 = drvui->modulate_gbl[id].vector_mult[0];     /* multiplier for first cell mod vector */
            i2 = drvui->modulate_gbl[id].vector_mult[1];     /* multiplier for second cell mod vector */
            i3 = drvui->modulate_gbl[id].vector_mult[2];     /* multiplier for third cell mod vector */
            arg = i1 * rx4[0] + i2 * rx4[1] + i3 * rx4[2];
            *occ += (float)(drvui->modulate_x[k].atom_occpar[0]*cos(arg)
                          + drvui->modulate_x[k].atom_occpar[1]*sin(arg));
        }
        if (*occ < drvui->atoms[atom_no].min_occ) {
            *occ = 0.;
            return;
        }
    }

    if (drvui->atoms[atom_no].atom_ismod & 1){ /* Fourier position modulation */
      if (!(drvui->atoms[atom_no].occ_ismod & 1)) { /* only if not already done in previous step */
        for (j=0; j<3; j++)
            rx4[j] *= 2.0*PI;
        }
        for (j=0; j<drvui->no_site_displace; j++) { /* loop through the fourier mod vectors */
            id = drvui->modulate_3x[j].atom_modpar_id-1;
            axis = drvui->modulate_3x[j].atom_modpar_axis;
            theatom = drvui->modulate_3x[j].atom_modpar_atom;
            if (theatom != atom_no) continue;
            i1 = drvui->modulate_gbl[id].vector_mult[0];     /* multiplier for first cell mod vector */
            i2 = drvui->modulate_gbl[id].vector_mult[1];     /* multiplier for second cell mod vector */
            i3 = drvui->modulate_gbl[id].vector_mult[2];     /* multiplier for third cell mod vector */
            arg = i1 * rx4[0] + i2 * rx4[1] + i3 * rx4[2];
            xadd[axis] += (float)(drvui->modulate_3x[j].atom_modpar[0]*cos(arg)
                         + drvui->modulate_3x[j].atom_modpar[1]*sin(arg));
        }
        for (k = 0; k < 3; k++) {
            rxadd[k] = 0.0;
            for (i1 = 0; i1 < 3; i1++) {
                rxadd[k] += (float)drvui->ss[sym_no][k][i1] * xadd[i1];
            }
            vert[k] += rxadd[k];
        }
    }

    if (drvui->atoms[atom_no].atom_ismod & 2){ /* Position modulation by a sawtooth function*/
        if ((drvui->atoms[atom_no].occ_ismod & 1) ^ (drvui->atoms[atom_no].atom_ismod & 1) ) {  /* undo any previous multiplication by 2pi */
            for (j=0; j<3; j++)
                rx4[j] /= 2.0*PI;
        }
        for (k = 0; k < 3; k++)
            xadd[k] = 2.0f * drvui->modulate_x[atom_no].atom_mod_sawtooth[k]
                * ( ((float)rx4[k] - drvui->modulate_x[atom_no].atom_mod_sawtooth[3])
                /  drvui->modulate_x[atom_no].atom_mod_sawtooth[4] );

        for (k = 0; k < 3; k++) {
            xadd[k] = 0.0;
            for (i1 = 0; i1 < 3; i1++) {
                rxadd[k] += (float)drvui->ss[sym_no][k][i1] * xadd[i1];
            }
            vert[k] += rxadd[k];
        }
    }
}

/* ************************************************************** */
/* ************************************************************** */

int modulate_uij (float vert[3], int ellips_no, int atom_no, int sym_no)
{
    double x4[3]={0.0,0.0,0.0}, rx4[3]={0.0,0.0,0.0};
    double arg;
    float eps_inv[3][3];
    int id, i1, i2, i3, j, k, theatom;
    int term;
    float biso;
    float uij[6];

    for (j=0; j<6; j++)
        uij[j] = drvui->ellips[ellips_no].ellips[j];  /* copy unmodulated values */
    for (k=0; k<drvui->modulated; k++) {
        x4[k] = drvui->phaseshift[k];
        for (j=0; j<3; j++)
            x4[k] += drvui->cell_vec[k][j] * vert[j];   /* contains x4=q1.r, x5=q2.r, x6=q3.r */
    }
    for (k=0; k<3; k++) {
        rx4[k] = drvui->ts_m[sym_no][k];
        for (j=0; j<3; j++) {
            eps_inv[j][k] = (float)drvui->ss_m[sym_no][j][k];
            rx4[k] += drvui->ss_m[sym_no][k][j] * x4[j];  /* rx4 is x4 transformed by the _mm sym op */
        }
        rx4[k] = fmod(rx4[k],1.0);
    }
    (void)matinv(eps_inv);   /* the 'magical' epsilon^(-1) (van Smaalen eq 14) */

    for (j=0; j<3; j++)
        rx4[j] *= 2.0*PI;
    for (k=0; k<drvui->no_site_U_terms; k++) { /* loop through the fourier mod vectors */
        id = drvui->modulate_3t[k].ellips_modpar_id-1;
        theatom = drvui->modulate_x[k].ellips_modpar_atom;
        if (theatom != atom_no) continue;
        term = drvui->modulate_3t[k].ellips_modpar_term;
        i1 = drvui->modulate_gbl[id].vector_mult[0];     /* multiplier for first cell mod vector */
        i2 = drvui->modulate_gbl[id].vector_mult[1];     /* multiplier for second cell mod vector */
        i3 = drvui->modulate_gbl[id].vector_mult[2];     /* multiplier for third cell mod vector */
        arg = i1 * rx4[0] + i2 * rx4[1] + i3 * rx4[2];
        uij[term] += (float)(drvui->modulate_3t[k].ellips_modpar[0]*cos(arg)
                      + drvui->modulate_3t[k].ellips_modpar[1]*sin(arg));
    }

/* FIXME? Does rotation by symmetry need to be done before making the modulation correction? */

/* convert from Uij to betaij */

    for (j = 0; j < 3; ++j)
        uij[j] *= drvui->rec_lat_con[j] * drvui->rec_lat_con[j] * 0.25f;
    uij[3] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[1] * 0.25f;
    uij[4] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[2] * 0.25f;
    uij[5] *= drvui->rec_lat_con[1] * drvui->rec_lat_con[2] * 0.25f;
/* multiply Uij by 8pi^2 */
    for (j = 0; j < 6; ++j)
        uij[j] *= 78.9568f;

/* calculate eigenvalues and eigenvectors */

    if (eigen (&biso, uij, drvui->ellips[ellips_no].ellips_RMS, drvui->ellips[ellips_no].ellips_EV) == 0) {
        fprintf (drvui->flout,"Ellipsoid for %c%c%c%c %d is non-positive definite and has "
                "been removed from the active list.\n", drvui->ellips[ellips_no].ellips_l[0],
                drvui->ellips[ellips_no].ellips_l[1], drvui->ellips[ellips_no].ellips_l[2],
                drvui->ellips[ellips_no].ellips_l[3], drvui->ellips[ellips_no].ellips_n);
        return(1);
    } else {
        fprintf (drvui->flout,"\n%c%c%c%c %3d Equivalent Isotropic B: %6.2f, U: %6.3f\n",
                 drvui->ellips[ellips_no].ellips_l[0], drvui->ellips[ellips_no].ellips_l[1],
                 drvui->ellips[ellips_no].ellips_l[2], drvui->ellips[ellips_no].ellips_l[3],
                 drvui->ellips[ellips_no].ellips_n, biso, biso / 78.957);
        fprintf (drvui->flout, "      RMS Amplitudes: %6.3f %6.3f %6.3f\n",
                 drvui->ellips[ellips_no].ellips_RMS[0], drvui->ellips[ellips_no].ellips_RMS[1],
                 drvui->ellips[ellips_no].ellips_RMS[2]);
        fprintf (drvui->flout, "      Eigen Vectors (by row):\n");
        for (j = 0; j < 3; j++) {
            fprintf (drvui->flout, "     ");
            for (k = 0; k < 3; k++)
                fprintf (drvui->flout, "     %8.5f", drvui->ellips[ellips_no].ellips_EV[j][k]);
            fprintf (drvui->flout, "\n");
        }
    }
    return 0;

}

/* ************************************************************** */
/* ************************************************************** */

void Output_Spheres (float *radii, int i)

/* routine to output the spheres with vertices defined in s_vert and o_vert
   The radius of the sphere is in radii */
{
  float glr, glg, glb;
  int k;
  int outside = 0;
  int l1;
  int omit;
  char sphere_col_p[40];
  char sphere_col_v[40];

  strcpy(sphere_col_p,drvui->spheres[i].sphere_col);
  strcpy(sphere_col_v,drvui->spheres[i].sphere_col);
  Transform_POV_Color(sphere_col_p);
  Transform_VRML_Color(sphere_col_v);
  for (k = 0; k < nvert; ++k) {
    if (clipflag != 0) {

      outside = 0;
      for (l1 = 0; l1 < 3; l1++)
        if (o_vert[3 * k + l1] < drvui->frames[drvui->frame_no].clip_lim[l1] - 0.01 ||
            o_vert[3 * k + l1] > drvui->frames[drvui->frame_no].clip_lim[l1 + 3] + 0.01)
          outside = 1;
    }
    
    omit = 0;
    for (l1=0;l1<nomits;l1++) {
      if (omit1[l1]==i && omit2[l1] ==k ) omit = 1;
    }

        
    if (omit == 0 && (clipflag == 0 || outside == 0)) {
      fprintf (drvui->fpoutp,
	       " /* Sphere for %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
	       drvui->spheres[i].sphere_l[0], drvui->spheres[i].sphere_l[1],
               drvui->spheres[i].sphere_l[2], drvui->spheres[i].sphere_l[3],
	       o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
      fprintf (drvui->fpoutp, " sphere{<%8.5f, %8.5f, %8.5f>, %8.5f\n",
	       s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2], radii[k]);
      fprintf (drvui->fpoutp, "  texture{pigment{color %s  }}\n", sphere_col_p);
      fprintf (drvui->fpoutp,
	       "   finish{phong %5.2f phong_size %5.2f reflection 0.1}\n",
	       drvui->Phong_Value, drvui->Phong_Size);
      fprintf (drvui->fpoutp, " }\n");
      (void) sscanf (sphere_col_v, "%f %f %f", &glr, &glg, &glb);
#ifdef OPENGL
      if (gl_view) {
        glPushMatrix ();
        glLoadName(i);
        glPushName(k);
        glTranslatef (s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        glColor3f (glr, glg, glb);
        glutSolidSphere (radii[k], 10, 10);
        glPopName();
        glPopMatrix ();
      }
#endif
#if 0
      if (doAsy) {
#endif      
        fprintf (drvui->fpouta,
                " // Sphere for %c%c%c%c at %8.5f %8.5f %8.5f  \n",
                drvui->spheres[i].sphere_l[0], drvui->spheres[i].sphere_l[1],
                drvui->spheres[i].sphere_l[2], drvui->spheres[i].sphere_l[3],
	       o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
        fprintf (drvui->fpouta, "draw(pic, shift (%8.5f, %8.5f, %8.5f)*scale3(%.2f)*unitsphere,rgb(%4.2f,%4.2f,%4.2f));\n",
	       s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2], radii[k],glr,glg,glb);
#if 0
      }
#endif
      if (no_comment == 0)
        fprintf (drvui->fpoutv,
                 "# Sphere for %c%c%c%c at %8.5f %8.5f %8.5f \n",
                 drvui->spheres[i].sphere_l[0], drvui->spheres[i].sphere_l[1],
                 drvui->spheres[i].sphere_l[2], drvui->spheres[i].sphere_l[3],
                 o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
      if (Vrml2) {
        fprintf (drvui->fpoutv, "  Transform {");
        fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f\n",
            s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, " children [\n Shape {");
        fprintf (drvui->fpoutv,
                 "appearance Appearance {material Material { diffuseColor %s} }\n",
                 sphere_col_v);
        fprintf (drvui->fpoutv, "  geometry Sphere { radius %8.5f}", radii[k]);
        fprintf (drvui->fpoutv, "}]}");
      } else {
        fprintf (drvui->fpoutv, " Separator {");
        fprintf (drvui->fpoutv, "  Transform {");
        fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n",
                 s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, "  Material {");
        fprintf (drvui->fpoutv, " shininess 0.3\n");
        fprintf (drvui->fpoutv, " diffuseColor %s}\n", sphere_col_v);
        fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n", radii[k]);
        fprintf (drvui->fpoutv, " }\n");
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

float P_to_C (float prob)

/* Routine to convert probability to ellipsoid scaling, C.  Given probability
   'prob', calculate value of C required for ellipsoid,

   C^2 = (delx/sigx)^2 + (dely/sigy)^2 + (delz/sigz)^2,

   where 'prob' is the probability that a random point in the distribution
   will fall inside the ellipsoid.  Table values taken from D.B. Owen,
   "Handbook of Statistical Tables", Addison-Wesley, Reading, Mass., 1962.

   N.B. The present version of the tables only handles probabilities from
   0.01 to 0.99.
*/
#define N_C 99
{
  static double C_table[N_C + 1] =
	{ 0.3389, 0.4299, 0.4951, 0.5479, 0.5932, 0.6334, 0.6699, 0.7035,
    0.7349, 0.7644,
    0.7924, 0.8192, 0.8447, 0.8694, 0.8932, 0.9162, 0.9386, 0.9605,
    0.9818, 1.0026,
    1.0230, 1.0430, 1.0627, 1.0821, 1.1012, 1.1200, 1.1386, 1.1570,
    1.1751, 1.1932,
    1.2110, 1.2288, 1.2464, 1.2638, 1.2812, 1.2985, 1.3158, 1.3330,
    1.3501, 1.3672,
    1.3842, 1.4013, 1.4183, 1.4354, 1.4524, 1.4695, 1.4866, 1.5037,
    1.5209, 1.5382,
    1.5555, 1.5729, 1.5904, 1.6080, 1.6257, 1.6436, 1.6616, 1.6797,
    1.6980, 1.7164,
    1.7351, 1.7540, 1.7730, 1.7924, 1.8119, 1.8318, 1.8519, 1.8724,
    1.8932, 1.9144,
    1.9360, 1.9580, 1.9804, 2.0034, 2.0269, 2.0510, 2.0757, 2.1012,
    2.1274, 2.1544,
    2.1824, 2.2114, 2.2416, 2.2730, 2.3059, 2.3404, 2.3767, 2.4153,
    2.4563, 2.5003,
    2.5478, 2.5997, 2.6571, 2.7216, 2.7955, 2.8829, 2.9912, 3.1365,
    3.3682, 5.9503
  };

  int i;
  double C;

  i = (int) (100.0 * prob);			 /* find correct table entry */
  if (i > N_C)
    i = N_C;
/* interpolate (linear) */
  C = C_table[i - 1] + (prob - 0.01 * i) * (C_table[i] - C_table[i - 1]);
  return ((float) C);
}

/* ************************************************************** */
/* ************************************************************** */

void plot_vrml_poly (int polyno)
{
  int i, j, k, l, m, n;
  float *out;
  int *point, equal;
  char edgecolor[40];

/* get space for unique vertices and a list to point to them */

  if ((out = zalloc ((unsigned) (draw_list * 3 * sizeof (float)))) == NULL) {
    fprintf (stderr, "Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  if ((point =  zalloc ((unsigned) (draw_list * sizeof (int)))) == NULL) {
    free (out);
    fprintf (stderr, "Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  if (!Vrml2)
    fprintf (drvui->fpoutv, "Coordinate3 { point [\n");
  n = 0;
  for (i = 0; i < draw_list; i++) {		 /* build list of unique points */
    j = poly_list[i];
    for (k = i + 1; k <= i + j; k++) {
      l = poly_list[k] - 1;
      for (m = 0; m < 3; m++)
        out[3 * n + m] = s_vert[3 * l + m];
      equal = 0;
      for (m = 0; m < n; m++) {			 /* check if new */
        if ((fabs (out[3 * m] - out[3 * n]) < 0.0002)
           && (fabs (out[3 * m + 1] - out[3 * n + 1]) < 0.0002)
           && (fabs (out[3 * m + 2] - out[3 * n + 2]) < 0.0002)) {
          point[k] = m;				 /* point previously found */
          equal = 1;
          break;
        }
      }
      if (equal == 0)
        point[k] = n++;				 /* new point - save it by incrementing n */
    }
    i = i + j;
  }
  --n;
  if (Vrml2)
    fprintf (drvui->fpoutv,
	     "}\n geometry IndexedFaceSet { coord Coordinate{ point [\n");
  for (l = 0; l < n; l++)			 /* output the unique set of coordinates */
    fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n", out[3 * l], out[3 * l + 1],
	     out[3 * l + 2]);
  fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n", out[3 * n], out[3 * n + 1],
	   out[3 * n + 2]);
  if (Vrml2) {
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, " coordIndex [\n");
  } else {
    fprintf (drvui->fpoutv, "}\n IndexedFaceSet { coordIndex [\n");
  }
  l = 0;
  for (i = 0; i < draw_list - 1; i++) {
    j = poly_list[i];
    l++;
/*    fprintf(drvui->fpoutv,"# %d-sided polygon:\n",j);     */
    for (k = 0; k < j; k++)
      fprintf (drvui->fpoutv, "%d,", point[l++]);
    if (i + j < draw_list - 2) {
      fprintf (drvui->fpoutv, "-1,\n");
    } else {
      if (Vrml2) {
        fprintf (drvui->fpoutv, "-1]\n");
      } else {
        fprintf (drvui->fpoutv, "-1]\n  }\n");
      }
    }
    i = i + j;
  }
  if (Vrml2) {
    fprintf (drvui->fpoutv, " solid FALSE\n");
    fprintf (drvui->fpoutv, " convex TRUE\n");
    fprintf (drvui->fpoutv, " creaseAngle 1.57075\n");
    fprintf (drvui->fpoutv, "  }\n }\n");
  }
  if (edges == 1) {
    if (drvui->polyhedra[polyno].poly_rad_edge != 0.)
      strcpy (edgecolor, drvui->polyhedra[polyno].poly_col_edge);
    else
      strcpy (edgecolor, drvui->col_edge);
    Transform_VRML_Color(edgecolor);

    if (Vrml2) {
      fprintf (drvui->fpoutv, " Shape {\n");
      fprintf (drvui->fpoutv,
	       "appearance Appearance {\n  material Material { diffuseColor %s} \n }\n",
	       edgecolor);
      fprintf (drvui->fpoutv,
	       "geometry IndexedLineSet {  coord Coordinate{ point [\n");
      for (l = 0; l < n; l++)			 /* output the unique set of coordinates */
        fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n", out[3 * l],
             out[3 * l + 1], out[3 * l + 2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n", out[3 * n],
           out[3 * n + 1], out[3 * n + 2]);
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, "  coordIndex [\n");
    } else {
      fprintf (drvui->fpoutv, " Material {\n   diffuseColor %s \n }\n", edgecolor);
      fprintf (drvui->fpoutv, "IndexedLineSet { coordIndex [\n");
    }
    l = 0;
    for (i = 0; i < draw_list - 1; i++) {
      j = poly_list[i];
      l++;
      for (k = 0; k < j; k++)
        fprintf (drvui->fpoutv, "%d,", point[l++]);
      if (i + j < draw_list - 2) {
        fprintf (drvui->fpoutv, "%d,-1,\n", point[l - j]);
      } else {
        if (Vrml2) {
          fprintf (drvui->fpoutv, "%d,-1]\n  ", point[l - j]);
        } else {
          fprintf (drvui->fpoutv, "%d,-1]\n  }\n }\n", point[l - j]);
        }
      }
      i = i + j;
    }
    if (Vrml2)
      fprintf (drvui->fpoutv, "}\n }\n");
  } else {
    if (!Vrml2)
      fprintf (drvui->fpoutv, " }\n");
  }
  free (out);					 /* release the space */
  free (point);
  return;
}

/* ************************************************************** */
/* ************************************************************** */

void print_sym (void)

/* print symmetry information */
{
  int i, j, k;

 if (drvui->origin1_flag) {
      fprintf(drvui->flout, "\nThe current space group has multiple origins. The generator \n"
          "used by DRAWxtl always selects the origin at a center of inversion.\n"
          "If your structure does not display correctly, add the following line\n"
          "to your input (str) file\n"
          "  xyzoff  %.3f %.3f %.3f\n\n",-drvui->origin_offset[0],
            -drvui->origin_offset[1], -drvui->origin_offset[2]);
      fprintf(stderr, "\n**** \aCheck the listing (.out) file for important\n"
          "      info regarding your space group.\n");
  }
  switch (drvui->sys) {
  case 1:
    fprintf (drvui->flout, "Triclinic - ");
    break;
  case 2:
    fprintf (drvui->flout, "Monoclinic - ");
    break;
  case 3:
    fprintf (drvui->flout, "Orthorhombic - ");
    break;
  case 4:
    fprintf (drvui->flout, "Tetragonal - ");
    break;
  case 5:
    fprintf (drvui->flout, "Hexagonal - ");
    break;
  case 6:
    fprintf (drvui->flout, "Cubic - ");
  }
  if (!drvui->acentric) {
    fprintf (drvui->flout, "Centric - ");
  }
  else {
    fprintf (drvui->flout, "Acentric - ");
  }
  switch (drvui->nbr) {
  case 1:
    fprintf (drvui->flout, "P lattice");
    break;
  case 2:
    fprintf (drvui->flout, "A-centered");
    break;
  case 3:
    fprintf (drvui->flout, "B-centered");
    break;
  case 4:
    fprintf (drvui->flout, "C-centered");
    break;
  case 5:
    fprintf (drvui->flout, "F-centered");
    break;
  case 6:
    fprintf (drvui->flout, "I-centered");
    break;
  case 7:
    fprintf (drvui->flout, "R-centered");
    break;
  default:
    fprintf (drvui->flout, "**** Unknown Lattice ***");
  }
  fprintf (drvui->flout, " with %d Symmetry Operators\n\n", drvui->ng);
  for (i = 1; i <= drvui->ng; ++i) {

    for (j = 1; j <= 3; ++j) {
      for (k = 1; k <= 3; ++k) {
	drvui->rss[i - 1][j - 1][k - 1] = (float) drvui->ss[i - 1][j - 1][k - 1];
	if (drvui->ss[i - 1][j - 1][k - 1] != 0) {
	  if (drvui->ss[i - 1][j - 1][k - 1] > 0) {
	    fprintf (drvui->flout, "+");
	  } else {
	    fprintf (drvui->flout, "-");
	  }
	  switch (k) {
	  case 1:
	    fprintf (drvui->flout, "x");
	    break;
	  case 2:
	    fprintf (drvui->flout, "y");
	    break;
	  case 3:
	    fprintf (drvui->flout, "z");
	  }
	}
      }
      if (drvui->ts[i - 1][j - 1] != 0.0) {
        if (drvui->ts[i - 1][j - 1] == 0.5)
          fprintf (drvui->flout, "+1/2");
        if (drvui->ts[i - 1][j - 1] == 0.25)
          fprintf (drvui->flout, "+1/4");
        if (drvui->ts[i - 1][j - 1] == 0.75)
          fprintf (drvui->flout, "+3/4");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.33333333) < 0.00001)
          fprintf (drvui->flout, "+1/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.66666667) < 0.00001)
          fprintf (drvui->flout, "+2/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.16666667) < 0.00001)
          fprintf (drvui->flout, "+1/6");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.83333333) < 0.00001)
          fprintf (drvui->flout, "+5/6");
      }
      if (j != 3)
        fprintf (drvui->flout, ",");
    }
    if (i != drvui->ng) {
      if (i % 4 != 0) {
        fprintf (drvui->flout, " ; ");
      } else {
        fprintf (drvui->flout, "\n");
      }
    }
  }
  Conv_Sym_Mat ();
  fprintf (drvui->flout, "\n");
}						 /* end of print_sym */

/* ************************************************************** */
/* ************************************************************** */

void Token_Strip (char string[], int no)

/*  string[] - input output string */
/*  no - number of tokens to strip */
/* routine to strip tokens from the leading part of a string */
{
  int i, j;
  int l = strlen(string);

  for (i = 0, j = 0; i < no; i++) {		 /* loop through tokens to be removed */
    while (j < l && string[j++] != ' ')
      string[j - 1] = ' ';			 /* characters to space */
    while (j < l && string[j] == ' ')
      j++;					 /* skip all spaces */
  }
  Blank_Strip (string);				 /* move stuff to left */
}

/* ************************************************************** */
/* ************************************************************** */

void usage (void)
{
  fprintf (stderr, "\n****************************************\n\n");
  fprintf (stderr, "  Crystal Structure Drawing Program V%s\n\n", VERSION);
  fprintf (stderr,
	   "Not enough command-line arguments have been specified.\n\n");
  fprintf (stderr, "The full usage is:\n\n");
  fprintf (stderr,
	   "DRAWxtl [-a] [-A start end step] [-b Xb Yb Zb] [-c] [-C] [-m M] [-M <path>] [-n] \n");
  fprintf (stderr,
	   "        [-O] [-o Xo Yo Zo] [-p xlo xhi ylo yhi zlo zhi] [-s/-S c sb] [-u] [-v Xr Yr Zr] <file>\n\n");
  fprintf (stderr,"-a autolabel, i.e. label all atoms\n");
  fprintf (stderr,"-A autogenerate a sequence of POV files for a modulated structure,\n");
  fprintf (stderr,"   covering the t range from \'start\' to \'end\' in steps of size \'step\'\n");
  fprintf (stderr,
	   "-b x-, y-, z-limits of view box in angstroms (20 20 20)\n");
  fprintf (stderr, "-c include comments in the VRML output\n");
  fprintf (stderr, "-C output multiple cameras to the VRML file\n");
  fprintf (stderr, "-F <map-file> read GSAS Fourier .grd file <map-file>\n");
  fprintf (stderr, "-G do NOT generate openGL graphical view\n");
  fprintf (stderr,
	   "-H do NOT display help information at startup of openGL\n");
  fprintf (stderr, "-m magnification factor for image (1.)\n");
  fprintf (stderr, "-M <executable> path to the MSMS surface generation program\n");
  fprintf (stderr, "-n no display of unit cell\n");
  fprintf (stderr, "-O create orthographic view\n");
  fprintf (stderr, "-o x-, y-, z-coordinates of origin (0.5 0.5 0.5)\n");
  fprintf (stderr,
	   "-p limits of fract. coord. along a, b, and c in display box (0 1 0 1 0 1)\n");
  fprintf (stderr,
	   "-Q quick read of input information but no graphics output\n");
  fprintf (stderr,
           "-S c sb: stereo image (using POV-Ray with unofficial stereopov patch)\n");
  fprintf (stderr,
           " <c> = 0/1 for cross-eyed setting, <sb> gives stereo base (eye separation)\n");  
  fprintf (stderr,
           "-s c sb: stereo image (using POV-Ray version 3.7 or later)\n");
  fprintf (stderr, "-u display axial orientation vector triple\n");
  fprintf (stderr, "-v xrot yrot zrot : View rotation angles (0. 0. 0.)\n");
  fprintf (stderr,
	   "<file> is the name of the file describing the structure.\n\n");
  fprintf (stderr, "Arguments enclosed in brackets [] are optional.\n");
  fprintf (stderr, "<file> is required.\n\n");
  exit (1);
}

#ifdef OPENGL
void resizeWindow (int w, int h)
{

  float ratio;
  GLfloat light_position[] = { 0.0f, 0.0f, 100.0f, 0.0f };
  GLfloat light_direction[] = { 0.0f, 0.0f, 1.0f };
  float cpx, cpy;

  /* Prevent a divide by zero, when window is too short
     (you cannot make a window of zero width). */
  if (h == 0)
    h = 1;

  ratio = 1.0f *w / h;

/* Set the viewport to be the entire window */
  glViewport (0, 0, w, h);			 /* urx,ury,llx,lly */

/* Reset the coordinate system before modifying */
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();				 /*initialize rotation matrix */

/* Set the correct perspective. */
  if (!gl_ortho)
      gluPerspective (17., ratio, 0.01, 1000.);	 /* view angle, aspect,near/far clip */
  else {
      if (w<=h)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
      else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
  }
  cpx = (POV_Max[0] + POV_Min[0]) / 2.0f;
  cpy = (POV_Max[1] + POV_Min[1]) / 2.0f;

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  gluLookAt (cpx, cpy, Scale * .50,    /* camera position */
             cpx, cpy, -1.0,           /* camera lookat point */
             0.0f, 1.0f, 0.0f);        /* camera "up" vector */
  glLightfv (GL_LIGHT0, GL_POSITION, light_position);
  glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);

  glClearColor (drvui->glback[0], drvui->glback[1], drvui->glback[2],0.0f);
/* Update size information */
  gl_height = h;
  gl_width = w;
}


void display_crystal (void)
{
  int i;
  char * p;
  GLdouble objx;
  GLdouble objy;
  GLdouble objz;
  int ypos = gl_height - 20;
  float cpx, cpy;
  
  char Text1[] = { "Use Arrow keys to move image. "
      "Use q or <esc> to exit, <HOME> to reset position" };
  char Text2[] = { "Left mouse button to rotate about x,y, z - right to zoom in/out."
                   "POV output file will match last orientation."};
  char Text3[] = { "'c' turns graphics cursor on, 'p' moves it to atom at mouse position" };
  char Text4[] = { "'a' moves graphics cursor to nearest atom, 'l' labels atom, 'b' labels bond" };
  char Text5[] = { "'o' toggles between orthographic and perspective mode"};

  GLfloat light_position[] = {0.0f, 0.0f, 100.0f, 0.0f};
  GLfloat light_direction[] = { 0.0f, 0.0f, 1.0f };
  GLdouble modelMatrix1[16];
  GLdouble projMatrix1[16];
  GLint viewport1[4];
  float ratio=(float)gl_width/(float)gl_height;

  glClear ((GLbitfield) (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));

/*initialize rotation matrix and set the correct perspective. */
/* we could just as well call resizeWindow(gl_width,gl_height); here */
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();				 

  if (!gl_ortho)
    gluPerspective (17., ratio, 0.01, 1000.);	 /* view angle, aspect,near/far clip */
  else {
    if (ratio<1.)
      glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
    else
      glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
  }
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  cpx = (POV_Max[0] + POV_Min[0]) / 2.0f;
  cpy = (POV_Max[1] + POV_Min[1]) / 2.0f;
  gluLookAt (cpx, cpy, Scale * .50,    /* camera position */
             cpx, cpy, -1.0,           /* camera lookat point */
             0.0f, 1.0f, 0.0f);        /* camera "up" vector */

  glLightfv (GL_LIGHT0, GL_POSITION, light_position);
  glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
  glColor3f (0.0f, 0.0f, 0.0f);
  glGetDoublev (GL_MODELVIEW_MATRIX, modelMatrix1);
  glGetDoublev (GL_PROJECTION_MATRIX, projMatrix1);
  glGetIntegerv (GL_VIEWPORT, viewport1);
  if (!view_inst) {
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text1 at 10,gl_height-20 pixels */
    i = 0;
    while (Text1[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text1[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text2 at 10,gl_height-32 pixels */
    i = 0;
    while (Text2[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text2[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text3 at 10,gl_height-44 pixels */
    i =  0;
    while (Text3[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text3[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text4 at 10,gl_height-56 pixels */
    i =  0;
    while (Text4[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text4[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text5 at 10,gl_height-68 pixels */
    i =  0;
    while (Text5[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text5[i++]);
    }
    ypos -= 20;
  }
/*     display_cursor(); */
  if (cur_show) {
    char Text1[100];
    char Text2[100];
    char Text3[100];
    strcpy(Text3,"");
    if (! ReadFourMap)
        sprintf (Text1,"Cursor: %.4f,%.4f,%.4f (step=%.2fA)", 
				cur_cen[0], cur_cen[1], cur_cen[2], cur_step);
    else
        sprintf (Text1,"Cursor: %.4f,%.4f,%.4f (step=%.2fA). Rho=%.3f",
		  cur_cen[0], cur_cen[1], cur_cen[2], cur_step,
		  InterpolateMap (cur_cen[0], cur_cen[1], cur_cen[2]));
    if (cur_atom[0] > 0) {
        sprintf(Text2," Atoms 1.%s",cur_name[0]);
        strcat(Text1, Text2);
    }
    if (cur_atom[1] > 0) {
      sprintf(Text2," 2.%s",cur_name[1]);
      strcat(Text1,Text2);
      sprintf(Text3,"d12=%6.3fA",dist12);
      if (cur_atom[2] > 0) {
        sprintf(Text2," 3.%s",cur_name[2]);
        strcat(Text1,Text2);
        sprintf(Text2,", d23=%6.3fA, a123=%5.2f",dist23,ang123);
        strcat(Text3,Text2);
        if (cur_atom[3] > 0) {
            sprintf(Text2," 4.%s",cur_name[3]);
            strcat(Text1,Text2);
            sprintf(Text2,", d34=%6.3fA, a234=%5.2f, tor1234=%5.2f",dist34,ang234,torsion_ang);
            strcat(Text3,Text2);
        }
      }
    }
    glPushMatrix();
    (void)gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text1 at 10,ypos pixels */
    for (p=Text1;*p;p++)
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, *p);
    ypos -= 12;
    (void)gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text3 12 pixels lower */
    if (strlen(Text3)) {
      for (p=Text3;*p;p++)
        glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, *p);
    }
    glPopMatrix();
  }
  glPushMatrix ();
  glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);
  glRotatef (zrot, 0.0f, 0.0f, 1.0f);
  glRotatef (yrot, 0.0f, 1.0f, 0.0f);
  glRotatef (xrot, 1.0f, 0.0f, 0.0f);
  glTranslatef (-(POV_Max[0] + POV_Min[0]) / 2.0f,
		-(POV_Max[1] + POV_Min[1]) / 2.0f,
		-(POV_Max[2] + POV_Min[2]) / 2.0f);
  draw_cursor ();          /* display openGL cursor */
  glCallList (crystalDL);
  for (drvui->frame_no=1; drvui->frame_no<=drvui->max_frame; drvui->frame_no++)
      generate_gl_texts();
  drvui->frame_no--; /* put frame_no back in range to allow addition of labels */    
  if (ShowMapLegend) MapLegend();
  glGetIntegerv(GL_VIEWPORT,viewport);             /* save matrices for later */
  glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
  glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
  glPopMatrix ();
  glutSwapBuffers ();
}

#ifdef OPENGL
void Dump_View(void)
{
  FILE *fp;
  char file[256];
  int state = GL2PS_OVERFLOW, buffsize = 0;
  int format = GL2PS_EPS;
  int sort;
  int options = GL2PS_DRAW_BACKGROUND|GL2PS_OCCLUSION_CULL|GL2PS_BEST_ROOT|GL2PS_SILENT;
  int nbcol = 16;
  GLint viewport[4];

  float cpx,cpy,cpz;
  float m[16];
  GLfloat mat_specular[] = {  1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat light_ambient[] = { AMBLEV, AMBLEV, AMBLEV, 1.0f };
  GLfloat light_specular[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat mat_shininess[] = { 0.8f };
  GLfloat light_position[] = {0.0f, 1.0f, 1.0f, 0.0f};
  GLfloat light_direction[] = { 0.0f, -1.0f, 0.0f };

  if (ReadFourMap)
     sort = GL2PS_SIMPLE_SORT;
   else
     sort = GL2PS_BSP_SORT;

  glGetIntegerv(GL_VIEWPORT,viewport);
  strcpy(file, drvui->File_Root);
  strcat(file,".ps");

  fp = fopen(file, "wb");

  while(state == GL2PS_OVERFLOW){
    buffsize += 1024*1024;
    gl2psBeginPage(file, "DRAWxtl 4.5", viewport, format, sort, options,
                   GL_RGBA, 0, NULL, nbcol, nbcol, nbcol,
                   buffsize, fp, file);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (M_cameras==0) {
        float ratio=(float)viewport[2]/(float)viewport[3];
        if (ratio <=1.)
         glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
       else
         glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
    } else {
        gluPerspective (17., (GLdouble)viewport[2]/(GLdouble)viewport[3], 0.01, 1000.);    
    }
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    cpx = (POV_Max[0] + POV_Min[0]) / 2.0f;
    cpy = (POV_Max[1] + POV_Min[1]) / 2.0f;
    cpz = (POV_Max[2] + POV_Min[2]) / 2.0f;
    glMaterialfv (GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv (GL_FRONT, GL_SHININESS, mat_shininess);
    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    glEnable (GL_COLOR_MATERIAL);
    glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable (GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glClearColor (drvui->glback[0], drvui->glback[1], drvui->glback[2],0.0f);
    glClear ((GLbitfield)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
    gluLookAt (cpx, cpy, Scale * .50,    /* camera position */
               cpx, cpy, -1.0,           /* camera lookat point */
               0.0f, 1.0f, 0.0f);        /* camera "up" vector */
    glLightfv (GL_LIGHT0, GL_POSITION, light_position);
    glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
    glColor3f(0.0f, 0.0f, 0.0f);
    glPushMatrix ();
    glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);
    quaternion_to_rotmatrix(&Rotq, m);

    glMultMatrixf(m);
/*    glTranslatef (-cpx,-cpy,-cpz); */
    draw_cursor();
    glCallList (crystalDL);
    generate_gl_texts();
    glPopMatrix ();


    state = gl2psEndPage();
  }

  fclose(fp);
  resizeWindow(gl_width, gl_height);
}
#endif
  
static void key (unsigned char k, int x, int y)
{
  int Sense = 1;
  int atomno;

  switch (k) {
  case 27:					 /* Escape */
  case 'Q':
  case 'q':
    close_and_exit();
  case 'D':
  case 'd':                           /* d - dump Postscript view */
#ifdef OPENGL
      if (gl_view) {
          Dump_View();
      }
      break;
#endif
  case 'G':
  case 'g':                           /* d - dump Postscript view */
#ifdef OPENGL
      if (gl_view) {
          Dump_Gif();
      }
      break;
#endif
  case 'H':
  case 'h':                           /* h - toggle display of help text */
    view_inst = 1 - view_inst;
    break;
  case 'C':
  case 'c':                           /* c - if cursor off, turn it on */
      if (cur_show) {                 /*     if cursor on, divide step by 5 */
      cur_step *= 0.2f;               /*     if resulting step < 0.01, turn cursor off */
      if (cur_step < 0.01) {
          cur_show = 0;
          cur_step = 0.5;
      }
    } else {
          cur_show = 1;
    }
    break;
  case 'x':                            /* x - move cursor along +x */
    move_cursor(0,1.0f);
    break;
  case 'y':                            /* y - move cursor along +y */
    move_cursor(1,1.0f);
    break;
  case 'z':                            /* z - move cursor along +z */
    move_cursor(2,1.0f);
    break;
  case 'X':                            /* X - move cursor along -x */
    move_cursor(0,-1.0f);
    break;
  case 'Y':                            /* Y - move cursor along -y */
    move_cursor(1,-1.0f);
    break;
  case 'Z':                            /* Z - move cursor along -z */
    move_cursor(2,-1.0f);
    break;
  case 'o':
  case 'O':                           /* O - toggle between orthographic and perspective projection */
    if (!gl_ortho) {
	    gl_ortho = 1;
        gl_size = max(POV_Max[0] - POV_Min[0],POV_Max[1] - POV_Min[1]);
        gl_size = max(gl_size, POV_Max[2] - POV_Min[2]);
    } else {
        gl_ortho = 0;
    }
    resizeWindow(gl_width,gl_height);
    break;
  case 'P':
  case 'p':
    atomno = find_proj_atom(x, y);
    Update_Cursor_List(cur_cen, atomno); /* add this position to cursor list */
    break;
  case 'a':
  case 'A':			/* find nearest atom, calculate distance */
    atomno = find_atom();
    Update_Cursor_List(cur_cen, atomno); /* add this position to cursor list */
    break;
  case 'M':
      Sense = -1;
  case 'm':
      atomno=Maximize_rho(Sense);      /* find local maximum (minimum if Sense < 0) */
      Update_Cursor_List(cur_cen, atomno); /* add this position to cursor list */
      break;
  case 'L':
  case 'l': {
      unsigned int j, nn, theatom;
      if (strlen(cur_name[0])==0) break;
      if (cur_atom[0]<1) break;
      theatom=0;			    

      if (cur_atom[1]>0) theatom=1;
      if (cur_atom[2]>0) theatom=2;
      if (cur_atom[3]>0) theatom=3;

      drvui->labels[drvui->nlabel].label_fn= drvui->frame_no;	  
      drvui->labels[drvui->nlabel].label_x[0]=cur_cen[0];
      drvui->labels[drvui->nlabel].label_x[1]=cur_cen[1];
      drvui->labels[drvui->nlabel].label_x[2]=cur_cen[2];
      nn=0;
      for (j=0;j<strlen(cur_name[theatom]);j++) {
        drvui->labels[drvui->nlabel].label_label[nn++]=cur_name[theatom][j];
      }
      drvui->labels[drvui->nlabel].label_label[nn]='\0';
      drvui->nlabel++;
      drvui->labels_moved = 1;
      break; }
  case 'B':
  case 'b': {
      int first,second;
      if (dist12<0.01) break;
/*fprintf(stderr,"atoms in cache : %d %d %d\n",cur_atom[0],cur_atom[1],cur_atom[2])' */
      if (cur_atom[0]<1) break;
      if (cur_atom[1]<1) break;
      first=cur_atom[0];
      second=cur_atom[1];
  
      drvui->labels[drvui->nlabel].label_fn = drvui->frame_no;	  
      drvui->labels[drvui->nlabel].label_x[0]=(o_vert[3*second]+o_vert[3*first])/2.0f;
      drvui->labels[drvui->nlabel].label_x[1]=(o_vert[3*second+1]+o_vert[3*first+1])/2.0f;
      drvui->labels[drvui->nlabel].label_x[2]=(o_vert[3*second+2]+o_vert[3*first+2])/2.0f;
      sprintf(drvui->labels[drvui->nlabel].label_label,"%.3f",dist12);
      drvui->nlabel++;
      drvui->labels_moved = 1;
      break; }
  case 'U':                            /* U Undo */
  case 'u':
      if (nomits>0) nomits--;
      fflush(drvui->fpoutp);
      rewind(drvui->fpoutp);
      ftruncate(fileno(drvui->fpoutp),0);
      fflush(drvui->fpoutv);
      rewind(drvui->fpoutv);
      ftruncate(fileno(drvui->fpoutv),0);
      fflush(drvui->fpouta);
      rewind(drvui->fpouta);
      ftruncate(fileno(drvui->fpouta),0);
      glDeleteLists(crystalDL, 1);
      crystal_compute(1,1);          /* regenerate drawing */
#if 0
      if (drvui->noshadow == 1 ) 
          fprintf (drvui->fpoutp, " no_shadow no_reflection}\n");
      else
          fprintf (drvui->fpoutp, " }\n");
      fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
      fprintf (drvui->fpoutp, "}\n");
      if (ShowMapLegend == 1 && drvui->frames[drvui->frame_no].slice > 1) {
        char label[30];
        float glr,glg,glb;
        int j;
        for (j=0;j<256;j++){
          float rho = Map_Info.rhomn+0.0039f*(float)j*(Map_Info.rhomx-Map_Info.rhomn);
          colorramp (rho,&glr,&glg,&glb);
          fprintf (drvui->fpoutp,"cylinder { <%5.3f,%5.3f,%5.3f>,<%5.3f,%5.3f,%5.3f>,0.01\n",
                  -Scale/20.+0.5,Scale/20.-2.5+0.0078f*(float)j,0.,
                  -Scale/20.+0.8,Scale/20.-2.5+0.0078f*(float)j,0.);
          fprintf (drvui->fpoutp," texture{pigment{color red %5.3f green %5.3f blue %5.3f} finish{diffuse 0.25}}\n}\n",glr,glg,glb);
          fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),\n",
                  -Scale/20.+0.5,Scale/20.-2.5+0.0078f*(float)j,0.,
                  -Scale/20.+0.8,Scale/20.-2.5+0.0078f*(float)j,0.);
          fprintf (drvui->fpouta,"rgb(%4.2f,%4.2f,%4.2f) );\n",glr,glg,glb);
        }
        fprintf(drvui->fpoutp,"cylinder { <%5.3f,%5.3f,%5.3f>,<%5.3f,%5.3f,%5.3f>,0.01\n",
	       -Scale/20.+0.4,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f, -Scale/20.+0.9,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f);
        fprintf(drvui->fpoutp," texture{pigment{color Black}}\n}\n");
        fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),\n",
	       -Scale/20.+0.4,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f, -Scale/20.+0.9,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f);
        fprintf (drvui->fpouta,"rgb(%4.2f,%4.2f,%4.2f) );\n",0.0f,0.0f,0.0f);
        for (j=0;j<6;j++) {
          float size;
          float fw=0.2f*(float)j*(Map_Info.rhomx-Map_Info.rhomn)+Map_Info.rhomn;
          sprintf(label,"% 5.3f",fw);
          size = drvui->label_scale * 2.0f * Text_Size;
          fprintf (drvui->fpoutp, "  text { ttf \"crystal.ttf\",\"%s\" 0.15,0\n", label);
          fprintf (drvui->fpoutp, "    scale <%4.2f,%4.2f,%4.2f>\n", size, size, size);
          fprintf (drvui->fpoutp,
		"    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n",
		-Scale/20.+1.,Scale/20.-2.5+0.4f*(float)j,0.0f);
          fprintf (drvui->fpoutp, "  }\n");
          fprintf (drvui->fpouta, " label(pic, \"%s\",(%8.5f,%8.5f,%8.5f));\n",
                  label, -Scale/20.+1.,Scale/20.-2.5+0.4f*(float)j,0.0f);

        }
      }
#endif
      (void) fflush (drvui->fpoutp);
      (void) fflush (drvui->fpouta);
      if (Vrml2)
          fprintf (drvui->fpoutv, "]\n");
      fprintf (drvui->fpoutv, " }\n");
      if (Vrml2) {
          if (yrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
          if (zrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
      }
      (void) fflush (drvui->flout);
      break;
  case 'R':                              /* Redo */
  case 'r':
      fprintf(stderr,"redo: nomits %d, omit1[nomits+1]=%d, omit2 %d\n",nomits,
           omit1[nomits+1],omit2[nomits+1]);
      if (omit2[nomits]+omit1[nomits] != 0) nomits++;
      fflush(drvui->fpoutp);
      rewind(drvui->fpoutp);
      ftruncate(fileno(drvui->fpoutp),0);
      fflush(drvui->fpoutv);
      rewind(drvui->fpoutv);
      ftruncate(fileno(drvui->fpoutv),0);
      fflush(drvui->fpouta);
      rewind(drvui->fpouta);
      ftruncate(fileno(drvui->fpouta),0);
      glDeleteLists(crystalDL, 1);
      crystal_compute(1,1);          /* regenerate drawing */
#if 0
      if (drvui->noshadow == 1 ) 
          fprintf (drvui->fpoutp, " no_shadow no_reflection}\n");
      else
          fprintf (drvui->fpoutp, " }\n");
      fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
      fprintf (drvui->fpoutp, "}\n");
      if (ShowMapLegend == 1 && drvui->frames[drvui->frame_no].slice > 1) {
        char label[30];
        float glr,glg,glb;
        int j;
        for (j=0;j<256;j++){
          float rho = Map_Info.rhomn+0.0039f*(float)j*(Map_Info.rhomx-Map_Info.rhomn);
          colorramp (rho,&glr,&glg,&glb);
          fprintf (drvui->fpoutp,"cylinder { <%5.3f,%5.3f,%5.3f>,<%5.3f,%5.3f,%5.3f>,0.01\n",
                  -Scale/20.+0.5,Scale/20.-2.5+0.0078f*(float)j,0.,
                  -Scale/20.+0.8,Scale/20.-2.5+0.0078f*(float)j,0.);
          fprintf(drvui->fpoutp," texture{pigment{color red %5.3f green %5.3f blue %5.3f} finish{diffuse 0.25}}\n}\n",glr,glg,glb);
          fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),\n",
                  -Scale/20.+0.5,Scale/20.-2.5+0.0078f*(float)j,0.,
                  -Scale/20.+0.8,Scale/20.-2.5+0.0078f*(float)j,0.);
          fprintf (drvui->fpouta,"rgb(%4.2f,%4.2f,%4.2f) );\n",glr,glg,glb);
        }
        fprintf(drvui->fpoutp,"cylinder { <%5.3f,%5.3f,%5.3f>,<%5.3f,%5.3f,%5.3f>,0.01\n",
	       -Scale/20.+0.4,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f, -Scale/20.+0.9,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f);
        fprintf(drvui->fpoutp," texture{pigment{color Black}}\n}\n");
        fprintf (drvui->fpouta," draw(pic, (%8.5f,%8.5f,%8.5f)--(%8.5f,%8.5f,%8.5f),\n",
	       -Scale/20.+0.4,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f, -Scale/20.+0.9,Scale/20.-2.5+0.0078f*256.*(0.f-Map_Info.rhomn)/(Map_Info.rhomx-Map_Info.rhomn),
                0.0f);
        fprintf (drvui->fpouta,"rgb(%4.2f,%4.2f,%4.2f) );\n",0.0f,0.0f,0.0f);
        for (j=0;j<6;j++) {
          float size;
          float fw=0.2f*(float)j*(Map_Info.rhomx-Map_Info.rhomn)+Map_Info.rhomn;
          sprintf(label,"% 5.3f",fw);
          size = drvui->label_scale * 2.0f * Text_Size;
          fprintf (drvui->fpoutp, "  text { ttf \"crystal.ttf\",\"%s\" 0.15,0\n", label);
          fprintf (drvui->fpoutp, "    scale <%4.2f,%4.2f,%4.2f>\n", size, size, size);
          fprintf (drvui->fpoutp,
		"    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n",
		-Scale/20.+1.,Scale/20.-2.5+0.4f*(float)j,0.0f);
          fprintf (drvui->fpoutp, "  }\n");
          fprintf (drvui->fpouta, " label(pic, \"%s\",(%8.5f,%8.5f,%8.5f));\n",
                  label, -Scale/20.+1.,Scale/20.-2.5+0.4f*(float)j,0.0f);
        }
      }
#endif      
      (void) fflush (drvui->fpoutp);
      (void) fflush (drvui->fpouta);
      if (Vrml2)
          fprintf (drvui->fpoutv, "]\n");
      fprintf (drvui->fpoutv, " }\n");
      if (Vrml2) {
          if (yrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
          if (zrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
      }
      (void) fflush (drvui->fpoutv);
      (void) fflush (drvui->flout);
      break;
  }
  glutPostRedisplay ();
}

static void specialkey (int k, int x, int y)
{
  switch (k) {
  case GLUT_KEY_LEFT:
    gl_pos_x -= 2.0f *gl_offset;

    break;
  case GLUT_KEY_RIGHT:
    gl_pos_x += 2.0f *gl_offset;

    break;
  case GLUT_KEY_DOWN:
    gl_pos_y -= 2.0f *gl_offset;

    break;
  case GLUT_KEY_UP:
    gl_pos_y += 2.0f *gl_offset;

    break;
  case GLUT_KEY_HOME:
    gl_pos_x = gl_pos_y = gl_pos_z = 0.0f;
    xrot = xrot_i;
    yrot = yrot_i;
    zrot = zrot_i;
    gl_offset = 1.0;
    tb_init(); /* reset trackball */
    break;
  default:
    break;
  }
  glutPostRedisplay ();
}

static void mouse (int button, int state, int x, int y)
{
  float trans[3];

  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
		  start_picking(x,y);
      } else if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
          if (drvui->nlabel == 1) return;
          Event = DRAG;
          if (c_pick == 0 ) {
              trans[0] = gl_pos_x;
              trans[1] = gl_pos_y;
              trans[2] = gl_pos_z;
              c_pick=pick_label(x,y,gl_width,gl_height, trans);
              Oldx = x;
              Oldy = y;
          }
      } else {
           Event = SPIN;
           Oldx = x;
           Oldy = y;
           c_pick = 0;
           return;
      }
    } else {
        Event = 0;
        c_pick = 0;
    }
  }

  if (button == GLUT_RIGHT_BUTTON) {
     if (state == GLUT_DOWN) {
         Event = ZOOM;
         Oldx = x;
         Oldy = y;
     } else {
         Event = 0;
     }
  }
}

static void mousemotion (int x, int y)
{
    double cosy;
    float m[16];
	float *v, dx, dy;
    float ratio;
    GLdouble objx, objy, objz, objx1, objy1, objz1;

    Mx = x;
    My = y;
    dx = (float)(x - Oldx) * Mscale / gl_width;
    dy = (float)(Oldy - y) * Mscale / gl_height;
    if (!Event || (dx == 0.0f && dy == 0.0f)) return;
    v = Trans;
    if (Event == SPIN ) {
         trackball();
         qmult(&Spin, &Rotq, &Rotq);
    } else if (Event == ZOOM) {
         gl_size -= (dx+dy)/2.0f;
         if (!gl_ortho) {
           gl_pos_z -= (dx+dy)/2.0f;
         } else {
           glMatrixMode(GL_PROJECTION);
           glLoadIdentity();
           ratio = 1.0f*gl_width/gl_height;
           if (gl_width<=gl_height)
             glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
           else
             glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
           glMatrixMode(GL_MODELVIEW);
	 }
         glutPostRedisplay ();
         return;
    } else if (Event == DRAG) {
      if (c_pick != 0) {
          float factor;
          if (drvui->nlabel == 1) return;
          if (gl_ortho)
              factor = gl_size * Scale *  0.00002f;
          else
              factor = (float)gl_width * Scale * 0.00003f;
          drvui->labels_moved = 1;
          gluUnProject(Oldx, viewport[3]-Oldy, 0, modelMatrix, projMatrix, viewport,
                      &objx, &objy, &objz);
          gluUnProject(x, viewport[3]-y, 0, modelMatrix, projMatrix, viewport,
                      &objx1, &objy1, &objz1);
          if (c_pick != drvui->triple[0]) {
              drvui->labels[c_pick].label_x[0] += factor * (float)(objx1 - objx);
              drvui->labels[c_pick].label_x[1] += factor * (float)(objy1 - objy);
              drvui->labels[c_pick].label_x[2] += factor * (float)(objz1 - objz);
          } else {
              offset[0] += 5.0f * factor * (float)(objx1 - objx);
              offset[1] += 5.0f * factor * (float)(objy1 - objy);
              offset[2] += 5.0f * factor * (float)(objz1 - objz);
          }
          Oldx=x;
          Oldy=y;
          glutPostRedisplay();
          return;
       }
    }  

/* recalculate rotation angles. */

    quaternion_to_rotmatrix(&Rotq, m);

    yrot = -(float)(asin(m[2]) * RAD);
    cosy = cos(yrot/RAD);
    if (fabs(cosy) > 1.0e-3) {
        xrot = (float)(atan2(m[6]/cosy,m[10]/cosy)*RAD);
        zrot = (float)(atan2(m[1]/cosy,m[0]/cosy)*RAD);
    } else {
        zrot = 0.0f;
        xrot = (float)(atan2(-m[9],m[5])*RAD);
    }
/* redraw the window */
    glutPostRedisplay ();
}

static void start_picking (int x, int y)
{
    GLint viewport[4];
    int hits;
    int i, n;
    float m[16];
    float cpx, cpy, cpz;
    float  ratio = 1.0f *gl_width / gl_height;
 
    memset(selectBuf,0,BUFSIZE);
    glGetIntegerv(GL_VIEWPORT,viewport);

    glPushMatrix();
    glSelectBuffer(BUFSIZE,selectBuf);
    glRenderMode(GL_SELECT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
         
    gluPickMatrix(x,viewport[3]-y,7,7,viewport);

    if (!gl_ortho)
      gluPerspective(17.,(GLdouble)ratio,0.01,1000.);
    else {
      if (gl_width <= gl_height)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
      else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
    } 
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glInitNames();
    glPushName(~0);


    cpx = (POV_Max[0] + POV_Min[0]) / 2.0f;
    cpy = (POV_Max[1] + POV_Min[1]) / 2.0f;
    cpz = (POV_Max[2] + POV_Min[2]) / 2.0f;
            
    gluLookAt (cpx, cpy, Scale * .50,   /* camera position */
               cpx, cpy, -1.0,          /* camera lookat point */
               0.0f, 1.0f, 0.0f);       /* camera "up" vector */
                                              
    glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);

    quaternion_to_rotmatrix(&Rotq,m);
    glMultMatrixf(m);

    glCallList(crystalDL);

    glPopMatrix();
    
    hits = glRenderMode(GL_RENDER);
    glFlush();
  
    if (hits > 0) {
         process_hits (hits,selectBuf);
    } else {
        float trans[3];
        trans[0] = gl_pos_x;
        trans[1] = gl_pos_y;
        trans[2] = gl_pos_z;
        hits = pick_label(x, y, gl_width, gl_height, trans);
        if (hits > 0 && hits <= drvui->nlabel) {
            if (hits != drvui->triple[0] && hits != drvui->triple[1]
                && hits != drvui->triple[2] && hits != drvui->triple[3]) {
                for (n=hits; n<drvui->nlabel; n++) {         /* remove label, but not a, b, c, or origin */
                    drvui->labels[n].label_fn = drvui->labels[n+1].label_fn;
                    strcpy(drvui->labels[n].label_label,drvui->labels[n+1].label_label);
                    for (i=0; i<3; i++)
                        drvui->labels[n].label_x[i] = drvui->labels[n+1].label_x[i];
                    for (i=0; i<4; i++)
                        if (drvui->triple[i] == n+1)
                            drvui->triple[i] = n;
                }
                drvui->nlabel--;
            }
        }
    }
    glutPostRedisplay();

}

void process_hits (GLint hits, GLuint buffer[])
{
    unsigned int i,j;
    GLuint names,*ptr,minZ, *ptrNames=NULL,numberOfNames;
    int handled=0;

    printf ("hits = %d\n", (int) hits);
    numberOfNames=0;
    
    ptr = (GLuint *) buffer;
    minZ = 0xffffffff;
    for (i = 0; i < (unsigned)hits; i++) { 
        names = *ptr;
        ptr++;
        if (*ptr < minZ && *(ptr+2) != -1) {
            numberOfNames = names;
            minZ = *ptr;
            ptrNames = ptr+2;
        }   
        ptr += names+2;
    } 
    if (numberOfNames==0) return;
	
    ptr = ptrNames;
    if (*ptr >200000) {
   	    for (j=0;j<(unsigned)nomits;++j) {
   		    if ((unsigned)omit1[j]==*ptr && (unsigned)omit2[j]==*(ptr+1)){
		        omit3[j]++;
		        if (omit3[j]>6) omit3[j]=1;
                handled=1;
            }
        }	
   	} 
    if (handled==0) {	
       if (*ptr >200000) omit3[nomits]=1;
       omit1[nomits]=*ptr;
       ptr++;
       omit2[nomits++]=*ptr;
    }
    fflush(drvui->fpoutp);
    rewind(drvui->fpoutp);
    ftruncate(fileno(drvui->fpoutp),0);
    fflush(drvui->fpoutv);
    rewind(drvui->fpoutv);
    ftruncate(fileno(drvui->fpoutv),0);
    fflush(drvui->fpouta);
    rewind(drvui->fpouta);
    ftruncate(fileno(drvui->fpouta),0);
    glDeleteLists(crystalDL, 1);
    crystal_compute(1,1);          /* regenerate drawing */
#if 0
    fprintf (drvui->fpoutp, " }\n");
    fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
    fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
    fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
    fprintf (drvui->fpoutp, "}\n");
#endif
    (void) fflush (drvui->fpoutp);
    if (Vrml2)
        fprintf (drvui->fpoutv, "]\n");
    fprintf (drvui->fpoutv, " }\n");
    if (Vrml2) {
        if (yrot != 0.0)
            fprintf (drvui->fpoutv, " }\n");
        if (zrot != 0.0)
            fprintf (drvui->fpoutv, " }\n");
    }

    (void) fflush (drvui->flout);
}

static int pick_label (int x, int y, int w, int h, float trans[3])
{
/* picks the label at the current cursor position */

    GLint viewport[4];
    int i;
    GLuint names,*ptr,minZ;
    GLuint *ptrNames = 0, numberOfNames=0;
    int hits;
    int label;
    float  ratio = 1.0f *(float)w / (float)h;

    glGetIntegerv(GL_VIEWPORT,viewport);
    memset(selectBuf,0,BUFSIZE);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glSelectBuffer(BUFSIZE,selectBuf);
    glRenderMode(GL_SELECT);

    gluPickMatrix(x,viewport[3]-y, 25, 25,viewport);
    if (!gl_ortho)
        gluPerspective(17.,ratio,0.01,1000.);
    else {
        if (gl_width <= gl_height)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-10000.f,10000.f);
        else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-10000.f,10000.f);
         }
                                            
    glMatrixMode(GL_MODELVIEW);
    glInitNames();
    glPushName(~0);
    glTranslatef (trans[0], trans[1], trans[2]);
    glRotatef (zrot, 0.0f, 0.0f, 1.0f);
    glRotatef (yrot, 0.0f, 1.0f, 0.0f);
    glRotatef (xrot, 1.0f, 0.0f, 0.0f);
    glTranslatef (-(POV_Max[0] + POV_Min[0]) / 2.0f,
		-(POV_Max[1] + POV_Min[1]) / 2.0f,
		-(POV_Max[2] + POV_Min[2]) / 2.0f);
    generate_gl_texts();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode (GL_MODELVIEW);
    glFlush();

    hits = glRenderMode(GL_RENDER);
    if (hits == 0) return(0);
    ptr = (GLuint *) selectBuf;
    minZ = 0xffffffff;
    for (i = 0; i < hits; i++) { 
        names = *ptr;
        ptr++;
        if (*ptr < minZ && *(ptr+2) != (GLuint)-1) {
            numberOfNames = names;
            minZ = *ptr;
            ptrNames = ptr+2;
        }  
        ptr += names+2;
    }
    if (numberOfNames==0) return(0);
    ptr = ptrNames;
    label=*ptr;
    if (label >= drvui->nlabel) return(0); /* not a label ? */
    return(label);
}

#endif

/* ************************************************************** */
/* ************************************************************** */

float *polygon_normal_3d ( int n, float v[] )
/* helper function for not_in_slab(), simplified from John Burkardt's geometry.c */
{
  int i;
  int j;
  float normal_norm;

  static float normal[3];
  float v1[3];
  float v2[3];
  float p[3];

   normal[0]=normal[1]=normal[2]=0.;

  for ( i = 0; i < 3; i++ )
  {
    v1[i] = v[i+1*3] - v[i+0*3];
  }

  for ( j = 2; j < n; j++ )
  {
    for ( i = 0; i < 3; i++ )
    {
      v2[i] = v[i+j*3] - v[i+0*3];
    }

    vcross (v1,v2,p);

    for ( i = 0; i < 3; i++ )
    {
      normal[i] = normal[i] + p[i];
    }

    for ( i = 0; i < 3; i++ )
    {
      v1[i]= v2[i];
    }

  }
  normal_norm = (float)sqrt (normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);

  if ( normal_norm != 0.0 )
  {
    for ( i = 0; i < 3; i++ )
    {
      normal[i] = normal[i] / normal_norm;
    }
  }

  return normal;
}

float polygon_solid_angle_3d(int n, float v[], float p[3])
/* helper function for not_in_slab(), simplified from John Burkardt's geometry.c */
{
  float a[3];
  double angle;
  float area = 0.0;
  float b[3];
  float c[3];
  int j;
  int jp1;
  float normal1[3];
  float normal1_norm;
  float normal2[3];
  float normal2_norm;
  float *plane;
  float r1[3];
  double s;
  float value;

  if (n<3) return 0.0;

  plane = polygon_normal_3d ( n, v );

  a[0] = v[0+(n-1)*3] - v[0+0*3];
  a[1] = v[1+(n-1)*3] - v[1+0*3];
  a[2] = v[2+(n-1)*3] - v[2+0*3];

  for ( j = 0; j < n; j++ )
  {
    r1[0] = v[0+j*3] - p[0];
    r1[1] = v[1+j*3] - p[1];
    r1[2] = v[2+j*3] - p[2];

    jp1 = j+1;
    if (jp1 > n-1) jp1=0;

    b[0] = v[0+jp1*3] - v[0+j*3];
    b[1] = v[1+jp1*3] - v[1+j*3];
    b[2] = v[2+jp1*3] - v[2+j*3];

    vcross (a,r1,normal1);

    normal1_norm = (float)sqrt(normal1[0]*normal1[0]+normal1[1]*normal1[1]+normal1[2]*normal1[2]);

    vcross(r1,b,normal2);

    normal2_norm = (float)sqrt(normal2[0]*normal2[0]+normal2[1]*normal2[1]+normal2[2]*normal2[2]);

    s = vdot ( normal1, normal2 )
      / ( normal1_norm * normal2_norm );

   if (s <= -1.) {
     angle=PI;
   } else if (s >= 1.) {
     angle = 0.;
   } else
     angle = acos ( s );

    vcross (a,plane, c);
    s =  vdot (b,c);

    if ( 0.0 < s )
      area = area + (float)(PI - angle);
    else
      area = area + (float)(PI + angle);

    a[0] = -b[0];
    a[1] = -b[1];
    a[2] = -b[2];

  }
  area = area - (float)(PI * ( double ) ( n - 2 ));

  if ( 0.0 < vdot ( plane, r1 ) ) 
    value = -area;
  else 
    value = area;

  return value;
}


int not_in_slab(float x, float y, float z)
/* adapted from polyhedron_contains_point_3d (Carvalho et al. Graphics Gems V)
   as contained in John Burkardt's geometry.c on http://www.csit.fsu.edu/~burkardt/ */

{
  float area;
  int face;
  int i,k;
  int node;
  int node_num_face=4;
  float v_face[3*4];
  float p[3];
  static const int face_point[4*6]={1,2,4,3, 1,3,6,5, 1,5,7,2, 5,6,8,7, 2,7,8,4, 3,4,8,6} ; 


  p[0]=x;
  p[1]=y;
  p[2]=z;

  area=0.0;
  for (face=0;face<6;face++) {

    for (k=0;k<4;k++) {
      node = face_point[k+face*4];
      for (i=0;i<3;i++) {
        v_face[i+k*3] = slabv[i+(node-1)*3];
      }
    }
    area = area + polygon_solid_angle_3d(node_num_face,v_face,p);
  }
  if (area < -2.* PI || 2. * PI < area ) 
    return 0;
  else
    return 1;
}

/* ************************************************************** */
/* ************************************************************** */

float dot0_3d (float x0, float y0, float z0, float x1, float y1, float z1,
	       float x2, float y2, float z2)
{
/* 3D dot product of (P1-P0) and (P2-P0), John Burkardt */

  return (x1 - x0) * (x2 - x0) + (y1 - y0) * (y2 - y0) +
          (z1 - z0) * (z2 - z0);
}

/* Initialize the trackball interface. */

void tb_init(void)
{

	/* Set the default trackball size and mouse scale factor. */

	Tbsize = 0.5f;
	Mscale = 2.5f;

	/* Default is no rotation or translation. */

	Spin.s = 1.0;
	vzero(Spin.v);
	Rotq.s = 1.0;
	vzero(Rotq.v);
	vzero(Trans);

	/* Initialize the other instance variables.  The rest are set the
	first time the user clicks the mouse. */

	Event = 0;
	idle_redraw_ = FALSE;
	changed = TRUE;
    Rotq = XYZ_Rot_to_Q(xrot, yrot, zrot);
}

void trackball(void)
{
	float p1x, p1y, p2x, p2y;
	float theta, t;
	XYZ p1, p2, axis, d;

	QUAT *q = &Spin;

	if (Mx == Oldx && My == Oldy) {

		/* Zero rotation. */

		q->s = 1.;
		vzero(q->v);
		return;
	}
	/* Scale the old and new mouse positions to (-1, 1). */

	window_to_tb((float)Oldx, (float)Oldy, &p1x, &p1y);
	window_to_tb((float)Mx, (float)My, &p2x, &p2y);

	/*First, figure out z-coordinates for projection of P1 and P2 to
	  the deformed sphere.                                            */

	vset(p1, p1x, p1y, tb_project_to_sphere(Tbsize, p1x, p1y));
	vset(p2, p2x, p2y, tb_project_to_sphere(Tbsize, p2x, p2y));

	/* Now the axis of rotation. */

	vcross(p1, p2, axis);

	/* Figure out how much to rotate around that axis. */

	vsub(p1, p2, d);
	t = vlength(d) * Mscale / 20.0f;
	if (t > 1.0) t = 1.0;
	if (t < -1.0) t = -1.0;
	theta = (float)asin(t);

	/* Return a rotation quaternion */

	axis_to_quaternion(axis, theta, q);
}

/* Map mouse click mx, my to a more convenient (-1.0, 1.0) range, based on
window size. */

void window_to_tb(float mx, float my, float *x, float *y)
{
	*x = (2.0f * mx) / (float)gl_width - 1.f;
	*y = 1.0f - (2.0f * my) / (float)gl_height;
}

/* Project an x, y pair onto a sphere of radius r OR a hyperbolic sheet if we
are away from the center of the sphere.  */

float tb_project_to_sphere(float r, float x, float y)
{
	float d, t, z;

	d = (float)sqrt(x * x + y * y);
	if (d < r * 0.70710678118654752440) {        /* inside sphere */
		z = (float)sqrt(r * r - d * d);
	} else {                        /* on hyperbola */
		t = r / (float)sqrt(2.0);
		z = t * t / d;
	}
	return (z);
}

/* Given an axis and an angle, compute a rotation quaternion. */

void axis_to_quaternion(float *axis, float theta, QUAT *quat)
{
	quat->s = (float)cos(theta / 2.0);
	vcopy(axis, quat->v);
	vnormalize(quat->v);
	vscale(quat->v, (float)sin(theta / 2.0));
}

/* Given two rotations, q1 and q2, expressed as quaternions,
quaternion multiplication yields the equivalent product rotation.
This routine also normalizes the result every COUNT times it is
called, to keep error from creeping in. */

#define COUNT 100

void qmult(QUAT *q1, QUAT *q2, QUAT *dest)
{
	XYZ v1, v2;
	QUAT q;
	static int count = 0;

	vcopy(q1->v, v1);
	vcopy(q2->v, v2);
	vscale(v1, q2->s);
	vscale(v2, q1->s);
	vcross(q1->v, q2->v, q.v);
	vadd(v1, q.v, q.v);
	vadd(v2, q.v, q.v);
	q.s = q1->s * q2->s - vdot(q1->v, q2->v);

	*dest = q;

	if (++count > COUNT) {
		count = 0;
		qnormalize(dest);
	}
}

/* Normalize greatest component, to avoid problems that occur when the
component we're normalizing gets close to zero (and the other components may
add up to more than 1.0 because of rounding error).  */

void qnormalize(QUAT *q)
{
	int which, i;
	float gr, v[4];

	vcopy(q->v, v);
	v[3] = q->s;

	which = 0;
	gr = v[0];
	for (i = 1; i < 4; i++) {
		if (fabs(v[i]) > fabs(gr)) {
			gr = v[i];
			which = i;
		}
	}

	v[which] = 0.;
	v[which] = (float)sqrt(1.0 - (v[0] * v[0] + v[1] * v[1] +
		v[2] * v[2] + v[3] * v[3]));

	/* Check to see if we need negative square root. */

	if (gr < 0.0) {
		v[which] = -v[which];
	}
	vcopy(v, q->v);
	q->s = v[3];
}

/* Build a rotation matrix from a rotation quaternion. */

void quaternion_to_rotmatrix(QUAT *q, float *m)
{
	float w, x, y, z;

	w = q->s;
	x = q->v[0];
	y = q->v[1];
	z = q->v[2];

	/* Column major order, used to left multiply points, i.e., P'=M*P */

	m[0]  = 1.f - 2.f * (y * y + z * z);
	m[1]  = 2.f * (x * y + w * z);
	m[2]  = 2.f * (x * z - w * y);
	m[3]  = 0.;

	m[4]  = 2.f * (x * y - w * z);
	m[5]  = 1.f - 2.f * (x * x + z * z);
	m[6]  = 2.f * (y * z + w * x);
	m[7]  = 0.;

	m[8]  = 2.f * (x * z + w * y);
	m[9]  = 2.f * (y * z - w * x);
	m[10] = 1.f - 2.f * (x * x + y * y);
	m[11] = 0.;

	m[12] = 0.;
	m[13] = 0.;
	m[14] = 0.;
	m[15] = 1.;
}

float *vnew(float x, float y, float z)
{
	float *v;

	v = zalloc(3 * sizeof(float));
	v[0] = x;
	v[1] = y;
	v[2] = z;

	return (v);
}

void vset(float *v, float x, float y, float z)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
}

void vcopy(float *vsrc, float *vdst)
{
	vdst[0] = vsrc[0];
	vdst[1] = vsrc[1];
	vdst[2] = vsrc[2];
}

void vzero(float *v)
{
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;
}

void vnormalize(float *v)
{
#if defined(_IEEE) || defined(_IEEE_FP)
	vscale(v, 1.0 / vlength(v));
#else
	float l;

	l = vlength(v);
	if (l > EPS) {
		vscale(v, 1.0f / l);
	}
#endif
}

float vlength(float *v)
{
	return (float)(sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}

void vscale(float *v, float f)
{
	v[0] *= f;
	v[1] *= f;
	v[2] *= f;
}

void vmult(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] * src2[0];
	dst[1] = src1[1] * src2[1];
	dst[2] = src1[2] * src2[2];
}

void vadd(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] + src2[0];
	dst[1] = src1[1] + src2[1];
	dst[2] = src1[2] + src2[2];
}

void vsub(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] - src2[0];
	dst[1] = src1[1] - src2[1];
	dst[2] = src1[2] - src2[2];
}

float vdot(float *v1, float *v2)
{
	return (v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]);
}

void vcross(float *v1, float *v2, float *cross)
{
	XYZ temp;

	  /* x        y       z         z       y */
	temp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
	temp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
	temp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
	vcopy(temp, cross);
}

QUAT XYZ_Rot_to_Q(float Xrot, float Yrot, float Zrot)
{
/* routine to convert the X,Y,Z rotations into a quaternion */
    float * axis;
    QUAT q1,q2,q3,qtemp,Rotq;

    axis = zalloc(3 * sizeof(float));
    axis[2] = axis[1] = 0.0;
    axis[0] = 1.0f;
    axis_to_quaternion(axis, (float)(Xrot/RAD), &q1);
    axis[0] = axis[2] = 0.0;
    axis[1] = 1.0f;
    axis_to_quaternion(axis, (float)(Yrot/RAD), &q2);
    axis[2] = 1.0f;
    axis[1] = axis[0] = 0.0;
    axis_to_quaternion(axis, (float)(Zrot/RAD), &q3);
    qmult(&q3, &q2, &qtemp);
    qmult(&qtemp, &q1, &Rotq);
    free(axis);
    return Rotq;
}

void print_map_info(int Quick)
{
    if (Quick) return;
    fprintf(drvui->flout,"Map Title: %s\n",Map_Info.title);
	fprintf(drvui->flout,"Map Cell Parameters: %7.3f %7.3f %7.3f %7.2f %7.2f %7.2f\n",
		Map_Info.lat_con[0], Map_Info.lat_con[1], Map_Info.lat_con[2],
		Map_Info.lat_con[3], Map_Info.lat_con[4], Map_Info.lat_con[5]);
	fprintf(drvui->flout,"Minimum and Maximum density: %8.1f %8.1f\n",Map_Info.rhomn, Map_Info.rhomx);
	fprintf(drvui->flout,"Map Steps parallel to a, b, c: %5d %5d %5d\n",Map_Info.map_int[0],
		Map_Info.map_int[1], Map_Info.map_int[2]);
	fprintf(drvui->flout,"Map Calculated Region:            Axis     Min       Max\n"
		"                                    x: %8.3f %8.3f\n"
		"                                    y: %8.3f %8.3f\n"
		"                                    z: %8.3f %8.3f\n",
		Map_Info.xlim[0], Map_Info.xlim[1],
		Map_Info.ylim[0], Map_Info.ylim[1],
		Map_Info.zlim[0], Map_Info.zlim[1]);
}

void Progress_Window(int type, char *label, float max)
{

/* Routine to simulate a progress bar. The first argument defines the type of call
 *      type == -1, initialize. The second argument gives the label, 3rd the maximum.
 *      type == -2, close. 2nd and 3rd arguments ignored.
 *      type == 0, 3rd argument contains progress argument. 2nd ignored.
 */

    static char loc_label[1000];
    static float loc_max;
    
    switch (type) {
        
    case -1:
        strcpy(loc_label, label);
        loc_max = max;
        break;
    case -2:
        fprintf(stderr,"%s: 100 percent\n",loc_label);
        break;
    default:
        fprintf(stderr,"%s: %3d percent\r",loc_label,(int) (100.0f * max/loc_max + 0.5f));
    }
}

/* ************************************************************** */
/* ************************************************************** */

void label_cell(void)
{
/* generate the labels corresponding to the unit-cell axes */
  int i, j;
  int used[4] = {0,0,0,0};
  
  if (Labels) {

      if (fabs(offset[0]) + fabs(offset[1]) + fabs(offset[2]) < 0.01)
          Locate_Triple();

/* find any labels that are already in the list */

      if (drvui->nlabel > 1) {
          for (i=1; i<drvui->nlabel; i++) {
            if (!strcmp(drvui->labels[i].label_label,"o")) {
                drvui->triple[0] = 0;
                used[0] = 1;
            }
            if (!strcmp(drvui->labels[i].label_label,"triple_vect")) {
                drvui->triple[0] = i;
                used[0] = 1;
            }
            if (!strcmp(drvui->labels[i].label_label,"a")) {
                drvui->triple[1] = i;
                used[1] = 1;
            }
            if (!strcmp(drvui->labels[i].label_label,"b")) {
                drvui->triple[2] = i;
                used[2] = 1;
            }
            if (!strcmp(drvui->labels[i].label_label,"c")) {
                drvui->triple[3] = i;
                used[3] = 1;
            }
          }
      }

/* following section adds labels to the list for those not already supplied */

      if (!Display_axes) {
          drvui->labels[drvui->nlabel].label_x[0]=-0.5f/drvui->lat_con[0];
          drvui->labels[drvui->nlabel].label_x[1]=-0.5f/drvui->lat_con[1];
          drvui->labels[drvui->nlabel].label_x[2]=-0.5f/drvui->lat_con[2];
          drvui->labels[drvui->nlabel].label_x[2]=1.0f+0.5f/drvui->lat_con[2];
          drvui->labels[drvui->nlabel].label_x[1]=1.0f+0.5f/drvui->lat_con[1];
          drvui->labels[drvui->nlabel].label_x[0]=1.0f+0.5f/drvui->lat_con[0];
      } else {
          drvui->labels[drvui->nlabel].label_x[0]=1.4f/drvui->lat_con[0];
          drvui->labels[drvui->nlabel].label_x[2]=1.4f/drvui->lat_con[2];
          drvui->labels[drvui->nlabel].label_x[1]=1.4f/drvui->lat_con[1];
      }
      if (!used[0]) {
          drvui->labels[drvui->nlabel].label_fn=1;
          if (!Display_axes) {
              drvui->labels[drvui->nlabel].label_x[0]=-0.5f/drvui->lat_con[0];
              drvui->labels[drvui->nlabel].label_x[1]=-0.5f/drvui->lat_con[1];
              drvui->labels[drvui->nlabel].label_x[2]=-0.5f/drvui->lat_con[2];
              drvui->triple[0] = 0;
              strcpy(drvui->labels[drvui->nlabel++].label_label,"o");
          } else {
              drvui->triple[0] = drvui->nlabel;
              drvui->labels[drvui->nlabel].label_x[0] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[1] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[2] = 0.0f;
              strcpy(drvui->labels[drvui->nlabel++].label_label,"triple_vect");
          }
      }
      if (!used[1]) {
          drvui->labels[drvui->nlabel].label_fn=1;
          if (!Display_axes) {
              drvui->labels[drvui->nlabel].label_x[0] = 1.0f+0.5f/drvui->lat_con[0];
              drvui->labels[drvui->nlabel].label_x[1] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[2] = 0.0f;
          } else {
              drvui->labels[drvui->nlabel].label_x[0] = 1.4f/drvui->lat_con[0];
              drvui->labels[drvui->nlabel].label_x[1] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[2] = 0.0f;
          }
          drvui->triple[1] = drvui->nlabel;
          strcpy(drvui->labels[drvui->nlabel++].label_label,"a");
      }
      if (!used[2]) {
          drvui->labels[drvui->nlabel].label_fn=1;
          if (!Display_axes) {
              drvui->labels[drvui->nlabel].label_x[0] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[1] = 1.0f+0.5f/drvui->lat_con[1];
              drvui->labels[drvui->nlabel].label_x[2] = 0.0f;
          } else {
              drvui->labels[drvui->nlabel].label_x[0] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[1] = 1.4f/drvui->lat_con[1];
              drvui->labels[drvui->nlabel].label_x[2] = 0.0f;
          }
          drvui->triple[2] = drvui->nlabel;
          strcpy(drvui->labels[drvui->nlabel++].label_label,"b");
      }
      if (!used[3]) {
          drvui->labels[drvui->nlabel].label_fn=1;
          if (!Display_axes) {
              drvui->labels[drvui->nlabel].label_x[0] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[1] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[2] = 1.0f+0.5f/drvui->lat_con[2];
          } else {
              drvui->labels[drvui->nlabel].label_x[0] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[1] = 0.0f;
              drvui->labels[drvui->nlabel].label_x[2] = 1.4f/drvui->lat_con[2];
          }
          drvui->triple[3] = drvui->nlabel;
          strcpy(drvui->labels[drvui->nlabel++].label_label,"c");
      }
      for (j=0; j<4; j++)
          used[j] = 1;
  }
}

#ifdef OPENGL

static void close_and_exit(void)
{
    FILE * ftmp;
    int i;
    char string[256];

    memset(string,0,256);
    fclose(drvui->fpoutp);
    drvui->fpoutp = fopen(drvui->Pov_File,"r");
    ftmp = fopen("pov.tmp","w");              /* open a temp file */
    while (!feof (drvui->fpoutp)) {
      if (!fgets (string, 256, drvui->fpoutp))        /* read a line */
          break;
      trim_string(string,256);
      if (!strncmp(string,"#declare xrot =",15))
          fprintf(ftmp,"#declare xrot =%7.1f;\n",xrot); /* replace xrot, etc. */
      else if (!strncmp(string,"#declare yrot =",15))
          fprintf(ftmp,"#declare yrot =%7.1f;\n",yrot);
      else if (!strncmp(string,"#declare zrot =",15))
          fprintf(ftmp,"#declare zrot =%7.1f;\n",zrot);
      else
          fprintf(ftmp,"%s\n",string);
    }
    fclose(ftmp);
    fclose(drvui->fpoutp);
    unlink(drvui->Pov_File);
    rename("pov.tmp",drvui->Pov_File);

    fprintf (drvui->flout,
	     "Updated values for xrot, yrot, zrot are %8.2f %8.2f %8.2f\n",
	     xrot, yrot, zrot);
    drvui->fpin = fopen(drvui->Str_File,"r");
    ftmp = fopen("pov.tmp","w");                      /* open a temp file */
    while (!feof (drvui->fpin)) {
      if (!fgets (string, 256, drvui->fpin))        /* read a line */
          break;
      trim_string(string,256);
      fprintf(ftmp,"%s\n",string);
      if (!strncmp(string,"end",3) || !strncmp(string,"END",3))
          break;                                      /* truncate input file after 'end' */
    }
    fclose(drvui->fpin);

    fprintf(ftmp,"rem - following lines indicate (possible) changes/additions to parameters\n");
    fprintf(ftmp,"view %.3f %.3f %.3f\n",xrot, yrot, zrot);
	if (nomits >0) {
		fprintf (drvui->flout, "Sites marked in the GUI: \n");
		fprintf (drvui->flout, "(add lines to input file to keep settings) \n");
        for (i=0;i<nomits;i++) {
		fprintf(drvui->flout,"special %d %d %d\n",omit1[i],omit2[i],omit3[i]);
		fprintf(ftmp,"special %d %d %d\n",omit1[i],omit2[i],omit3[i]);
        }
	}
	if (drvui->nlabel >1 && drvui->labels_moved) {
		fprintf(drvui->flout, "Labels: \n");
		fprintf(drvui->flout,"(these lines may need to be added to the input file)\n");
        for (i=1;i<drvui->nlabel;i++) {
            if (drvui->triple[0] == i)
                 continue;       /* the "triple_vect" is not printed */
	    fprintf(drvui->flout,"labeltext %6.3f %6.3f %6.3f %s\n",drvui->labels[i].label_x[0],
                    drvui->labels[i].label_x[1],drvui->labels[i].label_x[2],drvui->labels[i].label_label);
	    fprintf(ftmp,"labeltext %6.3f %6.3f %6.3f %s\n",drvui->labels[i].label_x[0],
                    drvui->labels[i].label_x[1],drvui->labels[i].label_x[2],drvui->labels[i].label_label);
        }
        if (Display_axes) {
            fprintf(drvui->flout,"vectors %.3f %.3f %.3f\n",offset[0], offset[1], offset[2]);
            fprintf(ftmp,"vectors %.3f %.3f %.3f\n",offset[0], offset[1], offset[2]);
        }
	}			
    fclose(ftmp);
    unlink(drvui->Str_File);
    rename("pov.tmp",drvui->Str_File);
    (void) fclose (drvui->fpoutv);
    (void) fclose (drvui->fpouta);
    (void) fclose (drvui->flout);
    exit (0);
}
#endif 

void *zalloc(size_t size)
{
    void *ret = malloc(size);
    if (ret)
        memset(ret, 0, size);
    return ret;
}
